#ifndef LV_SPEECHPORT_ABNF_H
#define LV_SPEECHPORT_ABNF_H

#include "LV_SRE_ParseTree.h"
#include "LV_SRE_Semantic.h"


/* temporary, until both grammar styles are backed by the same setup */
//#define LV_DECODE_USE_ABNF_GRAMMAR 256 


#ifdef __cplusplus
extern "C"
{
#endif
 
/************************************************************************************************************
 *
 *	functions for creating  ABNF grammars programatically at runtime.
 *
 ************************************************************************************************************/

/*
 * Load a single rule into the grammar set (gram).  calling AddRule(port,LHS,RHS)
 * is equivalent to writing an SRGS-ABNF rule $LHS = RHS;
 *
 * Input: hport		The speech port handle.
 * Input: gram		The grammarset to load into.
 * Input: LHS		The rulename.  May or may not be prefixed by "$"
 * Input: RHS		The rule expansion.
 */
LV_SRE_API int LVSpeechPortABNF_AddRule(HPORT hport, int gram, const char* LHS, const char* RHS);

/*
 * Remove a single rule from the grammar set(gram).
 *
 * Input: hport		The speech port handle.
 * Input: gram		The grammar set to remove the rule from.
 * Input: LHS		The rulename.  May or may not be prefixed by "$"
 */
LV_SRE_API int LVSpeechPortABNF_RemoveRule(HPORT hport, int gram, const char* LHS);

/*
 * Load a grammar file into the grammar set.
 *
 * Input: hport		The speech port handle.
 * Input: gram		The grammar set to load into
 * Input: uri		The location of the grammar file.
 */
LV_SRE_API int LVSpeechPortABNF_LoadGrammar(HPORT hport, int gram, const char* uri);
LV_SRE_API int LVSpeechPortABNF_LoadGrammarFromBuffer(HPORT hport, int gram, const char* label, const char* buffer);

/*
 * Clears out the ABNF grammar from the grammar set (gram).
 */
LV_SRE_API int LVSpeechPortABNF_ResetGrammar(HPORT hport, int gram);

/*
 * Makes one of your grammar rules the root rule, the rule that will be initially active when starting a decode.
 * Equivalent to 
 *		root $Root;
 * In the header of an SRGS-ABNF grammar file.
 * If your root rule is not in the grammar, or you do not call SetRoot, then the active root becomes the union of all
 * rules in the grammar.  You should always set a root rule.
 *
 * Input: hport		The speech port handle.
 * Input: gram		The grammar set being acted on.
 * Input: Root		The name of the new root rule.  May or may not be prefixed by "$"
 */
LV_SRE_API int LVSpeechPortABNF_SetRoot(HPORT hport, int gram, const char* Root);

/*
 * Checks the grammar for errors (left recursion, referenced rulenames that have no expansion).
 */
LV_SRE_API int LVSpeechPortABNF_ValidateGrammar(HPORT hport, int gram);





/************************************************************************************************************
 *
 *	functions for loading or unloading an ABNF grammar into the 
 *	collection of grammars usable by all speechports.  Pre loading a grammar avoids the runtime cost of
 *	compiling a grammar for use in the SRE.
 *                                                              
 ************************************************************************************************************/

/*
 * Load an SRGS-ABNF grammar file into the grammar manager.
 *
 * Input: uri	The location of the grammar you wish to load
 */
LV_SRE_API int LVSpeechPortABNF_LoadGlobalGrammar(const char* uri);

/*
 * Load an SRGS-ABNF grammar file into the grammar manager from a string containing the file contents.
 *
 * Input: label			The location of the grammar you wish to load.  The label should be a uri, but it need not
 *						resolve to any location, or use any known protocol.
 *
 * Input: file_buffer	The contents of the grammar, as a null-terminated string.
 */
LV_SRE_API int LVSpeechPortABNF_LoadGlobalGrammarFromBuffer(const char* label, const char* file_buffer);

/*
 * Unload an SRGS-ABNF grammar file from grammar manager. The grammar is identified by its full URI.
 *
 * Input: uri	the location of the grammar when it was loaded into the manager.
 */
LV_SRE_API int LVSpeechPortABNF_UnloadGlobalGrammar(const char* uri);




/************************************************************************************************************
 *
 *	Functions for Activating/Deactivating grammars for use in a decode.                                                               
 *                                                              
 ************************************************************************************************************/

/*
 * Add an active grammar to be used in the next decode.  If the grammar was preloaded, then a compiled 
 * version of the grammar is retrieved from memory.  Otherwise, the grammar is pulled from the uri
 * location and compiled.
 * 
 * Input: hport			The speechport that will use this active grammar set.
 * Input: uri			the uri descriptor of the grammar to be activated.
 */
LV_SRE_API int LVSpeechPortABNF_ActivateGlobalGrammar(HPORT hport, const char* uri);

/*
 * Add an active grammar to be used in the next decode.  The grammar is retrieved from the speechports 
 * set of modifiable grammars.
 * 
 * Input: hport			The speechport that will use this active grammar.
 * Input: grammarset	the id of the grammar to be activated.
 */
LV_SRE_API int LVSpeechPortABNF_ActivateGrammar(HPORT hport, int grammarset);

/*
 * Clears out the active grammar set.
 * 
 * Input: hport			The speechport that owns the active grammar set.
 */
LV_SRE_API int LVSpeechPortABNF_DeactivateGrammars(HPORT hport);


/************************************************************************************************************
 *
 *	Alternate Decode function, flags, etc.
 *
 ************************************************************************************************************/

/*
 *	Decode Flags (defined in LV_SRE.h):
 *
 *		LV_DECODE_USE_ABNF_GRAMMAR		
 *		Tells the decode to use Lumenvox's implementation of SRGS/ABNF grammars.  The return value of
 *		an ABNF decode is a parse string and parse tree compatible with the suggested return type laid out
 *		by the W3Cs Speech Recognition Grammar Specification.
 *
 *		LV_DECODE_SEMANTIC_INTERPRETATION
 *		If this is set, The speechport will use the parse tree and tag information in the tree to return
 *		an XML string compatible with W3Cs Semantic Interpretation working draft.  In the future, it will
 *		also return a C data structure representation of the semantic interpretation tree, to make it 
 *		easier to transform the semantic interpretation into results like ECMAScript source, EMMA, NLSML, etc.
 *
 *	Decode variables:
 *
 *		LV_ACTIVE_GRAMMAR_SET
 *		During the call to LV_SRE_Decode, if this variable is set for the "grammarset" variable, then your
 *		set of active grammars will be used.  Otherwise, only one of your speechports grammars will be used.
 *
 *
 *	eg: decoding with all the bells and whistles:
 *		LV_SRE_Decode(port, voicechannel, LV_ACTIVE_GRAMMAR_SET, LV_DECODE_USE_ABNF_GRAMMAR | 
 *		                                                         LV_DECODE_SEMANTIC_INTERPRENTATION);
 */

/*
 * Decode using SRGS ABNF grammar, and the active grammar set (convenience function).
 * Note:  the flag LV_DECODE_USE_ABNF_GRAMMAR and 
 *        the specifier LV_ACTIVE_GRAMMAR_SET are not needed with this function.
 *        LV_DECODE_SEMANTIC_INTERPRETATION is still needed for the interpretation XML string.
 *
 * Input: hport			LVSpeechport handle
 * Input: VoiceChannel	voice channel containing the audio for decode;
 * Input: flags			Decoding flags
 */
LV_SRE_API int LVSpeechPortABNF_Decode(HPORT hport, int VoiceChannel, unsigned int flags);
LV_SRE_API int LVSpeechPortABNF_DTMFDecode(HPORT hport, const char* dtmf_str, unsigned int flags);




/************************************************************************************************************
 *
 *	functions for retrieving information after a decode that used an ABNF grammar.
 *
 ************************************************************************************************************/

/* 
 * returns the number of parses for the SRE output.
 * If the value is zero, then the decode didn't match the active grammar.
 * If the value is greater than one, then the active grammar(s) were ambiguous for the decode.
 */
LV_SRE_API int LVSpeechPortABNF_GetNumberOfParses(HPORT hport, int VoiceChannel);

/*
 * Returns an utterance score for the decoded audio.  Higher scores are more likely to be correct than
 * lower scores.  The range of the score is 0 - 1000.
 *
 * Input: hport			The speechport used in the decode.
 * Input: VoiceChannel	The voice channel used in the decode.
 */
LV_SRE_API int LVSpeechPortABNF_GetUtteranceScore(HPORT hport, int VoiceChannel);

/*
 * Returns a logical parse string for the decode.
 *
 * See "http://www.w3.org/TR/speech-grammar/#AppH"
 *
 * Input: hport			The speechport used in the decode.
 * Input: VoiceChannel	The voice channel used in the decode.
 * Input: Index			A value from 0 to GetNumberOfParses - 1, indexing the parse trees for the decode.
 */
LV_SRE_API const char* LVSpeechPortABNF_GetParseTreeString(HPORT hport, int VoiceChannel, int Index);

/*
 * Logically equivalent to GetParseString, A parse tree handle allows traversal of the parse tree, 
 * letting the user see word confidence scores, start and end frames for each word, 
 * and phonetic pronunciations.
 *
 * See <LV_SRE_ParseTree.h>
 *
 * Input: hport			The speechport used in the decode.
 * Input: VoiceChannel	The voice channel used in the decode.
 * Input: Index			A value from 0 to GetNumberOfParses - 1, indexing the parse trees for the decode.
 */
LV_SRE_API H_SPT LVSpeechPortABNF_GetParseTreeHandle(HPORT hport, int VoiceChannel, int Index);

/*
 * Get semantic interpretation result in XML format
 * See "http://www.w3.org/TR/semantic-interpretation/#SI7."
 *
 * Input:	hport			The port used in the decode.
 *			voicechannel	The audio channel that the port used during decode.
 *			index			The index of the interpretation.  There is one semantic interpretation result
 *                          per parse tree, so the index ranges from 
 *							0 to LVSpeechPortABNF_GetNumberOfParses - 1 
 *
 * Note:  The flag LV_DECODE_SEMANTIC_INTERPRETATION must be set if you want to recieve 
 *        semantic interpretation results.
 *
 * ToDo:  GetInterpretationHandle(port, voicechannel, index, outputhandle) to return a C datastructure 
 *		  representation of the semantic interpretation object.
 *       
 */
LV_SRE_API const char* LVSpeechPortABNF_GetInterpretationString(HPORT hport, int VoiceChannel, int Index);

/*
 * Get a handle to Semantic Data. 
 *
 * Input:	hport			The port used in the decode.
 *			voicechannel	The audio channel that the port used during decode.
 *			index			The index of the interpretation.  There is one semantic interpretation result
 *                          per parse tree, so the index ranges from 
 *							0 to LVSpeechPortABNF_GetNumberOfParses - 1 
 *
 * Note:  The flag LV_DECODE_SEMANTIC_INTERPRETATION must be set if you want to recieve 
 *        semantic interpretation data.      
 */
LV_SRE_API H_SI_DATA LVSpeechPortABNF_GetInterpretationData(HPORT hport, int VoiceChannel, int Index);

/*
 * Get the list of grammars used in this decode.
 */
LV_SRE_API int LVSpeechPortABNF_GetNumberOfGrammarsUsedInDecode(HPORT hport, int VoiceChannel);
LV_SRE_API const char* LVSpeechPortABNF_GetDecodeGrammarLabel(HPORT hport, int VoiceChannel,int index);

#ifdef __cplusplus
}
#endif


#endif
