#include <math.h> // log10
#include <stdio.h>   // FILE
#include <io.h>	// access
#include <fstream>
using std::fstream;
#include <string>
using std::string;

#include "..\escutil\iv_str.h"   // FindCaseInsensitiveString
#include "..\TFServer\TFTCPClient_communication.h"
#ifdef TF_CLOSED_SOURCE
#include "..\tts\tts_code.h"
#endif

#include "brd_mt_code.h"

/* -------------------------------------------------------------------------
   Name:        CreateInstance

   Purpose:     This function must be implemented in the board DLL to create
                an instance of the specific board class.

                The board class must derive from the "BoardGeneric" class.

                TeleFlow server will call the CreateInstance function in the
                board DLL to create a new instance of the specific board
                class.
   ------------------------------------------------------------------------- */
BoardGeneric* CreateInstance(PIVRINFO pInfo)
{
   BoardGeneric* gen;

   gen = new BoardMT(pInfo);
   return gen;
}
/*CreateInstance*/

/* -------------------------------------------------------------------------
   Name:        ReleaseInstance
   ------------------------------------------------------------------------- */
void ReleaseInstance(BoardGeneric* brd)
{
   delete (BoardMT*) brd;
}
/*ReleaseInstance*/

// --------------------------------------------------------------------------
//  Name:       OpenHardwareDriver
// --------------------------------------------------------------------------
bool OpenHardwareDriver(Log* sysLog)
{
   HANDLE hProc;                  // retrieve handle of current process
   DWORD dwCurrentPriority;
   SYSTEM_INFO sysInfo;           // Win32 API system info structure
   DWORD iCount;                  // count of CPU cores in system
   DWORD dwProcessAffinityMask;   // current process affinity (retrieved and then modified)
   DWORD dwSystemAffinityMask;    // system affinity
   INT iShift;                    // bitwise shift to set affinity to only some of the cores
   bool bSuccess;                 // result from Win32 API calls

   // Retrieve pseudohandle to the current process
   hProc = GetCurrentProcess();

   // Reverse the process priority and CPU affinity settings made by TFServer
   // Current Ai-Logix support uses the Windows default
   GetSystemInfo(&sysInfo);
   iCount = sysInfo.dwNumberOfProcessors;
   sysLog->Print("   Number of processors: %i \n", iCount);

   if(iCount > 1)
   {
      bSuccess = GetProcessAffinityMask(hProc, &dwProcessAffinityMask, &dwSystemAffinityMask);
      if(bSuccess)
      {
         // Set affinity to all cores in the system
         sysLog->Print("   Process affinity mask 0x%x \n", dwProcessAffinityMask);
         sysLog->Print("   System affinity mask 0x%x \n", dwSystemAffinityMask);
         dwProcessAffinityMask = dwSystemAffinityMask;
         bSuccess = SetProcessAffinityMask(hProc, dwProcessAffinityMask);
         if(bSuccess)
         {
            sysLog->Print("   Processor affinity set to all cores\n");
         }
      }

      if(!bSuccess)
      {
         sysLog->Print("   WARNING: Failed to set process affinity\n");
         return(false);
      }
   }

   // Set process priority to NORMAL
   dwCurrentPriority = GetPriorityClass(hProc);
   // sysLog->Print("   Current Priority = 0x%x\n", dwCurrentPriority);
   // sysLog->Print("      NORMAL_PRIORITY_CLASS   = 0x%x\n", NORMAL_PRIORITY_CLASS);
   // sysLog->Print("      HIGH_PRIORITY_CLASS     = 0x%x\n", HIGH_PRIORITY_CLASS);
   // sysLog->Print("      REALTIME_PRIORITY_CLASS = 0x%x\n", REALTIME_PRIORITY_CLASS);
   //
   if(dwCurrentPriority != NORMAL_PRIORITY_CLASS)
   {
      sysLog->Print("   Set process priority class to NORMAL on Ai-Logix\n");

      if(SetPriorityClass(hProc, NORMAL_PRIORITY_CLASS))
      {
         sysLog->Print("   Process priority class set to NORMAL\n");
      }
      else
      {
         sysLog->Print("   WARNING: Failed to set process priority\n");
      }
   }
   else
   {
      sysLog->Print("   Current priority class (0x%x) is OK on Ai-Logix\n", dwCurrentPriority);
   }

   // Call SmartWorks system startup
   if(MTSysStartup()!= MT_RET_OK)
   {
      MT_RESULT error;
      unsigned short errorMessageLength = 127;
      char errorMessage[128];

      memset(errorMessage, 0, 128);
      MTGetLastError(&error, &errorMessageLength, errorMessage);
      sysLog->Print("ERROR: Failed to start Ai-Logix Hardware\n");
      sysLog->Print("      %X: %s\n", error, errorMessage);

      return(false);
   }

   return(true);
}
//OpenHardwareDriver

// --------------------------------------------------------------------------
//  Name:       CloseHardwareDriver
// --------------------------------------------------------------------------
bool CloseHardwareDriver(Log* sysLog)
{
   // Call SmartWorks system shutdown
   sysLog->Print("   Shutdown all telephony channels\n");
   if(MTSysShutdown()!= MT_RET_OK)
   {
      MT_RESULT error;
      unsigned short errorMessageLength = 127;
      char errorMessage[128];

      memset(errorMessage, 0, 128);
      MTGetLastError(&error, &errorMessageLength, errorMessage);
      sysLog->Print("ERROR: Failed to stop Ai-Logix Hardware\n");
      sysLog->Print("      %X: %s\n", error, errorMessage);

      return(false);
   }

   // AudioCodes recommends waiting to give time for MTSysShutdown to complete
   // Windows SCM will allow us to sleep for 20 seconds, less time it took to get to this point
   sysLog->Print("   Waiting 18 seconds...\n");
   Sleep(18000);
   sysLog->Print("   Exiting\n");

   return(true);
}
//CloseHardwareDriver

CINTERFACE void MTEventCallback(CHANNEL eventChannel, int eventCount, MT_EVENT* events, LPARAM params)
{
   if(params == NULL)
   {
      return;
   }

   CALLBACK_INFO* info = (CALLBACK_INFO*) params;
   CHANNEL channel = info->channel;
   SignalQueue<MT_EVENT>* eventQueue = info->eventQueue;
   SignalQueue<MT_EVENT>* callTriggerQueue = info->callTriggerQueue;
   PIVRINFO pInfo = info->pInfo;
   AiLAudioState* audioState = info->audioState;
   GUILog* ivLog = pInfo->ivLog;
   BoardMT* ivBoard = (BoardMT*)(pInfo->ivBoard);
   MT_EVENT event;

   if(ivBoard == NULL)
   {
      // Cannot process events until board object is fully created
      return;
   }

   if(eventChannel != channel)
   {
      ivLog->Print("%d events came from port %d\n", eventCount, eventChannel + 1);
      return;
   }

   for(int counter = 0; counter < eventCount; counter++)
   {
      event = events[counter];

      // Process EVT_STREAMIN_WATERMARK and EVT_STREAMIN_DROPPED asynchronously
      if((event.EventCode == EVT_STREAMIN_WATERMARK) || (event.EventCode == EVT_STREAMIN_DROPPED))
      {
         audioState->UpdateStreams(channel);
      }
      // Throw away digits while dialing
      else if(audioState->IsDialing() && (event.EventCode == EVT_DIGIT))
      {
         unsigned char digit;
         // Remove digit from board queue
         MTReadDigit(eventChannel, &digit);
      }
      else if(audioState->IsDialing() && ((event.EventCode == EVT_ONHOOK) || (event.EventCode == EVT_LOOP_DROP) || (event.EventCode == EVT_REVERSAL)))
      {
         // Ignore these events while dialing
      }
      else
      {
         // DEBUG: Let the event know what position it will have in the queue
         // 'UserStatus' is a user-defined field provided by SmartWorks API
         event.UserStatus = eventQueue->Size() + 1;
         BoardMT::T_ACTIVITY_MODE actMode = ivBoard->CurrentActivityMode();
         // Control how events are queued, and control management of event queues
         if(actMode == BoardMT::PASSIVE)
         {
            // Only Call Trigger event queue is active at this time
            callTriggerQueue->Add(event);
         }
         else if(actMode == BoardMT::ACTIVE)
         {
            // Both Call Trigger and board event queues are active at this time
            callTriggerQueue->Add(event);
            eventQueue->Add(event);
         }
         else if(actMode == BoardMT::ACTIVE_ONLY)
         {
            // Only board event queue is active in this mode
            eventQueue->Add(event);
         }
         // NOTE: No limit on size of queues
         // All events that are queued MUST be processed by the application, and
         // during normal operation, they will be. ACTIVE/PASSIVE queues are only
         // populated when the system is in the appropriate mode. Queue size will
         // only become a problem if something is preventing the application from
         // responding in a timely manner. IF that should happen, address the real
         // issue; for example:
         //    When the process priority of the TeleFlow thread was set to HIGH
         //    and core affinity was set to 1 to half of the cores in the system.
         //    The Ai-Logix thread had higher priority and so this call-back function
         //    was adding events faster than the main application could process them.
         //    This has been fixed in build 2010.5.13 and later

         if(event.EventCode == EVT_MFTONE)
         {
            unsigned char mfTone;
            unsigned char toneDirection;
            MTReadMFTone(eventChannel, &mfTone, &toneDirection);
         }
      }
   }
};

AiLAudioState::AiLAudioState(CommunicationHub<TFServerCommunicationPacket::TF_DESTINATION_ID>* commHub) : AudioState(commHub)
{
   Reset();
}

unsigned long AiLAudioState::AudioBufferSize(SOUND_FORMAT format)
{
	switch(format)
   {
   	case TF_WAV_8KHZ_16BIT:	return(3200);	// WAV format- 8000 sampled rate and 16bits
   	case TF_WAV_8KHZ_16BIT_G610:	return(325);	   // Microsoft GSM
      default: return(3200);
   }
}

bool AiLAudioState::IsVoiceRecorded(void)
{
   return(isVoiceRecorded);
}

void AiLAudioState::SetVoiceRecorded(bool voiceRecorded)
{
   isVoiceRecorded = voiceRecorded;
}

void AiLAudioState::ResetVoiceRecorded(void)
{
   isVoiceRecorded = false;
}

bool AiLAudioState::IsDialing()
{
   return(isDialing);
}

void AiLAudioState::StartDialing()
{
   isDialing = true;
}

void AiLAudioState::StopDialing()
{
   isDialing = false;
}

void AiLAudioState::Reset(void)
{
   isVoiceRecorded = false;
   isDialing = false;
   streamFormat = TF_WAV_8KHZ_16BIT_G610;
   if(recordFile.is_open())
   {
      recordFile.close();
   }
   AudioState::Reset();
}

bool AiLAudioState::StartRecording(const char* fileName, SOUND_FORMAT format)
{
   recordFile.open(fileName, std::ios::in | std::ios::out | std::ios::trunc | std::ios::binary);

   if(!recordFile.is_open())
   {
      return(false);
   }

   if(!AudioState::StartRecording(fileName, format))
   {
      if(recordFile.is_open())
      {
         recordFile.close();
      }
      return(false);
   }

   // Make a header for zero length file
   switch(RecordFormat())
   {
      // WAV format - 8000 sampled rate and 16bits
      case TF_WAV_8KHZ_16BIT:
      {
         PCM_WAVE_HEADER header;
         MakeWavePCMHeader(&header, 0, MT_PCM_Raw_16bit);
         recordFile.write((const char*) &header, sizeof(PCM_WAVE_HEADER));
         break;
      }

      // Microsoft GSM
      case TF_WAV_8KHZ_16BIT_G610:
      {
         GSM_WAVE_HEADER header;
         MakeWaveGSMHeader(&header, 0);
         recordFile.write((const char*) &header, sizeof(GSM_WAVE_HEADER));
         break;
      }
   }

   return(true);
}

void AiLAudioState::StopRecording(void)
{
   int recordingSize;

   if(recordFile.is_open())
   {
      recordFile.flush();

      // Determine the size of the audio
      recordingSize = recordFile.tellp() - std::ios::beg + 1;

      // This is the position of the byte most recently written not the byte about to be written
      recordFile.seekp(std::ios::beg - 1);

      // Write a wave header for size of audio data less the size of the wave header
      // since the wave header will overwrite the first little bit of audio data
      switch(RecordFormat())
      {
         // WAV format - 8000 sampled rate and 16bits
         case TF_WAV_8KHZ_16BIT:
         {
            PCM_WAVE_HEADER header;
            MakeWavePCMHeader(&header, recordingSize - sizeof(PCM_WAVE_HEADER), MT_PCM_Raw_16bit);
            recordFile.write((const char*) &header, sizeof(PCM_WAVE_HEADER));
            break;
         }

         // Microsoft GSM
         case TF_WAV_8KHZ_16BIT_G610:
         {
            GSM_WAVE_HEADER header;
            MakeWaveGSMHeader(&header, recordingSize - sizeof(GSM_WAVE_HEADER));
            recordFile.write((const char*) &header, sizeof(GSM_WAVE_HEADER));
            break;
         }
      }

      recordFile.close();
   }

   AudioState::StopRecording();
}

bool AiLAudioState::UpdateStreams(CHANNEL curChannel)
{
   unsigned long length = AudioBufferSize(IsRecording() ? RecordFormat() : StreamFormat());
   MT_RESULT result = MTStreamBufIn(curChannel, 0, audioBuffer.buffer, &length);

   if(result != MT_RET_OK)
   {
      return(false);
   }

   if(IsRecording() && (recordFile.is_open()))
   {
      recordFile.write((const char*) audioBuffer.buffer, length);
   }

   if(IsStreamingPermitted())
   {
      audioBuffer.format = IsRecording() ? RecordFormat() : StreamFormat();
      StreamAudio((const char*) &audioBuffer, sizeof(SOUND_FORMAT) + length);
   }

   return(true);
}

MTBoardEvent::MTBoardEvent(void)
{
	// Do nothing
}

MTBoardEvent::~MTBoardEvent(void)
{
	// Do nothing
}

bool MTBoardEvent::SameEvent(BoardEvent* comparisonEvent)
{
	return(MTEvent.EventCode == ((MTBoardEvent*) comparisonEvent)->MTEvent.EventCode);
}

BoardMT::BoardMT(PIVRINFO pIVRInfo) : BoardGeneric(pIVRInfo)
{
   short adapter;
   short channelOnAdapter;
   MTADAPTER_XINFO adapterInfo;
   unsigned long adapterInfoSize = sizeof(MTADAPTER_XINFO);
	unsigned short eventMask;
   unsigned long monitorEventMask;
   MT_PSTN pstnParams;
   unsigned long pstnParamsSize = sizeof(MT_PSTN);
   MT_ACTPARAMS actParams;            // activity detection parameters
   unsigned long actParamsSize = sizeof(MT_ACTPARAMS);
   MT_VOICEDETECTPARAMS voiceDetectParams;
   unsigned long voiceDetectParamsSize = sizeof(MT_VOICEDETECTPARAMS);
   char brd_tt;      // flush digits out of Ai-logix queue & display them in the log

   preferredFormat = TF_WAV_8KHZ_16BIT;

   if(strcmp(pInfo->pszProtocol, "nocc") == 0)
   {
      ivLog->Print("   Set activity mode to Active/Passive\n");
      SetActivityMode(ACTIVE);   // active to start, allowed to change to passive
   }
   else
   {
      ivLog->Print("   Set activity mode to Active Only\n");
      SetActivityMode(ACTIVE_ONLY); // active only, not allowed to change
   }

   // Warning: Be careful to delete base class' audioState before overwriting
   // it with the derived class' audioState
   delete BoardGeneric::audioState;
   BoardGeneric::audioState = audioState = new AiLAudioState(pInfo->commHub);

   // Assume success
 	bBrdInit = true;

	// Initialise curLine
	curLine = (pInfo->iPort < 1) ? -1 : pInfo->iPort - 1;

   if(pInfo->iPort > 0)
   {
	   cs.Enter();

   	// Initialise curChannel
		if(MTGetGCI(0, 0, &curChannel)!= MT_RET_OK)
	   {
   	  	ivLog->Print("ERROR: Failed to get GCI\n");
         PrintMTError();
       	bBrdInit = false;
	   }
      curChannel += (CHANNEL) (pInfo->iPort - 1);
      otherChannel = -1;		// by default current port is not in a Switched state

		if(MTOpenChannel(curChannel)!= MT_RET_OK)
	   {
   	  	ivLog->Print("ERROR: Failed to open port\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      // Get adapter information
		if(MTGetGCIMap(curChannel, &adapter, &adapterType, &channelOnAdapter)!= MT_RET_OK)
	   {
   	  	ivLog->Print("ERROR: Failed to get GCI Map\n");
         PrintMTError();
       	bBrdInit = false;
	   }

		// Set the wait object handle for the MT board event queue
   	waitObjects[4] = eventQueue.GetSignal();

	   // Stop all activity on this channel of the board
		if(MTStopChannel(curChannel)!= MT_RET_OK)
	   {
   	  	ivLog->Print("ERROR: Failed to stop MT channel events\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      // Set the board event queue mask
   	eventMask = SE_LCURRENT_CHANGE	// Enables EVT_LOOP_DROP, EVT_LOOP_ON
   				 | SE_RING	            // Enables EVT_RINGS
         	    | SE_ACT	            // Enables EVT_MON_ACTIVITY, EVT_MAX_ACTIVITY
         	    | SE_SIL	            // Enables EVT_MON_SILENCE, EVT_MAX_SILENCE
         	    | SE_OFFH	            // Enables EVT_OFFHOOK
            	 | SE_ONH 	            // Enables EVT_ONHOOK
   	          | SE_LREV	            // Enables EVT_REVERSAL
   	          | SE_LVOLTAGE_CHANGE;	// Enables EVT_LVOLTAGE_OFFHOOK, EVT_LVOLTAGE_NOTOFFHOOK

		// Set board events
	   if(MTSetEventFilters(curChannel, eventMask, 1) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to set MT events mask\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      // Set the configuration for activity detection
      if(MTChInputGetACTDParams(curChannel, 0, &actParams, &actParamsSize) !=  MT_RET_OK)
      {
         ivLog->Print("ERROR: Failed to get MT activity detection parameters\n");
         PrintMTError();
         bBrdInit = false;
      }
      actParams.min_silence = 200;       // minimum silence deglitch time 200 ms
      actParams.min_activity = 200;      // minimum activity deglitch time 200 ms
      if(MTSetACTDParams(curChannel, &actParams, &actParamsSize) !=  MT_RET_OK)
      {
         ivLog->Print("ERROR: Failed to set MT activity detection parameters\n");
         PrintMTError();
         bBrdInit = false;
      }
      ivLog->Print("   Activity detection parameters:\n");
      ivLog->Print("      threshold_low = %5.2f\n", actParams.threshold_low);
      ivLog->Print("      threshold_high = %5.2f\n", actParams.threshold_high);
      ivLog->Print("      min_silence = %i\n", actParams.min_silence);
      ivLog->Print("      max_silence = %i\n", actParams.max_silence);
      ivLog->Print("      min_activity = %i\n", actParams.min_activity);
      ivLog->Print("      max_activity = %i\n", actParams.max_activity);

      // Set board event signal callback function
      callbackParams.channel = curChannel;
      callbackParams.eventQueue = &eventQueue;
      callbackParams.callTriggerQueue = &callTriggerQueue;
      callbackParams.pInfo = pInfo;
      callbackParams.audioState = audioState;
      //
	   if(MTSetEventCallback(curChannel, MTEventCallback, (LPARAM) &callbackParams) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to set MT event callback function\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      if(EnableReorderToneDetection() != IVR_SUCCESS)
      {
         bBrdInit = false;
      }

		// Set the monitor event mask
   	monitorEventMask =  MONI_BUSY1		// Enables EVT_MON_BUSY
                        | MONI_BUSY2		// Enables EVT_MON_BUSY_TRUNK
                        | MONI_DIAL		   // Enables EVT_MON_DIAL
                        | MONI_CALLWAITING
                     	| MONI_RINGBACK1	// Enables EVT_MON_RINGBACK
                     	| MONI_RINGBACK2	// Enables EVT_MON_DRINGBACK
//                        | MONI_SIGNAL_CYCLE
   	                  | MONI_REVERSAL	// Enables EVT_REVERSAL
                        | MONI_VOICE
                        | MONI_SIT1
                        | MONI_SIT2
                        | MONI_SIT3
                        | MONI_SIT4
                        | MONI_SIT5
                        | MONI_DIAL1
                        | MONI_DIAL2
                        | MONI_DIAL3
                        | MONI_FAX1
                        | MONI_FAX2
                        | MONI_UTONE;

		// Set monitor events
	   if(MTSetMoni(curChannel, monitorEventMask) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to set MT monitor events mask\n");
         PrintMTError();
       	bBrdInit = false;
	   }

	   if(MTChGetVoiceDetectParams(curChannel, &voiceDetectParams, &voiceDetectParamsSize) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to get MT voice parameters\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      // The default value of 100 ms causes premature voice detection during PlaceCall
      voiceDetectParams.voice_min = 200;

	   if(MTChSetVoiceDetectParams(curChannel, &voiceDetectParams, &voiceDetectParamsSize) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to set MT voice parameters\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      // Set ABCD bit signaling
      if(MTGetAdapterXInfo(adapter, &adapterInfo, &adapterInfoSize) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to get adapter extra info\n");
         PrintMTError();
       	bBrdInit = false;
	   }
      else
      {
      	MT_RESULT result = MT_RET_OK;

         for(unsigned short framerIndex = 0; (framerIndex < adapterInfo.NumFramers) && (result == MT_RET_OK); framerIndex++)
         {
            result = MTEnableFramerABCDSignalingEvent(adapter, framerIndex);
         }

         if((result != MT_RET_OK) && (result != MT_RET_NOT_AVAILABLE))
         {
  	   	   ivLog->Print("WARNING: Failed to enable ABCD signalling bits\n");
            PrintMTError();
            bBrdInit = false;
         }
      }

	   if(MTGetPSTNParams(curChannel, &pstnParams, &pstnParamsSize) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to get PSTN defaults\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      // Set default PSTN
      pstnParams.flash_char				= '&';	// default is '&'
      pstnParams.flash_time				= 500;	// default is 500 ms
      pstnParams.wink_delay				= 150;	// default is 150 ms
      pstnParams.wink_time					= 150;	// default is 150 ms
      pstnParams.pause_char				= ',';	// default is ','
      pstnParams.pause_time				= 2000;	// default is 2000 ms
      pstnParams.pulse_on					= 40;		// default is 40 ms
      pstnParams.pulse_off					= 60;		// default is 60 ms
      pstnParams.pulse_delay_time		= 1000;	// default is 1000 ms
      // Ai-Logix recommends 250 ms instead of the default
      // pstnParams.loop_deglitch      = 250;   // default is 50 ms
      pstnParams.loop_deglitch         = 750;   // attempting to solve flash-hook problems for CET - 2010/10/05
      pstnParams.loop_delay				= 10;		// default is 10 ms
      pstnParams.ring_on_time				= 300;	// default is 300 ms
      pstnParams.ring_off_time			= 500;	// default is 500 ms
      pstnParams.ring_reset				= 8000;	// default is 8000 ms
      pstnParams.ring_count				= 1;		// default is 1
      pstnParams.ring_trigger				= 0;		// default is 0
      pstnParams.ring_autoanswer			= 0;		// default is 0 (0 = disabled, 1 = enabled)
      pstnParams.cpm_start_delay			= 1000;	// default is 1000 ms
      pstnParams.offhook_delay			= 500;	// default is 500 ms
      pstnParams.min_rings_no_answer	= 4;		// default is 4
      pstnParams.max_sig_time				= 6500;	// default is 6500 ms
      pstnParams.max_sil_time				= 40000;	// default is 40000 ms

		// Set PSTN defaults
	   if(MTSetPSTNParams(curChannel, &pstnParams, &pstnParamsSize) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to set PSTN defaults\n");
         PrintMTError();
       	bBrdInit = false;
	   }

		// Set monitor mode default to monitor and terminate
      if(MTChCPMControl(curChannel, CPMMODE_MONI) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to set CPM mode\n");
         PrintMTError();
       	bBrdInit = false;
	   }

  		// Set streaming buffer size to thrice the audio buffer size
      // If the streaming buffer is too large than audio gets missed at the end of recordings
	   if(MTSetStreamingConfig(curChannel, 3*audioState->AudioBufferSize()) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to set streaming buffer size\n");
         PrintMTError();
       	bBrdInit = false;
	   }

      switch(adapterType)
      {
         case VOICE_RESOURCE:             // For channels on VR6400, VR3200, VR6409, VR3209
         case ANALOG_VDS:                 // For channels on PT1600, PT800, PT1609, PT809
         case DIGITAL_PASSIVE_T1_NONE:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
         case DIGITAL_PASSIVE_T1_ISDN:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
         case DIGITAL_PASSIVE_T1_RBS:     // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
         case DIGITAL_PASSIVE_T1_NFAS:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
         case DIGITAL_PASSIVE_E1_NONE:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
         case DIGITAL_PASSIVE_E1_ISDN:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
         case DIGITAL_PASSIVE_E1_CAS:     // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
         case DIGITAL_PASSIVE_E1_NFAS:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
         case DIGITAL_PASSIVE_E1_DASS2:   // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
         case DIGITAL_PASSIVE_NGX:        // For channels on Passive NGX
         {
            // Nothing to do for passive boards
            break;
         }

         case ANALOG_FXO_LS:              // For channels on AT1600, AT800, AT1609, AT809
         case ANALOG_FXS_LS_E:            // For station channels on AT4X12
         case ANALOG_VFXO_LS_E:           // For channels on LD809, LD409, LD2409, LD1609
         case ANALOG_VFXS_LS_E:           // For station channels on LD101, LD201
         case DIGITAL_TERMINATE_T1_NONE:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         case DIGITAL_TERMINATE_T1_ISDN:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         case DIGITAL_TERMINATE_T1_RBS:   // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         case DIGITAL_TERMINATE_T1_NFAS:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         case DIGITAL_TERMINATE_E1_NONE:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         case DIGITAL_TERMINATE_E1_ISDN:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         case DIGITAL_TERMINATE_E1_CAS:   // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         case DIGITAL_TERMINATE_E1_NFAS:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
         {
   		   // Put this channel is on hook
   	      if(MTOnHook(curChannel) != MT_RET_OK)
      	   {
           		ivLog->Print("ERROR: Failed to place MT channel on hook\n");
               PrintMTError();
          	   bBrdInit = false;
   	      }
         }
      }

      // Clear the board digit queue for this channel
      // This algorithm is better than MTClearDTMFDigits() because we can output each digit to the log
      ivLog->Print("   Flush board-level digit queue\n");
      brd_tt = -1;
      while(brd_tt != 0)
      {
         MTReadDigit(curChannel, &((unsigned char) brd_tt));
         if(brd_tt != NULL)
         {
            ivLog->Print("      digit '%c'\n", brd_tt);
         }
         else
         {
            ivLog->Print("      end of digits\n");
         }
      }

   	// Clear the board event queue for this channel
      ivLog->Print("   Flush events\n");
	   if(MTFlushEvents(curChannel) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to flush MT channel events\n");
         PrintMTError();
       	bBrdInit = false;
	   }

	   cs.Leave();
   }

   ANILength = 0;

   if(bBrdInit)
   {
      ivLog->Print("Driver successfully opened\n");
      ivLog->Print("   Adapter Type 0x%x\n", adapterType);
      if(pInfo->iPort > 0)
      {
         switch(adapterType)
         {
            case VOICE_RESOURCE:             // For channels on VR6400, VR3200, VR6409, VR3209
            {
               trunkType = TF_TRUNK_TYPE_NONE;
               break;
            }
            case ANALOG_VDS:                 // For channels on PT1600, PT800, PT1609, PT809
            case ANALOG_FXO_LS:              // For channels on AT1600, AT800, AT1609, AT809
            case ANALOG_FXS_LS_E:            // For station channels on AT4X12
            case ANALOG_VFXO_LS_E:           // For channels on LD809, LD409, LD2409, LD1609
            case ANALOG_VFXS_LS_E:           // For station channels on LD101, LD201
            {
               trunkType = TF_TRUNK_TYPE_ANALOG;
               break;
            }
            case DIGITAL_PASSIVE_T1_NONE:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_PASSIVE_T1_ISDN:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_PASSIVE_T1_RBS:     // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_PASSIVE_T1_NFAS:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_TERMINATE_T1_NONE:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_T1_ISDN:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_T1_RBS:   // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_T1_NFAS:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            {
               trunkType = TF_TRUNK_TYPE_T1;
               break;
            }
            case DIGITAL_PASSIVE_E1_NONE:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_ISDN:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_CAS:     // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_NFAS:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_DASS2:   // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_TERMINATE_E1_NONE:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_E1_ISDN:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_E1_CAS:   // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_E1_NFAS:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            {
               trunkType = TF_TRUNK_TYPE_E1;
               break;
            }
            case DIGITAL_PASSIVE_NGX:        // For channels on Passive NGX
            {
               trunkType = TF_TRUNK_TYPE_STATION;
               break;
            }
            default:
            {
               if((adapterType | ANALOG_TERMINATE_CARD) ||
                  (adapterType | ANALOG_PASSIVE_CARD) ||
                  (adapterType | ANALOG_CARD))
               {
                  trunkType = TF_TRUNK_TYPE_ANALOG;
               }
               else
               {
                  trunkType = TF_TRUNK_TYPE_INVALID;
               }
            }
         }
      }
   }
   else
   {
      ivLog->Print("Driver failed to open successfully\n");
   }
}

BoardMT::~BoardMT(void)
{
   if(pInfo->iPort > 0)
   {
	   // Reset any previous switch states that may have been created by this thread
	   UnSwitch();

      // Reset board event signal callback function
	   if(MTClearEventCallback(curChannel) != MT_RET_OK)
   	{
     		ivLog->Print("ERROR: Failed to clear MT event callback function\n");
         PrintMTError();
       	bBrdInit = false;
	   }

	   // Stop all activity on this channel of the board
		if(MTStopChannel(curChannel)!= MT_RET_OK)
      {
         ivLog->Print("ERROR: Failed to stop events on port\n");
         PrintMTError();
      }

      if(MTCloseChannel(curChannel) != MT_RET_OK)
      {
         ivLog->Print("ERROR: Failed to close port\n");
         PrintMTError();
      }
   }

   ivLog->Print("Shutting down\n");
}

// Returns the name of the event code number
const char* BoardMT::GetEventName(int event)
{
   MTGetEventCodeDescription(event, eventNameBuffer, 256);
   return(eventNameBuffer);
}

bool BoardMT::CallerIDEvent(MT_EVENT& event, char* callingDigits)
{
   char ANIString[MAX_DIGIT_BUFFER_SIZE];
   size_t pNumberStart;
   int iLenDigits;
   MT_RESULT result;
   string strData;

   if(event.EventCode != EVT_CALLID_STOP)
   {
      ivLog->Print("         ERROR: Invalid board event 0x%x\n", event.EventCode);
      return(false);
   }
   if(callingDigits == NULL)
   {
      ivLog->Print("         ERROR: pointer to retrieve callingDigits is NULL\n");
      return(false);
   }

   memset(ANIString, '\0', MAX_DIGIT_BUFFER_SIZE);

   ANILength = (unsigned short) (event.SubReason < MAX_DIGIT_BUFFER_SIZE - 1 ? event.SubReason : MAX_DIGIT_BUFFER_SIZE - 1);
   // ivLog->Print("         data length = %i\n", ANILength);

   // get Caller ID data from SmartWorks
   result = MTGetCallerID(curChannel, &ANILength, (unsigned char*) ANIString);
   if(result != MT_RET_OK)
   {
      ivLog->Print("         Failed to retrieve caller ID\n");
      PrintMTError();
      return(false);
   }
   // ivLog->Print("         data = '%s'\n", ANIString);

   // copy to STL string for easier manipulation
   strData.assign(ANIString, ANILength);

   switch (strData.at(0))
   {
      case 0x80:
      {
         // ivLog->Print("         format = MDMF\n");
         break;
      }
      case 0x04:
      {
         ivLog->Print("         format = SDMF (not supported)\n");
         break;
      }
      default:
      {
         ivLog->Print("         format unknown\n");
         break;
      }
   }

   // Note: The following works on MDMF (Multiple Data Message Format) only
   //       SDMF (Single Data Message Format) will return empty string

   pNumberStart = strData.rfind(0x02);    // data type 0x02 = phone number
   if(pNumberStart != string::npos)
   {
      pNumberStart = pNumberStart + 2; // skip separator characters
      // ivLog->Print("         number start = %i\n", pNumberStart);
      strcpy(callingDigits, strData.substr(pNumberStart).c_str());
      // Trim the checksum digit from the end
      iLenDigits = strlen(callingDigits)-1;
      // ivLog->Print("         digits length = %i\n", iLenDigits);
      callingDigits[iLenDigits] = '\0';
      // ivLog->Print ("         callingDigits: %s\n", callingDigits);
   }
   else
   {
      ivLog->Print("         Number not present\n");
      strcpy(callingDigits, "");
   }

   return(true);
}
//BoardMT::CallerIDEvent

IV_STATUS BoardMT::AudioStopEvent(MT_EVENT* event)
{
   switch(event->FuncCode)
   {
      // Playing from file
      case F_PLAYB:
      case F_PLAYF:
      {
         audioState->StopPlaying();
         break;
      }

      // Recording from stream
      case F_RECS:
      {
         if(audioState->IsRecording())
         {
            audioState->StopRecording();
         }

         if((event->EventCode == EVT_TERMSILENCE) || (event->EventCode == EVT_TERMDIGIT))
         {
            return(StopStreaming());
         }
         else if(event->EventCode == EVT_STREAMIN_STOP)
         {
            audioState->StopStreaming();

            if((audioState->Streamers() > 0) && audioState->RestartStreaming())
            {
               ivLog->Print("      Restart backgound recording due to event '%s', FuncCode F_RECS\n", GetEventName(event->EventCode));
               return(StartStreaming(audioState->StreamFormat()));
            }
            else
            {
               ivLog->Print("      Pause Streamers due to event '%s', FuncCode F_RECS\n", GetEventName(event->EventCode));
               audioState->PauseStreamers();
            }
         }
         break;
      }

      // Recording to buffer
      case F_RECB:
      {
         if(audioState->Streamers() > 0)
         {
            ivLog->Print("      Restart backgound recording due to event '%s', FuncCode F_RECB\n", GetEventName(event->EventCode));
            return(StartStreaming(audioState->StreamFormat()));
         }
         else
         {
            ivLog->Print("      Pause Streamers due to event '%s', FuncCode F_RECB\n", GetEventName(event->EventCode));
            audioState->PauseStreamers();
         }
         break;
      }
   }

   return(IVR_SUCCESS);
}

IV_STATUS BoardMT::AudioStopRecord(void)
{
   return(audioState->IsRecording() ? StopStreaming() : IVR_SUCCESS);
}

IV_STATUS BoardMT::StartStreaming(SOUND_FORMAT format)
{
 	MT_IO_CONTROL ioCtrl;
   MT_RESULT result;
   MTBoardEvent boardEvent;
   IV_STATUS status;

   if(audioState->IsStreaming())
   {
      IV_STATUS status = StopStreaming();
      if(status != IVR_SUCCESS)
      {
         return(status);
      }
   }

	// Clear parameter
   memset(&ioCtrl, 0, sizeof(MT_IO_CONTROL));

   // Set file/audio format
   ioCtrl.FileType = GetEncoding(audioState->StreamFormat());

   // Set streaming to not terminate on Call Progress Monitoring events
   ioCtrl.CpmMode = CPMMODE_MONI;

   // Set streaming watermark
   ioCtrl.Length = audioState->AudioBufferSize(audioState->StreamFormat());

   // Enable/disable automatic gain control
   if(MTAGCControl(curChannel, true) != MT_RET_OK)
   {
      ivLog->Print("ERROR: Failed to set AGC\n");
      PrintMTError();
   }

   // Start streaming
   ivLog->Print("      BoardMT::StartStreaming: MTStartStreaming, ioCtrl.MaxSilence = %i\n", ioCtrl.MaxSilence);   // @@@ debug
   result = MTStartStreaming(curChannel, MT_STREAM_INPUT, &ioCtrl);
   if((result != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to stream on port\n");
      PrintMTError();
    	return(IVR_FAIL);
   }

   ivLog->Print("      Wait for streaming start\n");

	// Wait for board to signal record completed
   while((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS)
   {
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_STREAMIN_START:
         {
            ivLog->Print("      Streaming started\n");   // @@@ debug
            // Update audioState
            audioState->StartStreaming();
            audioState->StartStreamers(format);
            return(IVR_SUCCESS);
         }

         default:
         {
            ivLog->Print("      WARNING: StartStreaming consumed event '%s'\n",
                   GetEventName(boardEvent.MTEvent.EventCode));
         }
      }
   }

   return(status);
}

IV_STATUS BoardMT::StopStreaming(void)
{
   MTBoardEvent boardEvent;
   MT_RESULT result;

   if(!audioState->IsStreaming())
   {
      return(IVR_SUCCESS);
   }

   ivLog->Print("      BoardMT::StopStreaming: MTStopStreaming called\n");   // @@@ debug:

   if((result = MTStopStreaming(curChannel, MT_STREAM_INPUT, MT_STREAM_STOP_NOW)) != MT_RET_OK)
   {
      if(result == MT_RET_SERVICE_NOT_STARTED)
      {
         ivLog->Print("      Streaming not running\n");
         return(IVR_SUCCESS);
      }
      else
      {
         ivLog->Print("      ERROR: Failed to stop stream on port\n");
         PrintMTError();
         return(IVR_FAIL);
      }
   }

   return(AudioWaitForRecordStop());
}

void BoardMT::Reset(void)
{
   short adapter;
   short channelOnAdapter;
   MTADAPTER_XINFO adapterInfo;
   unsigned long adapterInfoSize = sizeof(MTADAPTER_XINFO);
   unsigned short eventMask;
   unsigned long monitorEventMask;
   MT_PSTN pstnParams;
   unsigned long pstnParamsSize = sizeof(MT_PSTN);
   MT_VOICEDETECTPARAMS voiceDetectParams;
   unsigned long voiceDetectParamsSize = sizeof(MT_VOICEDETECTPARAMS);

   if(pInfo->iPort > 0)
   {
      // Close
      {
	      // Reset any previous switch states that may have been created by this thread
   	   UnSwitch();

	      // Stop all activity on this channel of the board
		   MTStopChannel(curChannel);

         MTCloseChannel(curChannel);
      }

      // Open
      {
         audioState->Reset();

         cs.Enter();

         // Get adapter information
	   	if(MTGetGCIMap(curChannel, &adapter, &adapterType, &channelOnAdapter)!= MT_RET_OK)
	      {
   	  	   ivLog->Print("ERROR: Failed to get GCI Map\n");
            PrintMTError();
          	bBrdInit = false;
	      }

   		if(MTOpenChannel(curChannel) != MT_RET_OK)
	      {
   	     	ivLog->Print("ERROR: Failed to open port\n");
            PrintMTError();
          	bBrdInit = false;
	      }

         // Set the board event queue mask
      	eventMask = SE_LCURRENT_CHANGE	// Enables EVT_LOOP_DROP, EVT_LOOP_ON
   				 | SE_RING	            // Enables EVT_RINGS
         	    | SE_ACT	            // Enables EVT_MON_ACTIVITY, EVT_MAX_ACTIVITY
         	    | SE_SIL	            // Enables EVT_MON_SILENCE, EVT_MAX_SILENCE
         	    | SE_OFFH	            // Enables EVT_OFFHOOK
            	 | SE_ONH 	            // Enables EVT_ONHOOK
   	          | SE_LREV	            // Enables EVT_REVERSAL
   	          | SE_LVOLTAGE_CHANGE;	// Enables EVT_LVOLTAGE_OFFHOOK, EVT_LVOLTAGE_NOTOFFHOOK

   		// Set board events
	      if(MTSetEventFilters(curChannel, eventMask, 1) != MT_RET_OK)
   	   {
     		   ivLog->Print("ERROR: Failed to set MT events mask\n");
            PrintMTError();
          	bBrdInit = false;
	      }

         if(EnableReorderToneDetection() != IVR_SUCCESS)
         {
            bBrdInit = false;
         }

	   	// Set the monitor event mask
   	   monitorEventMask =  MONI_BUSY1		// Enables EVT_MON_BUSY
                        | MONI_BUSY2		// Enables EVT_MON_BUSY_TRUNK
                        | MONI_DIAL		   // Enables EVT_MON_DIAL
                        | MONI_CALLWAITING
                     	| MONI_RINGBACK1	// Enables EVT_MON_RINGBACK
                     	| MONI_RINGBACK2	// Enables EVT_MON_DRINGBACK
//                        | MONI_SIGNAL_CYCLE
   	                  | MONI_REVERSAL	// Enables EVT_REVERSAL
                        | MONI_VOICE
                        | MONI_SIT1
                        | MONI_SIT2
                        | MONI_SIT3
                        | MONI_SIT4
                        | MONI_SIT5
                        | MONI_DIAL1
                        | MONI_DIAL2
                        | MONI_DIAL3
                        | MONI_FAX1
                        | MONI_FAX2
                        | MONI_UTONE;

   		// Set monitor events
	      if(MTSetMoni(curChannel, monitorEventMask) != MT_RET_OK)
   	   {
     		   ivLog->Print("ERROR: Failed to set MT monitor events mask\n");
            PrintMTError();
          	bBrdInit = false;
	      }

   	   if(MTChGetVoiceDetectParams(curChannel, &voiceDetectParams, &voiceDetectParamsSize) != MT_RET_OK)
      	{
     	   	ivLog->Print("ERROR: Failed to get MT voice parameters\n");
            PrintMTError();
          	bBrdInit = false;
	      }

         // The default value of 100 ms causes premature voice detection during PlaceCall
         voiceDetectParams.voice_min = 200;

	      if(MTChSetVoiceDetectParams(curChannel, &voiceDetectParams, &voiceDetectParamsSize) != MT_RET_OK)
      	{
        		ivLog->Print("ERROR: Failed to set MT voice parameters\n");
            PrintMTError();
       	   bBrdInit = false;
   	   }

         // Set ABCD bit signaling
         if(MTGetAdapterXInfo(adapter, &adapterInfo, &adapterInfoSize) != MT_RET_OK)
   	   {
     		   ivLog->Print("ERROR: Failed to get adapter extra info\n");
            PrintMTError();
          	bBrdInit = false;
	      }
         else
         {
         	MT_RESULT result = MT_RET_OK;

            for(unsigned short framerIndex = 0; (framerIndex < adapterInfo.NumFramers) && (result == MT_RET_OK); framerIndex++)
            {
               result = MTEnableFramerABCDSignalingEvent(adapter, framerIndex);
            }

            if((result != MT_RET_OK) && (result != MT_RET_NOT_AVAILABLE))
            {
  	      	   ivLog->Print("WARNING: Failed to enable ABCD signalling bits\n");
               PrintMTError();
               bBrdInit = false;
            }
         }

   	   if(MTGetPSTNParams(curChannel, &pstnParams, &pstnParamsSize) != MT_RET_OK)
      	{
     	   	ivLog->Print("ERROR: Failed to get PSTN defaults\n");
            PrintMTError();
          	bBrdInit = false;
	      }

         // Set default PSTN
         pstnParams.flash_char				= '&';	// default is '&'
         pstnParams.flash_time				= 500;	// default is 500 ms
         pstnParams.wink_delay				= 150;	// default is 150 ms
         pstnParams.wink_time					= 150;	// default is 150 ms
         pstnParams.pause_char				= ',';	// default is ','
         pstnParams.pause_time				= 2000;	// default is 2000 ms
         pstnParams.pulse_on					= 40;		// default is 40 ms
         pstnParams.pulse_off					= 60;		// default is 60 ms
         pstnParams.pulse_delay_time		= 1000;	// default is 1000 ms
         // Ai-Logix recommends 250 ms instead of the default
         // pstnParams.loop_deglitch      = 250;   // default is 50 ms
         pstnParams.loop_deglitch         = 750;   // attempting to solve flash-hook problems for CET - 2010/10/05
         pstnParams.loop_delay				= 10;		// default is 10 ms
         pstnParams.ring_on_time				= 300;	// default is 300 ms
         pstnParams.ring_off_time			= 500;	// default is 500 ms
         pstnParams.ring_reset				= 8000;	// default is 8000 ms
         pstnParams.ring_count				= 1;		// default is 1
         pstnParams.ring_trigger				= 0;		// default is 0
         pstnParams.ring_autoanswer			= 0;		// default is 0 (0 = disabled, 1 = enabled)
         pstnParams.cpm_start_delay			= 1000;	// default is 1000 ms
         pstnParams.offhook_delay			= 500;	// default is 500 ms
         pstnParams.min_rings_no_answer	= 4;		// default is 4
         pstnParams.max_sig_time				= 6500;	// default is 6500 ms
         pstnParams.max_sil_time				= 40000;	// default is 40000 ms

   		// Set PSTN defaults
	      if(MTSetPSTNParams(curChannel, &pstnParams, &pstnParamsSize) != MT_RET_OK)
   	   {
     		   ivLog->Print("ERROR: Failed to set PSTN defaults\n");
            PrintMTError();
          	bBrdInit = false;
	      }

   		// Set monitor mode default to monitor and terminate
         if(MTChCPMControl(curChannel, CPMMODE_MONI) != MT_RET_OK)
   	   {
     		   ivLog->Print("ERROR: Failed to set CPM mode\n");
            PrintMTError();
          	bBrdInit = false;
	      }

     		// Set streaming buffer size to thrice the audio buffer size
         // If the streaming buffer is too large than audio gets missed at the end of recordings
	      if(MTSetStreamingConfig(curChannel, 3*audioState->AudioBufferSize()) != MT_RET_OK)
      	{
        		ivLog->Print("ERROR: Failed to set streaming buffer size\n");
            PrintMTError();
       	   bBrdInit = false;
   	   }

         switch(adapterType)
         {
            case VOICE_RESOURCE:             // For channels on VR6400, VR3200, VR6409, VR3209
            case ANALOG_VDS:                 // For channels on PT1600, PT800, PT1609, PT809
            case DIGITAL_PASSIVE_T1_NONE:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_PASSIVE_T1_ISDN:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_PASSIVE_T1_RBS:     // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_PASSIVE_T1_NFAS:    // For channels on DP6400-T1, DP3200-T1, DP6409-T1, DP3209-T1
            case DIGITAL_PASSIVE_E1_NONE:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_ISDN:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_CAS:     // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_NFAS:    // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_E1_DASS2:   // For channels on DP6400-E1, DP3200-E1, DP6409-E1, DP3209-E1
            case DIGITAL_PASSIVE_NGX:        // For channels on Passive NGX
            {
               // Nothing to do for passive boards
               break;
            }

            case ANALOG_FXO_LS:              // For channels on AT1600, AT800, AT1609, AT809
            case ANALOG_FXS_LS_E:            // For station channels on AT4X12
            case ANALOG_VFXO_LS_E:           // For channels on LD809, LD409, LD2409, LD1609
            case ANALOG_VFXS_LS_E:           // For station channels on LD101, LD201
            case DIGITAL_TERMINATE_T1_NONE:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_T1_ISDN:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_T1_RBS:   // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_T1_NFAS:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_E1_NONE:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_E1_ISDN:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_E1_CAS:   // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            case DIGITAL_TERMINATE_E1_NFAS:  // For channels on DT6400-T1, DT3200-T1, DT6409-T1, DT3209-T1
            {
   		      // Put this channel is on hook
   	         if(MTOnHook(curChannel) != MT_RET_OK)
         	   {
              		ivLog->Print("ERROR: Failed to place MT channel on hook\n");
                  PrintMTError();
             	   bBrdInit = false;
   	         }
            }
         }

         ttQueue.Clear();

         // Clear the board digit queue for this channel
         if(MTClearDTMFDigits(curChannel) != MT_RET_OK)
         {
            ivLog->Print("ERROR: Failed to clear MT digit queue\n");
            PrintMTError();
            bBrdInit = false;
         }

	      eventQueue.Clear();
	      callTriggerQueue.Clear();

      	// Clear the board event queue for this channel
	      if(MTFlushEvents(curChannel) != MT_RET_OK)
   	   {
     		   ivLog->Print("ERROR: Failed to flush MT channel events\n");
            PrintMTError();
          	bBrdInit = false;
	      }

   	   cs.Leave();
      }

      ANILength = 0;
   }
}

IV_STATUS BoardMT::StartReorderToneDetection(void)
{
   MT_CPMTONE reorderCPM;
   unsigned long cpmtoneParamsSize = sizeof(MT_CPMTONE);

   // Clear structure
   memset(&reorderCPM, 0, sizeof(MT_CPMTONE));

   // First user tone index available
   reorderCPM.index = 15;

   // First user tone ID available
   reorderCPM.signalname = 0;

   // Set as user tone
   reorderCPM.signaltype = USR_TONE;

   // Enable tone detection
   reorderCPM.signal.fEnable = 1;

   // Disable identification as SIT
   reorderCPM.signal.fDetectSIT = 0;

   // ???
   reorderCPM.signal.fDetectLate = 0;

   // Disable EVT_MON every time the cycle is detected
   reorderCPM.signal.fDetectCycle = 0;

   // ???
   reorderCPM.signal.fDetectLost = 1;

   // Only require 2 cycles before detection is signalled
   reorderCPM.signal.nCycleDetect = 2;

   // Filters for reorder tone (0 for unused frequencies)
   reorderCPM.signal.Freq[0] = (unsigned short) reorderTone.frequency1;
   reorderCPM.signal.Freq[1] = (unsigned short) reorderTone.frequency2;

   // Minimum/maximum pulse widths for reorder tone (0 for unused pulses)
   reorderCPM.signal.Pulse[0].onmin = (unsigned short) reorderTone.minimumOnTime;
   reorderCPM.signal.Pulse[0].onmax = (unsigned short) reorderTone.maximumOnTime;
   reorderCPM.signal.Pulse[0].offmin = (unsigned short) reorderTone.minimumOffTime;
   reorderCPM.signal.Pulse[0].offmax = (unsigned short) reorderTone.maximumOffTime;

   if(MTSetCpmToneParams(curChannel, &reorderCPM, &cpmtoneParamsSize) != MT_RET_OK)
   {
      ivLog->Print("      ERROR: StartReorderToneDetection::MTSetCpmToneParams failed\n");
      PrintMTError();
      return(IVR_FAIL);
   }

   return(IVR_SUCCESS);
}

IV_STATUS BoardMT::StopReorderToneDetection(void)
{
   MT_CPMTONE reorderCPM;
   unsigned long cpmtoneParamsSize = sizeof(MT_CPMTONE);

   // Clear structure
   memset(&reorderCPM, 0, sizeof(MT_CPMTONE));

   // First user tone index available
   reorderCPM.index = 15;

   // First user tone ID available
   reorderCPM.signalname = 0;

   // Set as user tone
   reorderCPM.signaltype = USR_TONE;

   // Disable tone detection
   reorderCPM.signal.fEnable = 0;

   // Disable identification as SIT
   reorderCPM.signal.fDetectSIT = 0;

   // ???
   reorderCPM.signal.fDetectLate = 0;

   // Disable EVT_MON every time the cycle is detected
   reorderCPM.signal.fDetectCycle = 0;

   // ???
   reorderCPM.signal.fDetectLost = 1;

   // Only require 2 cycles before detection is signalled
   reorderCPM.signal.nCycleDetect = 2;

   if(MTSetCpmToneParams(curChannel, &reorderCPM, &cpmtoneParamsSize) != MT_RET_OK)
   {
      ivLog->Print("      ERROR: StopReorderToneDetection::MTSetCpmToneParams failed\n");
      PrintMTError();
      return(IVR_FAIL);
   }

   return(IVR_SUCCESS);
}

IV_STATUS BoardMT::Switch(int newPort, bool bListen)
{
	MT_RESULT result;

	if(bListen)
   {
		ivLog->Print("      Switch current port (%d) to port %d (listen only)\n", pInfo->iPort, newPort);
		ivLog->Print("      ERROR: Listen-only switch connections not supported on Ai-Logix hardware\n");
      return(IVR_NOTSUPPORTED);
   }
   else
   {
		ivLog->Print("      Switch current port (%d) to port %d\n", pInfo->iPort, newPort);
   }

   // Reset any previous switch state(s) created by this thread
   UnSwitch();

	// Switch FIRST channel onto H.100 bus
	// ===================================

	// forward connection
	// ------------------
   // note: NMS uses the terms "local" and "mvip"
	ivLog->Print("         board %d   channel %d  TALK  -> H.100 %d:%d\n",
					 lines[curLine].board,
                curChannel,
                1, curChannel
               );
   result = MTSetCTRoute(curChannel,					// const CHANNEL nCannel,
                         DIR_TALK,						// const UCHAR Direction,
                         1,								// const int Stream,
                         (int)curChannel				// const int Slot
                        );
	if((result != MT_RET_OK))
	{
		ivLog->Print("ERROR: Failed to make forward connection on current port\n");
		PrintMTError();
      return(IVR_FAIL);
	}

	// reverse connection
	// ------------------
	ivLog->Print("         board %d   channel %d LISTEN <- H.100 %d:%d\n",
					 lines[curLine].board,
                curChannel,
                0, curChannel
               );
   result = MTSetCTRoute(curChannel,					// const CHANNEL nCannel,
                         DIR_LISTEN,					// const UCHAR Direction,
                         0,								// const int Stream,
                         (int)curChannel				// const int Slot
                        );
	if((result != MT_RET_OK))
	{
		ivLog->Print("ERROR: Failed to make reverse connection on current port\n");
		PrintMTError();
      return(IVR_FAIL);
	}

	// Switch OTHER channel onto H.100 bus
	// ===================================

	// forward connection
	// ------------------
	ivLog->Print("         board %d   channel %d  TALK  -> H.100 %d:%d\n",
					 lines[curLine].board,
                (newPort-1),
                0, curChannel
               );
   result = MTSetCTRoute((CHANNEL)(newPort-1),		// const CHANNEL nCannel,
                         DIR_TALK,						// const UCHAR Direction,
                         0,								// const int Stream,
                         (int)curChannel				// const int Slot
                        );
	if((result != MT_RET_OK))
	{
		ivLog->Print("ERROR: Failed to make forward connection on port %d\n", newPort);
		PrintMTError();
      return(IVR_FAIL);
	}

	// reverse connection
	// ------------------
	ivLog->Print("         board %d   channel %d LISTEN <- H.100 %d:%d\n",
					 lines[curLine].board,
                (newPort-1),
                1, curChannel
               );
   result = MTSetCTRoute((CHANNEL)(newPort-1),		// const CHANNEL nCannel,
                         DIR_LISTEN,					// const UCHAR Direction,
                         1,								// const int Stream,
                         (int)curChannel				// const int Slot
                        );
	if((result != MT_RET_OK))
	{
		ivLog->Print("ERROR: Failed to make reverse connection on port %d\n", newPort);
		PrintMTError();
      return(IVR_FAIL);
	}

   otherChannel = (CHANNEL)(newPort-1);

   // Done.
   // -----
	ivLog->Print("      Connected.\n");
   return(IVR_SUCCESS);
}
//BoardMT::Switch

IV_STATUS BoardMT::UnSwitch(void)
{
	MT_RESULT result;
   int stream;
   int slot;

   //  QuerySwitch();	   // DEBUG only: displays current switch setting(s)

	// Check that we have a port, and that it has an open Switch connection that it initiated
   if(((pInfo->iPort) > 0) && (otherChannel >= 0))
   {
		ivLog->Print("      Disconnecting switch state(s)\n");

		// Query CURRENT channel connection
		// ================================
      // Note: current port should always be open.
      // Don't need to check for that result when we make MT API calls

      // query forward connection
      // ------------------------
      result = MTGetCTRoute(curChannel,		// const CHANNEL nChannel,
                            DIR_TALK,			// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK))
		{
			ivLog->Print("ERROR: Failed to get forward connection on current port\n");
			PrintMTError();
			return(IVR_FAIL);
		}
      if (stream == -1)
      {
      	// not connected
			ivLog->Print("         board %d   channel %d  TALK  not connected \n",
			             lines[curLine].board,
			             curChannel
			            );
      }
      else
      {
	      // reset forward connection
	      // ------------------------
			result = MTResetCTRoute(curChannel,		// const CHANNEL nChannel,
			                        DIR_TALK,		// const UCHAR Direction,
			                        stream,			// const int Stream
			                        slot				// const int Slot
			                       );
			if((result != MT_RET_OK))
			{
				ivLog->Print("ERROR: Failed to reset forward connection on current port\n");
				PrintMTError();
				return(IVR_FAIL);
			}

			ivLog->Print("         board %d   channel %d  TALK  reset \n",
			             lines[curLine].board,
			             curChannel
			            );
		}

      // query reverse connection
      // ------------------------
      result = MTGetCTRoute(curChannel,		// const CHANNEL nChannel,
                            DIR_LISTEN,		// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK))
		{
			ivLog->Print("ERROR: Failed to get reverse connection on current port\n");
			PrintMTError();
			return(IVR_FAIL);
		}
      if (stream == -1)
      {
			ivLog->Print("         board %d   channel %d LISTEN not connected \n",
			             lines[curLine].board,
			             curChannel
			            );
      }
      else
      {
			// reset reverse connection
			// ------------------------
			result = MTResetCTRoute(curChannel,		// const CHANNEL nChannel,
			                        DIR_LISTEN,		// const UCHAR Direction,
			                        stream,			// const int Stream
			                        slot				// const int Slot
			                       );
			if((result != MT_RET_OK))
			{
				ivLog->Print("ERROR: Failed to reset forward connection on current port\n");
				PrintMTError();
				return(IVR_FAIL);
			}

			ivLog->Print("         board %d   channel %d LISTEN reset\n",
			             lines[curLine].board,
			             curChannel
			            );
		}

		// Query OTHER channel connection
		// ==============================
      // Note: other port could be closed.
      // Need to check for that result when we make MT API calls

      // query forward connection
      // ------------------------
      result = MTGetCTRoute(otherChannel,		// const CHANNEL nChannel,
                            DIR_TALK,			// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK) && (result != MT_RET_CHANNEL_NOT_OPENED))
		{
			ivLog->Print("ERROR: Failed to get forward connection on port %d\n", (otherChannel+1));
			PrintMTError();
			return(IVR_FAIL);
		}

      if ((result == MT_RET_CHANNEL_NOT_OPENED) || (stream == -1))
      {
			ivLog->Print("         board %d   channel %d  TALK  not connected \n",
			             lines[curLine].board,
			             otherChannel
			            );
      }
      else
      {
			// reset forward connection
			// ------------------------
			result = MTResetCTRoute(otherChannel,	// const CHANNEL nChannel,
			                        DIR_TALK,		// const UCHAR Direction,
			                        stream,			// const int Stream
			                        slot				// const int Slot
			                       );
			if((result != MT_RET_OK))
			{
				ivLog->Print("ERROR: Failed to reset forward connection on current port\n");
				PrintMTError();
				return(IVR_FAIL);
			}

			ivLog->Print("         board %d   channel %d  TALK  reset\n",
			             lines[curLine].board,
			             otherChannel
			            );
      }

      // query reverse connection
      // ------------------------
      result = MTGetCTRoute(otherChannel,		// const CHANNEL nChannel,
                            DIR_LISTEN,		// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK) && (result != MT_RET_CHANNEL_NOT_OPENED))
		{
			ivLog->Print("ERROR: Failed to get forward connection on current port\n");
			PrintMTError();
			return(IVR_FAIL);
		}

      if ((result == MT_RET_CHANNEL_NOT_OPENED) || (stream == -1))
      {
			ivLog->Print("         board %d   channel %d LISTEN not connected \n",
			             lines[curLine].board,
			             otherChannel
			            );
      }
      else
      {
			// reset reverse connection
			// ------------------------
			result = MTResetCTRoute(otherChannel,	// const CHANNEL nChannel,
			                        DIR_LISTEN,		// const UCHAR Direction,
			                        stream,				// const int Stream
			                        slot					// const int Slot
			                       );
			if((result != MT_RET_OK))
			{
				ivLog->Print("ERROR: Failed to reset forward connection on current port\n");
				PrintMTError();
				return(IVR_FAIL);
			}

			ivLog->Print("         board %d   channel %d LISTEN reset\n",
			             lines[curLine].board,
			             otherChannel
			            );
      }

      // Set other port at BoardMT level to indicate we are no longer connected
      otherChannel = -1;
   }
	return(IVR_SUCCESS);
}
//BoardMT::UnSwitch


IV_STATUS BoardMT::QuerySwitch(void)
{
	MT_RESULT result;
   int stream;
   int slot;

	ivLog->Print ("      ------- \n");
	ivLog->Print ("      Query Switch\n");
	ivLog->Print ("      ------- \n");
	ivLog->Print ("         curLine      = %d\n", curLine);
	ivLog->Print ("         pInfo->iPort = %d\n", pInfo->iPort);
	ivLog->Print ("         curChannel   = %d\n", curChannel);
	ivLog->Print ("         otherChannel = %d\n", otherChannel);
	ivLog->Print ("      ------- \n");

	ivLog->Print ("         Current connections: \n");
   if((pInfo->iPort) > 0)
   {
		// Query CURRENT channel connection
		// ================================

      // forward connection
      // ------------------
      result = MTGetCTRoute(curChannel,		// const CHANNEL nChannel,
                            DIR_TALK,			// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK))
		{
			ivLog->Print("ERROR: Failed to get forward connection on current port\n");
			PrintMTError();
			return(IVR_FAIL);
		}
		ivLog->Print("         board %d   channel %d  TALK  -> H.100 %d:%d\n",
						 lines[curLine].board,
	                curChannel,
	                stream, slot
	               );

      // reverse connection
      // ------------------
      result = MTGetCTRoute(curChannel,		// const CHANNEL nChannel,
                            DIR_LISTEN,		// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK))
		{
			ivLog->Print("ERROR: Failed to get reverse connection on current port\n");
			PrintMTError();
			return(IVR_FAIL);
		}
		ivLog->Print("         board %d   channel %d LISTEN <- H.100 %d:%d\n",
						 lines[curLine].board,
	                curChannel,
	                stream, slot
	               );
   }
   if(otherChannel >= 0)
   {
		// Query OTHER channel connection
		// ==============================

      // forward connection
      // ------------------
      result = MTGetCTRoute(otherChannel,		// const CHANNEL nChannel,
                            DIR_TALK,			// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK))
		{
			ivLog->Print("ERROR: Failed to get forward connection on port %d\n", otherChannel+1);
			PrintMTError();
			return(IVR_FAIL);
		}
		ivLog->Print("         board %d   channel %d  TALK  -> H.100 %d:%d\n",
						 lines[curLine].board,
	                otherChannel,
	                stream, slot
	               );

		// reverse connection
		// ------------------
      result = MTGetCTRoute(otherChannel,		// const CHANNEL nChannel,
                            DIR_LISTEN,		// const UCHAR Direction,
                            &stream,			// int * const pStream
                            &slot				// int * const pSlot
                           );
		if((result != MT_RET_OK))
		{
			ivLog->Print("ERROR: Failed to get reverse connection on port %d\n", otherChannel+1);
			PrintMTError();
			return(IVR_FAIL);
		}
		ivLog->Print("         board %d   channel %d LISTEN <- H.100 %d:%d\n",
						 lines[curLine].board,
	                otherChannel,
	                stream, slot
	               );
	}
	ivLog->Print ("      ------- \n");
	return(IVR_SUCCESS);
}
//BoardMT::QuerySwitch

IV_STATUS BoardMT::PlaceCall(const char* digits, const char*, int numrings, callMode mode,
   int flashdur, bool connectOnCED, bool connectOnSIT, int*, int silenceTimeout,
   int, ISDNNumberPlan, ISDNNumberType, ISDNNumberPlan, ISDNNumberType)

{
	MTBoardEvent boardEvent;
   IV_STATUS status;
   MT_IO_CONTROL ioParameters;
	MT_PSTN originalPSTNParameters;
	MT_PSTN pstnParameters;
	unsigned long pstnParametersSize = sizeof(MT_PSTN);
   LARGE_INTEGER waitTimeout;
   MT_RESULT result;
   char* trimmedDigits;
   char* trimmedIndex;
   char validDigits[] = "1234567890ABCDabcd*#";
   bool bContinue = true;
   bool dialToneDetected = false;
   bool voiceDetected = false;

   if(!bBrdInit)
   {
      ivLog->Print("      ERROR: Port is not initialised\n");
      return(IVR_FATAL);
   }

   if(silenceTimeout > 0)
   {
		ivLog->Print("      Initial voice length detection is not supported on this hardware\n");
      return(IVR_NOTSUPPORTED);
   }

   if((mode == newLine) && lineOffHook)
   {
      ivLog->Print("      ERROR: Cannot use a new line if port is already in use\n");
    	return(IVR_FATAL);
   }
   else if((mode == existingLine) && !lineOffHook)
   {
      ivLog->Print("      ERROR: Cannot use existing line if port is not off hook\n");
    	return(IVR_FATAL);
   }

	// Get current PSTN parameters
   if(MTGetPSTNParams(curChannel, &originalPSTNParameters, &pstnParametersSize) != MT_RET_OK)
   {
      PrintMTError();
   	return(IVR_FAIL);
   }

   // Set PSTN parameters for this function
   pstnParameters = originalPSTNParameters;
   pstnParameters.min_rings_no_answer = (unsigned short) numrings;

	// Set PSTN parameters with new minimum # of rings for timeout
   if(MTSetPSTNParams(curChannel, &pstnParameters, &pstnParametersSize) != MT_RET_OK)
   {
      PrintMTError();
   	return(IVR_FAIL);
   }

	// Assure blank I/O parameters
	memset(&ioParameters, 0, sizeof(MT_IO_CONTROL));

   ioParameters.CpmMode = MONI_USE_CHANNEL_AND_TASK | TERM_USE_TASK_ONLY;
   ioParameters.TermEnable = TERM_LOOP_DROP
                           | TERM_CONNECT
                           | TERM_NO_ANSWER
                           | TERM_INTERCEPT
                           | TERM_MON_UTONE
                           | TERM_MON_BUSY1
                           | TERM_MON_BUSY2
                           | TERM_MON_FAX1
                           | TERM_MON_FAX2
                           | TERM_MON_SIT1
                           | TERM_MON_SIT2
                           | TERM_MON_SIT3
                           | TERM_MON_SIT4
                           | TERM_MON_SIT5;

//   lineOffHook = true;

   trimmedIndex = trimmedDigits = new char[strlen(digits) + 1];

   // Strip away unusable characters
   for(unsigned int index = 0; index < strlen(digits); index++)
   {
      if((strchr(validDigits, digits[index]) != NULL)
         || (digits[index] == pstnParameters.flash_char) || (digits[index] == pstnParameters.pause_char))
      {
         trimmedIndex[0] = digits[index];
         trimmedIndex++;
      }
   }

   trimmedIndex[0] = '\0';

   audioState->StartDialing();

   if(mode == newLine)
   {
      // Dial digits
      if(((result = MTCallString(curChannel, trimmedDigits, &ioParameters)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
      {
         lineOffHook = false;
      	ivLog->Print("      ERROR: Failed to place call\n");
         PrintMTError();
         MTSetPSTNParams(curChannel, &originalPSTNParameters, &pstnParametersSize);
         delete[] trimmedDigits;
         Sleep(200);
         return(IVR_FAIL);
      }
   }
   else
   {
      if(flashdur > 0)
      {
   		if((status = FlashHook(flashdur)) != IVR_SUCCESS)
         {
            MTSetPSTNParams(curChannel, &originalPSTNParameters, &pstnParametersSize);
            ivLog->Print("      ERROR: Failed to Flash Hook port\n");
            delete[] trimmedDigits;
            Sleep(200);
         	return(status);
         }

       	Sleep(500);
      }

      if(GenerateDigits(trimmedDigits) != IVR_SUCCESS)
      {
         MTSetPSTNParams(curChannel, &originalPSTNParameters, &pstnParametersSize);
         delete[] trimmedDigits;
         Sleep(200);
      	return(status);
      }
   }

   delete[] trimmedDigits;
   MTSetPSTNParams(curChannel, &originalPSTNParameters, &pstnParametersSize);

   // Wait 5 s for connection, 6 s per ring, and 10 s of buffer before timing out
   waitTimeout.QuadPart = -(50000000 + 50000000*numrings + 100000000);
   SetWaitableTimer(hGenericTimer, &waitTimeout, 0, 0, 0, false);

   // Wait for board to signal dialing complete
   while(bContinue && ((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS))
   {
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_MON_VOICE:
         case EVT_MON_MACHINE:
         case EVT_MON_HUMAN:
         {
            voiceDetected = true;
            break;
         }

      	case EVT_CPM_STOP:
         {
         	CancelWaitableTimer(hGenericTimer);

				// Determine what happened when the board dialed
			   switch(boardEvent.MTEvent.SubReason)
			   {
			   	case CPM_CONNECT:
               case CPM_MON_CONNECT:
			      {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", "CONNECT");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", voiceDetected ? "VOICE" : "CONNECT");
      			   status = IVR_SUCCESS;
                  break;
			      }

               case CPM_MON_VOICE:
               case CPM_MON_MACHINE:
               case CPM_MON_HUMAN:
			      {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", "CONNECT");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", "VOICE");
      			   status = IVR_SUCCESS;
                  break;
			      }

			      case CPM_MON_BUSY:
               case CPM_MON_BUSY1:
               case CPM_MON_BUSY2:
			      {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", "BUSY");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", "BUSY");
      			   status = IVR_BUSY;
                  break;
			      }

			      case CPM_NO_ANSWER:
			      {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", "NOANSWER");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", "MAX_RINGS");
      			   status = IVR_TIMEOUT;
                  break;
			      }

               case CPM_NO_DIAL:
               {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", "FAILURE");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", "NO_DIALTONE");
      			   status = IVR_FAIL;
                  break;
               }

			      case CPM_INTERCEPT:
               case CPM_MON_SIT1:
               case CPM_MON_SIT2:
               case CPM_MON_SIT3:
               case CPM_MON_SIT4:
               case CPM_MON_SIT5:
               {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", connectOnSIT ? "CONNECT" : "FAILURE");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", "SIT");
                  status = connectOnSIT ? IVR_SUCCESS : IVR_FAIL;
                  break;
               }

               case CPM_MON_FAX1:
               case CPM_MON_FAX2:
               {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", connectOnCED ? "CONNECT" : "FAILURE");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", "CED");
                  status = connectOnCED ? IVR_SUCCESS : IVR_FAIL;
                  break;
               }

			      default:
			      {
                  pInfo->variables->SetSystem("@CALLOUT_RESULT", "FAILURE");
                  pInfo->variables->SetSystem("@CALLOUT_REASON", "UNKNOWN");

           	   	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
                  	"         User Status: %d\n         SubReason: %d\n"
                     "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
                     boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
                  status = IVR_FAIL;
			      }
				}

            bContinue = false;
            break;
			}

         case EVT_MON_RINGBACK:
         {
            break;
         }

         case EVT_MON_DIAL1:
         case EVT_MON_DIAL2:
         case EVT_MON_DIAL3:
         {
            dialToneDetected = true;
            break;
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

   audioState->StopDialing();
  	CancelWaitableTimer(hGenericTimer);

  	if(status == IVR_GENERIC_TIMEOUT)
   {
      if(dialToneDetected)
      {
         pInfo->variables->SetSystem("@CALLOUT_RESULT", "NOANSWER");
         pInfo->variables->SetSystem("@CALLOUT_REASON", "TIMEOUT");
         status = IVR_TIMEOUT;
      }
      else
      {
         pInfo->variables->SetSystem("@CALLOUT_RESULT", "FAILURE");
         pInfo->variables->SetSystem("@CALLOUT_REASON", "NO_DIALTONE");
         status = IVR_FAIL;
      }
   }

   if(mode == existingLine)
   {
      IV_STATUS status2;

      if((status2 = FlashHook(flashdur)) != IVR_SUCCESS)
      {
         return(status2);
      }
   }

   if((mode == newLine) && (status == IVR_SUCCESS))
   {
      lineOffHook = true;
      callStart.Update();
      callEnd.Invalidate();

      if(disableReorderToneDetection)
      {
         ivLog->Print("      Reorder tone detection disabled!\n");
      }
   }

  	return(status);
}

void BoardMT::PrintMTError(void)
{
   MT_RESULT error;
   unsigned short errorMessageLength = 127;
   char errorMessage[128];

   memset(errorMessage, 0, 128);

   MTGetLastError(&error, &errorMessageLength, errorMessage);

   ivLog->Print("      %X: %s\n", error, errorMessage);
}

IV_STATUS BoardMT::ProcessBoardEvent(BoardEvent* boardEvent, bool& bFound)
{
	MT_EVENT* event = &(((MTBoardEvent*) boardEvent)->MTEvent);
   IV_STATUS status = IVR_SUCCESS;
   char evt_tt;
   char brd_tt = 0;
   MT_RESULT result;

	// MT events are always valid (some NMS events aren't valid events)
	bFound = true;

   // Get event from our event queue
	if(eventQueue.Get(*event))
   {
      SYSTEMTIME eventTime;      // to output timestamp on board events
      FileTimeToSystemTime((FILETIME*)&(event->TimeStamp), &eventTime);

      // For debugging purposes
      if(event->EventCode == EVT_DIGIT)
      {
         ivLog->Print("Board event: %s: '%c' [qpos %i] %.2d:%.2d:%.2d.%.3d\n", GetEventName(event->EventCode), event->SubReason, event->UserStatus,
                      eventTime.wHour,
                      eventTime.wMinute,
                      eventTime.wSecond,
                      eventTime.wMilliseconds);
      }
      else
      {
         ivLog->Print("Board event: %s [qpos %i] %.2d:%.2d:%.2d.%.3d\n", GetEventName(event->EventCode), event->UserStatus,
                      eventTime.wHour,
                      eventTime.wMinute,
                      eventTime.wSecond,
                      eventTime.wMilliseconds);
      }
      if(strcmp(GetEventName(event->EventCode), "Unknown") == 0)
      {
         ivLog->Print("Event: %d\n", event->EventCode);
      }

      // Special processing for event
   	switch(event->EventCode)
	   {
         case EVT_DIGIT:
         {
            // Capture event digit from SubReason field
            evt_tt = (unsigned char) event->SubReason;

            // Remove digit from board queue
            result = MTReadDigit(curChannel, &((unsigned char) brd_tt));

            if(result != MT_RET_OK)
            {
               ivLog->Print("ERROR: MTReadDigit() failed with result of %i\n", result);
               ivLog->Print("   eventChannel = %i, curChannel = %i\n", event->Channel, curChannel);
               return(IVR_FATAL);
            }
            else if(brd_tt != NULL)
            {
               ivLog->Print("      digit '%c'\n", brd_tt);
            }

            // Put digit in our touch tone queue
            ttQueue.Add(evt_tt);

            // Delete contents of Ai-Logix digit queue, we already have the digit from the event
            while(brd_tt != 0)
            {
               MTReadDigit(curChannel, &((unsigned char) brd_tt));
               if(brd_tt != NULL)
               {
                  ivLog->Print("      digit '%c'\n", brd_tt);
               }
            }

            // Limit queue size to 100
            if(ttQueue.Size() > 100)
            {
               ttQueue.Get(brd_tt);
            }

            return(IVR_SUCCESS);
         }

			// These events indicate that the caller or board hung up
   		case EVT_ONHOOK:
	      case EVT_LOOP_DROP:
   	   case EVT_REVERSAL:
         case EVT_UTONE_ON:
         case EVT_MON_DIAL1:
         case EVT_MON_DIAL2:
         case EVT_MON_DIAL3:
	      {
   	      // Don't signal hangup if we are already hung up or dialing out
     			if(lineOffHook && !audioState->IsDialing())
	         {
	            audioState->RestartStreaming(false);
	            if(audioState->IsRecording())
	            {
	               status = AudioStopRecord();
	            }
	            else if(audioState->IsStreaming())
	            {
	               status = PauseStreamers();
	            }

	            if((status == IVR_SUCCESS) && audioState->IsPlaying())
	            {
	               status = AudioStopPlay();
	            }

	            if(status != IVR_SUCCESS)
	            {
	               return(status);
	            }

	         	return(Hangup(true));
            }
            // end if(lineOffHook)
            return(IVR_SUCCESS);
   		}

      	case EVT_LOOP_ON:
      	case EVT_OFFHOOK:
         {
            return(IVR_SUCCESS);
         }

         case EVT_EOF:
         case EVT_INTERDIGIT:
         case EVT_IO_COMPLETED:
         case EVT_MAXBYTES:
         case EVT_MAXTIME:
         case EVT_STOP:
         case EVT_TERMSILENCE:
         case EVT_TERMDIGIT:
         case EVT_STREAMIN_STOP:
         {
            return(AudioStopEvent(event));
         }

         // A board level failure has occurred
         case EVT_ERROR:
         {
         	ivLog->Print("      ERROR: A board level failure has occurred\n"
            	"         Time: %d\n         Status: %d\n         SubReason: %d\n"
               "         Termination: %d\n", event->UserStatus, event->SubReason, event->XtraInfo);
            return(IVR_FAIL);
         }

			// A system error has occurred
         case EVT_SYS_ERROR:
         {
         	ivLog->Print("      ERROR: A system error has occurred\n         Time: %d\n"
            "         Status: %d\n         SubReason: %d\n         Termination: %d\n",
            	event->UserStatus, event->SubReason, event->XtraInfo);
            return(IVR_FAIL);
         }

         case EVT_DISK_FULL:
         {
         	ivLog->Print("      ERROR: Disk is too full during record\n");
            return(IVR_FAIL);
			}

	    	default:
   	   {
  	   		return(IVR_SUCCESS);
	     	}
      }
   }
   // We could not find the signalled event
	else
	{
		ivLog->Print("      ERROR: Signalled MT board event on channel %d was not in queue!\n", curChannel);
  		return(IVR_FATAL);
   }
}
//BoardMT::ProcessBoardEvent

IV_STATUS BoardMT::Wait(int timeout, bool interruptable)
{
   LARGE_INTEGER waitTimeout;
   IV_STATUS status;
	MTBoardEvent boardEvent;

   if(lineOffHook && interruptable)
   {
      char ttDig;

   	if((status = QueryDigit(ttDig)) != IVR_SUCCESS)
      {
         return(status);
      }

      if(ttDig != '\0')
      {
         ivLog->Print("      Interrupted!\n");
         return(IVR_SUCCESS);
      }
   }

   if(timeout == 0)
   {
      return(IVR_SUCCESS);
   }

   waitTimeout.QuadPart = -(LONGLONG) timeout * 10000;
   SetWaitableTimer(hGenericTimer, &waitTimeout, 0, 0, 0, false);

   while((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS)
   {
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_DIGIT:
         {
            if(interruptable)
            {
               ivLog->Print("      Interrupted!\n");
              	CancelWaitableTimer(hGenericTimer);
               return(IVR_SUCCESS);
            }

            break;
         }

         default:
         {
            // Do nothing
         }
      }
   }

  	CancelWaitableTimer(hGenericTimer);

   return(status == IVR_GENERIC_TIMEOUT ? IVR_SUCCESS : status);
}

IV_STATUS BoardMT::SetParameter(const char* pszParm, DWORD dwParmValue)
{
   MT_RESULT result;

   if(!bBrdInit)
   {
      ivLog->Print("      ERROR: Board in an invalid state\n");
      return(IVR_FAIL);
   }

   if(stricmp(pszParm, "MTEnableCallerID()") == 0)
   {
      // Call SmartWorks function to set caller ID sensitivity
      USHORT nLevel = (USHORT)dwParmValue;
      if((nLevel < 0x02) || (nLevel > 0x80))
      {
         ivLog->Print("      ERROR: Invalid Caller ID sensitivity level %i (0x%x)\n", nLevel, nLevel);
         return(IVR_FATAL);
      }
      ivLog->Print("         MTEnableCallerID( (channel) %i, (sensitivity level) %i )\n", curChannel, nLevel);
      result = MTEnableCallerID(curChannel, nLevel);
      if(result != MT_RET_OK)
      {
         PrintMTError();
         return(IVR_FATAL);
      }
      return(IVR_SUCCESS);
   }

   // All other cases are not supported
   return(IVR_NOTSUPPORTED);
}

IV_STATUS BoardMT::CallTrigger(const char* triggerList, const char*& trigger, char*& triggerInfo, long timeout)
{
	MTBoardEvent 		boardEvent;
   IV_STATUS 			status;
   MT_CC_CALL_INFO  *pCallInfo;
   string            InfoString;
   bool              bFound;            // Used to exit the loop when one of the desired events is found

	triggerInfo = NULL;

   if(!bBrdInit)
   {
      ivLog->Print("      ERROR: Port is not initialised\n");
      return(IVR_FATAL);
   }

   if(CurrentActivityMode() == ACTIVE_ONLY)
   {
      ivLog->Print("      ERROR: Call Trigger not permitted in active configuration (protocol '%s')\n",
                   (pInfo->pszProtocol==NULL)?"NULL":pInfo->pszProtocol);
      ivLog->Print("      Protocol setting must be 'nocc' to use this step\n");
      return(IVR_FATAL);
   }

   if(triggerList == NULL)
   {
      ivLog->Print("      ERROR: Invalid trigger list\n");
      return(IVR_FATAL);
   }

   // Activate Call Trigger event signal
   SetGenericEvent(callTriggerQueue.GetSignal());

   // Deactivate board-level event signal, clear that queue and disable it
   ivLog->Print("      Disable and clear board event queue\n");
   ClearBoardEvent();
   if(SetActivityMode(PASSIVE) != IVR_SUCCESS)
   {
      ivLog->Print("      ERROR: Failed to disable board event queue\n");
      return(IVR_FATAL);
   }
   //
   eventQueue.Clear();

   // Wait for events
   bFound = false;
   while(!bFound
         && (((status = WaitForEvent(&boardEvent, timeout)) == IVR_SUCCESS)
             || (status == IVR_GENERIC_EVENT)))
   {
      if(status == IVR_GENERIC_EVENT)
      {
         std::string::size_type index = 0;
         MT_EVENT event;
         SYSTEMTIME eventTime;      // to output timestamp on board events

         callTriggerQueue.Get(event);
         trigger = GetEventName(event.EventCode);
         FileTimeToSystemTime((FILETIME*)&(event.TimeStamp), &eventTime);

         ivLog->Print("Call Trigger event: '%s' [qpos %i] %.2d:%.2d:%.2d.%.3d\n",
                      trigger, event.UserStatus,
                      eventTime.wHour,
                      eventTime.wMinute,
                      eventTime.wSecond,
                      eventTime.wMilliseconds);

         // Parse through list looking for event
         while(!bFound && ((index = FindCaseInsensitiveString(triggerList + index, trigger)) != std::string::npos))
         {
            // Make sure it isn't part of an event name that contains the desired event name
            // i.e. enclosed by commas or the beginning/end of list
            if(((index == 0) || (triggerList[index - 1] == ','))
               && ((triggerList[index + strlen(trigger)] == ',') || (triggerList[index + strlen(trigger)] == '\0')))
            {
               ivLog->Print("      Event is in the list of desired - @@@\n");
               RecreateTimer(hWaitTimer);   // Stop the timer
               ClearGenericEvent();         // Dectivate Call Trigger event signal

               // Capture Call information on ISDN call connected events - TF Feb.20/04
               switch (event.EventCode)
               {
	               case EVT_CALL_CONNECTED:
                  {
   		            // Shallow copy - no delete!
	      	         pCallInfo = (MT_CC_CALL_INFO*) event.ptrBuffer;
            	      // Get ISDN call direction
               	   InfoString = (pCallInfo->CallSource == MT_CC_INCOMING_CALL ? "IN" : "OUT");
	                  InfoString += ",";
   	               // Get ISDN ANI
      	            InfoString += (string)((char*)(pCallInfo->CallerNumber.Digits));
            	      InfoString += ",";
               	   // Get ISDN DNIS
	                  InfoString += (string)((char*)(pCallInfo->CalledNumber.Digits));

			   			triggerInfo = new char[InfoString.length() + 1];
				   	   strcpy(triggerInfo, InfoString.c_str());

					   	// Additional ISDN Information available:
						   //		ivLog->Print("         ISDN Call Info:\n", event.UserStatus);
   						//		ivLog->Print("            event.UserStatus = %i\n", event.UserStatus);
	   					//		ivLog->Print("            event.EventCode = %i\n", event.EventCode);
		   				//		ivLog->Print("            event.SubReason = %i\n", event.SubReason);
			   			//		ivLog->Print("            event.XtraInfo = %i\n", event.XtraInfo);
				   		//		ivLog->Print("            event.FuncCode = %i\n", event.FuncCode);
					   	//		ivLog->Print("            event.Board = %i\n", event.Board);
						   //		ivLog->Print("            event.Channel = %i\n", event.Channel);
   						//		ivLog->Print("            event.ptrBuffer = %p\n", event.ptrBuffer);
	   					//		ivLog->Print("            event.DataLength = %i\n", event.DataLength);
		   				//		ivLog->Print("            pCallInfo = event.ptrBuffer);
			   			//		ivLog->Print("                CallInfo.CallRef = %i\n", pCallInfo->CallRef);
				   		//		ivLog->Print("                CallInfo.CallSource = %i : %s\n", pCallInfo->CallSource, pCallInfo->CallSource == MT_CC_INCOMING_CALL ? "IN" : "OUT");
					   	//		ivLog->Print("                CallInfo.CallState = %i\n", pCallInfo->CallState);
						   //		ivLog->Print("                CallInfo.CallerNumber.TypeOfNumber = %i\n", pCallInfo->CallerNumber.TypeOfNumber);
   						//		ivLog->Print("                CallInfo.CallerNumber.NumberingPlan = %i\n", pCallInfo->CallerNumber.NumberingPlan);
	   					//		ivLog->Print("                CallInfo.CallerNumber.NumberOfDigits = %i\n", pCallInfo->CallerNumber.NumberOfDigits);
		   				//		ivLog->Print("                CallInfo.CallerNumber.Digits = '%s'\n", pCallInfo->CallerNumber.Digits);
			   			//		ivLog->Print("                CallInfo.CalledNumber.TypeOfNumber = %i\n", pCallInfo->CalledNumber.TypeOfNumber);
				   		//		ivLog->Print("                CallInfo.CalledNumber.NumberingPlan = %i\n", pCallInfo->CalledNumber.NumberingPlan);
					   	//		ivLog->Print("                CallInfo.CalledNumber.NumberOfDigits = %i\n", pCallInfo->CalledNumber.NumberOfDigits);
						   //		ivLog->Print("                CallInfo.CalledNumber.Digits = '%s'\n", pCallInfo->CalledNumber.Digits);
                     break;
                  }

                  case EVT_CALLID_STOP:
                  {
                     triggerInfo = new char[MAX_DIGIT_BUFFER_SIZE];
                     if(!CallerIDEvent(event, triggerInfo))
                     {
                        ivLog->Print("         ERROR: Failed to retrieve Caller ID\n");
                        triggerInfo[0] = '\0';
                        return(IVR_FATAL);
                     }
                     break;
                  }

   					case EVT_FUNCTION_BUTTON_PRESSED:
                  {
   						// MT_EVENT:
                     //		LARGE_INTEGER  TimeStamp;          // In FILETIME format
		   				//		ULONG          UserStatus;         // Not used by API
			   			//		ULONG          EventCode;          // Event Code
				   		//		ULONG          SubReason;          // Event sub-reason
					   	//		ULONG          XtraInfo;           // Extra information, e.g. termination
						   //		ULONG          FuncCode;           // Function active when this event occurred
   						//		USHORT         Board;              // Board ID
	      				//		USHORT         Channel;            // Global Channel ID (GCI)
			   			//		PVOID          ptrBuffer;          // Related play/record buffer pointer
				   		//		PVOID          ptrXtraBuffer;      // Pointer to xtra buffer
					   	//		ULONG          XtraBufferLength;   // Length of buffer pointed by ptrXtraBuffer
						   //		ULONG          XtraDataLength;     // Length of data in buffer pointed by ptrXtraBuffer
   						//		ULONG          EventFlag;          // Flags of the following:
	   					//													  //		0x00000001: 1 - Appl created the event
		   				//													  //		            0 - NTi DLL created the event
			   			//													  //		0x00000002: 1 - Appl allocated ptrtraBuffer
				   		//													  //		            0 - NTi DLL allocated ptrXtraBuffer
					   	ivLog->Print("         MT_EVENT: \n");
						   ivLog->Print("         (board %i, channel %i) \n", event.Board, event.Channel);
   						ivLog->Print("            UserStatus       = %i (not used)\n", event.UserStatus);
	   					ivLog->Print("            EventCode        = EVT_FUNCTION_BUTTON_PRESSED \n");
		   				ivLog->Print("            SubReason        = %i \n", event.SubReason);
			   			ivLog->Print("            XtraInfo         = %i \n", event.XtraInfo);
				   		ivLog->Print("            FuncCode         = %i \n", event.FuncCode);
					   	ivLog->Print("            ptrBuffer        = %p \n", event.ptrBuffer);
						   ivLog->Print("            DataLength       = %i \n", event.DataLength);
   						ivLog->Print("            ptrXtraBuffer    = %p \n", event.ptrXtraBuffer);
	   					ivLog->Print("            XtraBufferLength = %i \n", event.XtraBufferLength);
		   				ivLog->Print("            XtraDataLength   = %i \n", event.XtraDataLength);
			   			ivLog->Print("            EventFlag        = 0x%x \n", event.EventFlag);

				   	   // Big enough for a 32 bit number
            	      triggerInfo = new char[11];
               	   sprintf(triggerInfo, "%d", event.SubReason);

   						break;
                  }

   					case EVT_MESSAGE_CHANGE:
                  {
   	   	         if(event.ptrBuffer != NULL)
                     {
	            	      // Get message string
	         	         InfoString = (string)(char*)(event.ptrBuffer);
				   			triggerInfo = new char[InfoString.length() + 1];
					   	   strcpy(triggerInfo, InfoString.c_str());
                     }
                     else
                     {
							   triggerInfo = new char[1];
                        triggerInfo[0] = '\0';
                     }
   	            	break;
                  }

   					case EVT_INCOMING_SIGCHANGE:
   					case EVT_OUTGOING_SIGCHANGE:
                  {
                     ULONG bitmask = 1 << 31;
                     string outstring = "signal change: ABCD ";
                     string bitstring;

                     for(int octetcount = 3; octetcount >= 0; octetcount--)
                     {
                        for(int bitcount = 7; bitcount >= 0; bitcount--)
                        {
                           bitstring += event.SubReason & bitmask ? '1' : '0';
                           if(octetcount <= 1 && bitcount <= 3)
                           {
                              outstring += event.SubReason & bitmask ? '1' : '0';
                           }
                           bitmask = bitmask >> 1;
                        }
                        if(octetcount == 1)
                        {
                           outstring += " -> ABCD ";
                        }
                     }
                     ivLog->Print("      %s\n", outstring.c_str());
   					   // Big enough for 32 binary digits
               	   triggerInfo = new char[33];
                  	sprintf(triggerInfo, "%s", bitstring.c_str());
   	            	break;
                  }

         	      default:
                  {
   					   // Big enough for a 32 bit number
               	   triggerInfo = new char[11];
                  	sprintf(triggerInfo, "%d", event.SubReason);
                  }
			   	} // end switch

               bFound = true;           // Found an event we were looking for, break out of the loop
               status = IVR_SUCCESS;
            }

            // Increment so we search after current result
            index += strlen(trigger);
         } // end while
      } // end if
   }
   // end while

   if(!bFound)
   {
      // Got here without finding a desired event, so timer has not been stopped yet
      RecreateTimer(hWaitTimer);   // Stop the timer
      ClearGenericEvent();         // Dectivate Call Trigger event signal
      trigger = "";                // Event retrieved s/b empty string (not found)
   }

   // Re-activate board-level event queue
   ivLog->Print("      Enable board event queue\n");
   if(SetActivityMode(ACTIVE) != IVR_SUCCESS)
   {
      ivLog->Print("      ERROR: Failed to enable board event queue\n");
      return(IVR_FATAL);
   }
   SetBoardEvent(eventQueue.GetSignal());   // Reactivate board-level event signal

   if(status == IVR_GENERIC_TIMEOUT)
   {
      return(IVR_TIMEOUT);
   }

   return(status);
}
//BoardMT::CallTrigger

IV_STATUS BoardMT::FlashHook(long flashTime)
{
	MTBoardEvent boardEvent;
   IV_STATUS status;
	MT_PSTN originalPSTNParameters;
	MT_PSTN pstnParameters;
	MT_RESULT result;
   MT_IO_CONTROL ioControl;

	if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Cannot stream while port not off-hook\n");
   	return(IVR_FATAL);
   }

   ivLog->Print("         Flashing hook for %d ms\n", flashTime);

   if(flashTime == 0)
   {
      return(IVR_SUCCESS);
   }

	// Get current PSTN parameters
	unsigned long parametersSize = sizeof(MT_PSTN);
   if(MTGetPSTNParams(curChannel, &originalPSTNParameters, &parametersSize) != MT_RET_OK)
   {
      PrintMTError();
   	return(IVR_FAIL);
   }

   // Set PSTN parameters for this function
   pstnParameters = originalPSTNParameters;
   pstnParameters.flash_char = '&';
   pstnParameters.flash_time = (unsigned short) flashTime;

	// Set PSTN parameters with new flash time
   if(MTSetPSTNParams(curChannel, &pstnParameters, &parametersSize) != MT_RET_OK)
   {
      PrintMTError();
   	return(IVR_FAIL);
   }

   memset(&ioControl, 0, sizeof(MT_IO_CONTROL));

   // Dial flash hook
	if(((result = MTDialString(curChannel, "&", &ioControl)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to flash hook\n");
      PrintMTError();
      return(IVR_FAIL);
   }

	// Set PSTN parameters with original flash time
   if(MTSetPSTNParams(curChannel, &originalPSTNParameters, &parametersSize) != MT_RET_OK)
   {
      PrintMTError();
   	return(IVR_FAIL);
   }

   audioState->StartDialing();

	// Wait for board to signal flash hook completed
   while((status = WaitForEvent(&boardEvent, flashTime)) == IVR_SUCCESS)
   {
      switch(boardEvent.MTEvent.EventCode)
      {
      	case EVT_OFFHOOK:
         {
            break;
         }

      	case EVT_LOOP_REVERSE:
         case EVT_LVOLTAGE_MIDDLE:   // can indicate 'loop voltage detected' on LD cards
         {
  		   	return(IVR_SUCCESS);
         }

         case EVT_DIALED:
         {
            audioState->StopDialing();
	   	   break;
         }

         case EVT_IO_COMPLETED:
         {
  		   	return(IVR_SUCCESS);
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

   audioState->StopDialing();

   return(status == IVR_TIMEOUT ? Hangup(true) : status);
}

IV_STATUS BoardMT::Hangup(bool returnHangup, char *) // pszCause)
{
	MTBoardEvent boardEvent;
   IV_STATUS status;
   MT_RESULT result;

   if(!bBrdInit)
   {
      ivLog->Print("      ERROR: Port is not initialised\n");
      return(IVR_FATAL);
   }

   // Call deemed to start with incoming call event; lineOffHook may not have not been set;
   // therefore always update callEnd on Hangup.
   callEnd.Update();

   if(!lineOffHook)
   {
   	return(returnHangup ? IVR_HANGUP : IVR_SUCCESS);
   }

   // Clear touch-tones from the buffer
   FlushDigits();

   ivLog->Print("      Hanging up\n");

	// Reset switch state (if any)
	status = UnSwitch();
	if ((status != IVR_SUCCESS) && (status != IVR_NOTSUPPORTED))
	{
		return (status);
	}

	// Put this channel on hook
   if(((result = MTOnHook(curChannel)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to put port on hook\n");
      PrintMTError();
      return(IVR_FAIL);
   }

   // This must be set before MTOnHook to avoid recursive Hangup calls
   lineOffHook = false;

   if(audioState->IsStreaming())
   {
      StopStreaming();
   }

	// Wait for board to signal on hook completed
   while(((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS) || (status == IVR_HANGUP))
   {
      switch(boardEvent.MTEvent.EventCode)
      {
      	case EVT_ONHOOK:
         case EVT_IO_COMPLETED:
         {
            ivLog->Print("      Call done\n");
            return(returnHangup ? IVR_HANGUP : IVR_SUCCESS);
         }

         case EVT_DIGIT:
         case EVT_EOF:
         case EVT_INTERDIGIT:
         case EVT_MAXBYTES:
         case EVT_MAXTIME:
         case EVT_STOP:
         {
         	break;
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

  	return(status);
}

IV_STATUS BoardMT::Pickup(void)
{
	MTBoardEvent boardEvent;
   IV_STATUS status;
   MT_RESULT result;

   if(!bBrdInit)
   {
      ivLog->Print("      ERROR: Port is not initialised\n");
      return(IVR_FATAL);
   }

	if(lineOffHook)
   {
      ivLog->Print("      ERROR: Port already in use\n");
   	return(IVR_FATAL);
   }

	// Put this channel off hook
   if(((result = MTOffHook(curChannel)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to put port off hook\n");
      PrintMTError();

   	return(IVR_FAIL);
   }

	// Wait for board to signal off hook
   while((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS)
   {
      switch(boardEvent.MTEvent.EventCode)
      {
      	case EVT_OFFHOOK:
         case EVT_IO_COMPLETED:
         {
            lineOffHook = true;

            if((audioState->Streamers() > 0) && audioState->IsStreamingPermitted())
            {
               StartStreaming(audioState->StreamFormat());
            }

            if(disableReorderToneDetection)
            {
               ivLog->Print("      Reorder Tone detection disabled!\n");
            }

            if(pInfo->portLicense == UNLICENSED)
            {
               pInfo->ivLog->Print ("      THIS PORT IS UNLICENSED, PLAYING EVALUATION MESSAGE\n");
               status = pInfo->ivBoard->AudioPlayEvalMessage();
            }
            return(status);
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

   return(status);
}

IV_STATUS BoardMT::WaitForCall(long timeout, int numrings, bool pickup, char* calledDigits,
   char* callingDigits, char* handsetDisplay, int)
{
	MTBoardEvent boardEvent;
   IV_STATUS status = IVR_GENERIC_EVENT;
   LARGE_INTEGER waitTimeout;
   HANDLE hInterRingTimer = CreateWaitableTimer(0, false, 0);
   LARGE_INTEGER interRingTimeout;

   if(!bBrdInit)
   {
      CloseHandle(hInterRingTimer);
      ivLog->Print("      ERROR: Port is not initialised\n");
      return(IVR_FATAL);
   }

	if(lineOffHook)
   {
      CloseHandle(hInterRingTimer);
      ivLog->Print("      ERROR: Port already in use\n");
   	return(IVR_FATAL);
   }

   // Reset callerIDLength
   ANILength = 0;

   if(calledDigits != NULL)
   {
      calledDigits[0] = '\0';
   }

   if(callingDigits != NULL)
   {
      callingDigits[0] = '\0';
   }

   if(handsetDisplay != NULL)
   {
      handsetDisplay[0] = '\0';
   }

   // Minimum number of rings is 1 (0 never answers)
   if(numrings == 0)
   {
      numrings = 1;
   }

   if(MTFlushEvents(curChannel) != MT_RET_OK)
   {
      CloseHandle(hInterRingTimer);
      PrintMTError();
   	return(IVR_FAIL);
   }

   eventQueue.Clear();

	// Set timeout timer
   if(timeout != FOREVER)
   {
      waitTimeout.QuadPart = -(LONGLONG) timeout * 10000000;
      SetWaitableTimer(hGenericTimer, &waitTimeout, 0, 0, 0, false);
   }

   // Set inter-ring timer to 8 seconds
   interRingTimeout.QuadPart = -80000000;
   SetGenericEvent(hInterRingTimer);

   // Inter-ring timeout loop
   while(status == IVR_GENERIC_EVENT)
   {
   	// Wait until enough rings have occurred
      for(int counter = 0; (counter < numrings) && ((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS);)
      {
  	      switch(boardEvent.MTEvent.EventCode)
         {
            case EVT_RINGS:
            {
               SetWaitableTimer(hInterRingTimer, &interRingTimeout, 0, 0, 0, false);
               counter++;
               break;
            }

            default:
  	         {
               if(!CallerIDEvent(boardEvent.MTEvent, callingDigits))
               {
               // For debugging purposes
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
               }
            }
         }
      }
   }

	CancelWaitableTimer(hGenericTimer);
  	CancelWaitableTimer(hInterRingTimer);
   ClearGenericEvent();
   CloseHandle(hInterRingTimer);

   if(status == IVR_SUCCESS)
   {
      // Set call start with the incoming call event
      callStart.Update();
      callEnd.Invalidate();

      return(pickup ? Pickup() : IVR_SUCCESS);
   }
   else if(status == IVR_GENERIC_TIMEOUT)
   {
      return(IVR_TIMEOUT);
   }

   return(status);
}

IV_STATUS BoardMT::FlushDigits(void)
{
   // Remove digits from the software touch tone queue
   ttQueue.Clear();

   return(IVR_SUCCESS);
}

int BoardMT::BuildDTMF(const char* filter)
{
	int mask = 0;

   for(; *filter != '\0'; filter++)
   {
      switch(*filter)
      {
         case '0': mask |= MT_DIGIT_0; 	break;
         case '1': mask |= MT_DIGIT_1; 	break;
         case '2': mask |= MT_DIGIT_2; 	break;
         case '3': mask |= MT_DIGIT_3; 	break;
         case '4': mask |= MT_DIGIT_4; 	break;
         case '5': mask |= MT_DIGIT_5; 	break;
         case '6': mask |= MT_DIGIT_6; 	break;
         case '7': mask |= MT_DIGIT_7; 	break;
         case '8': mask |= MT_DIGIT_8; 	break;
         case '9': mask |= MT_DIGIT_9; 	break;
         case 'A': mask |= MT_DIGIT_A; 	break;
         case 'B': mask |= MT_DIGIT_B; 	break;
         case 'C': mask |= MT_DIGIT_C; 	break;
         case 'D': mask |= MT_DIGIT_D; 	break;
         case '*': mask |= MT_DIGIT_STAR;	break;
         case '#': mask |= MT_DIGIT_POUND;break;
         case 'X': mask = 0xFFFF;			break;
      }
   }

   return(mask);
}

IV_STATUS BoardMT::GetDigits(
      unsigned count,   // Number of digits to collect.
      char*& digits,    // Digit buffer returned from GetDigits
      int newTime,      // Timeout between key presses in ms.
      const char* ttOkay,     // Digits that are valid.
      const char* ttDone)     // Digits that terminate TT collection.
{
	MTBoardEvent boardEvent;
   IV_STATUS status = IVR_SUCCESS;
   bool terminate = false;
   LARGE_INTEGER waitTimeout;
   string digitString;
   char tt;

//   if(!lineOffHook)
//   {
//      ivLog->Print("      ERROR: Port not off hook\n");
//      digits = new char[1];
//      digits[0] = '\0';
//      return(IVR_FATAL);
//   }

   // newTime is in ms
   waitTimeout.QuadPart = -(LONGLONG) newTime * 10000;

	// If there is an X anywhere in the string then all digits are acceptable
   if(strchr(ttOkay, 'X'))
   {
   	ttOkay = "0123456789ABCD*#";
   }

	// If there is an X anywhere in the string then all digits are acceptable
   if(strchr(ttDone, 'X'))
   {
   	ttDone = "0123456789ABCD*#";
   }

   // RSet timeout timer
   if(newTime != FOREVER)
   {
      SetWaitableTimer(hGenericTimer, &waitTimeout, 0, 0, 0, false);
   }

	// Until enough digits or other terminating condition
   while(((count > 0) || (count == INFINITE)) && !terminate && !ttQueue.Empty())
   {
      // Get digit from software touch tone queue
  		ttQueue.Get(tt);

      // Check for termination digit
      if((tt != '\0') && strchr(ttDone, tt))
      {
         terminate = true;
      }

      // Keep touch tone only if it is in ttOkay
      if((tt != '\0') && strchr(ttOkay, tt))
      {
         digitString += tt;
         if(count != INFINITE)
         {
            count--;
         }
      }
   }

	// Until enough digits or other terminating condition
	while(((count > 0) || (count == INFINITE)) && !terminate)
   {
      // Wait for digit event from board event queue
      status = WaitForEvent(&boardEvent);

      switch(status)
      {
         case IVR_GENERIC_TIMEOUT:
         {
            // If we timed out with no digits return IVR_TIMEOUT
            status = digitString.length() == 0 ? IVR_TIMEOUT : IVR_SUCCESS;
            terminate = true;
            break;
         }

         case IVR_SUCCESS:
         {
				if(boardEvent.MTEvent.EventCode == EVT_DIGIT)
            {
         		// Reset timeout timer
            	if(newTime != FOREVER)
         	   {
            	   SetWaitableTimer(hGenericTimer, &waitTimeout, 0, 0, 0, false);
         	   }

               // Wait for digit to show up in software queue
	   	   	WaitForSingleObject(ttQueue.GetSignal(), INFINITE);

               // Get digit from software touch tone queue
  	   	      ttQueue.Get(tt);

               // Check for termination digit
               if((tt != '\0') && strchr(ttDone, tt))
               {
                  terminate = true;
               }

               // Keep touch tone only if it is in ttOkay
               if((tt != '\0') && strchr(ttOkay, tt))
               {
                  digitString += tt;
                  if(count != INFINITE)
                  {
                     count--;
                  }
               }
            }

            break;
         }

         default:
         {
            terminate = true;
         }
      }
   }

   // Stop timeout timer
	CancelWaitableTimer(hGenericTimer);

   digits = new char[digitString.length() + 1];
   strcpy(digits, digitString.c_str());
   strupr(digits);

	return(status);
}

IV_STATUS BoardMT::GenerateDigits(const char* ttStr)
{
	MTBoardEvent boardEvent;
   IV_STATUS status;
   MT_IO_CONTROL dialParams;
	MT_RESULT result;

   if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Port not off hook\n");
      return(IVR_FATAL);
   }

   FlushDigits();

   memset(&dialParams, 0, sizeof(dialParams));

	// Dial digits
	if(((result = MTDialString(curChannel, ttStr, &dialParams)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      PrintMTError();
   	return(IVR_FAIL);
   }

   // Wait for board to signal that dialing is complete. Is there a better event to wait for?
   while((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS)
   {
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_DIALED:
         case EVT_IO_COMPLETED:
         {
            Sleep(200);
            FlushDigits();
            return(IVR_SUCCESS);
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

   Sleep(200);
   FlushDigits();
  	return(status);
}

IV_STATUS BoardMT::QueryDigit(char& digit)
{
   if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Port not off hook\n");
      return(IVR_FATAL);
   }

	// Get a copy of the first digit in the queue (does not remove from queue!)
	if(!ttQueue.Peek(digit))
   {
      // No digit in queue so return NULL
   	digit = '\0';
   }

   return(IVR_SUCCESS);
}

IV_STATUS BoardMT::AudioBeep(bool interruptable)
{
	MTBoardEvent boardEvent;
   MT_PLAYTONE tone;
   MT_IO_CONTROL toneIO;
   char ttDig;
   IV_STATUS status;
   MT_RESULT result;

   if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Port not off hook\n");
      return(IVR_FATAL);
   }

	QueryDigit(ttDig);
   if(interruptable && (ttDig != '\0'))
   {
      ivLog->Print("      Interrupted!\n");
      return(IVR_SUCCESS);
   }

	// Clear parameter
	memset(&tone, 0, sizeof(MT_PLAYTONE));

	// Set tone parameters
	tone.freq1 = 1000;   // 1000 Hz
   tone.ampl1 = 0;   	// Regular volume
   tone.durtone = 200;	// 200 ms duration

	// Clear parameter
	memset(&toneIO, 0, sizeof(MT_IO_CONTROL));

   toneIO.Buffer = (unsigned char*) &tone;
   toneIO.Length = sizeof(MT_PLAYTONE);

   // Play tone
   if(((result = MTPlayTone(curChannel, &toneIO)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to play tone on port\n");
      PrintMTError();
      return(IVR_FAIL);
   }

   // Wait for board to signal tone completed
   while((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS)
   {
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_TERMSILENCE:
         case EVT_TONEP_DONE:
         case EVT_IO_COMPLETED:
         {
            return(IVR_SUCCESS);
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

   return(status);
}

short BoardMT::GetVolumeLevel(int volLevel)
{
	// Adjust linear volLevel (-3 to 3) to 64dB to +24dB
   return((short) (20*log10(((double) volLevel + 4)/7)));
}

unsigned int BoardMT::GetEncoding(SOUND_FORMAT format)
{
	switch(format)
   {
   	case TF_WAV_8KHZ_16BIT:	return(MT_PCM_Raw_16bit);	// WAV format- 8000 sampled rate and 16bits
   	case TF_WAV_8KHZ_16BIT_G610:	return(MT_MSGSM);	   // Microsoft GSM
      default: return(MT_PCM_Raw_16bit);
   }
}

IV_STATUS BoardMT::StartRecordFile(const char* filename, bool interruptable,
   SOUND_FORMAT format, int, int maxRecTime, int maxSilence, bool disableBeep,
   bool automaticGainControl)
{
	MTBoardEvent boardEvent;
	MT_IO_CONTROL ioCtrl;
   MT_RESULT result;
   char ttDig;
   IV_STATUS status;

   if(!bBrdInit)
   {
      ivLog->Print("      ERROR: Port is not initialised\n");
      return(IVR_FATAL);
   }

   if(audioState->IsRecording())
   {
      ivLog->Print("      ERROR: Already recording\n");
      return(IVR_FATAL);
   }

   // Only query for non-passive records
   if(lineOffHook)
   {
      // Beep has to occur before interrupt check in case it is interrupted
      if(!disableBeep)
      {
         status = AudioBeep(interruptable);

         // So that beep doesn't show up in recording
         Sleep(100);

         if(status != IVR_SUCCESS)
         {
            return(status);
         }
      }

      // Check for pending digit
	   QueryDigit(ttDig);

   	// Check for interruption
	   if(interruptable  && (ttDig != '\0'))
      {
         // Skip record if interruptable and digit pending
  	      ivLog->Print("      Interrupted!\n");
       	return(IVR_SUCCESS);
      }

   }

   if(pInfo->portLicense == UNLICENSED)
   {
		// Disable recording until we can make the Ai-Logix timeout work - TF Aug.15/03
      ivLog->Print ("      THIS PORT IS UNLICENSED, RECORDING TERMINATED\n");
      return(IVR_NOTSUPPORTED);
   }

   format = (format == TF_BY_EXTENSION) ? GetExtensionFormat(filename) : format;

	// Clear parameter
	memset(&ioCtrl, 0, sizeof(MT_IO_CONTROL));

	// Set maximum record time
   ioCtrl.MaxTime = (maxRecTime == FOREVER) ? 0 : maxRecTime;

   // Set silence timeout
   ioCtrl.MaxSilence = (maxSilence == FOREVER) ? 0 : maxSilence;

   // Terminate on digit
   // Doesn't work in the AiLogix PassPort 2.2.0 API so trapped via EVT_DIGIT
   ioCtrl.EndDigit = interruptable ? '@' : '\0';

	// Inter-digit delay starts after the first digit
   ioCtrl.StartControl = INTERDIGIT_DELAY;

   // Set file/audio format
   ioCtrl.FileType = GetEncoding(format);

   // Set streaming to not terminate on Call Progress Monitoring events
   ioCtrl.CpmMode = CPMMODE_MONI;

   // Set streaming watermark
   ioCtrl.Length = audioState->AudioBufferSize(format);

   // Enable/disable automatic gain control
   if(MTAGCControl(curChannel, automaticGainControl) != MT_RET_OK)
   {
   	ivLog->Print("ERROR: Failed to set AGC\n");
      PrintMTError();
	}

   // Check for streaming
   if(audioState->IsStreaming())
   {
      status = StopStreaming();

      if(status != IVR_SUCCESS)
      {
         return(status);
      }
   }

   // Record from stream
   ivLog->Print("      BoardMT::StartRecordFile: MTStartStreaming, ioCtrl.MaxSilence = %i\n", ioCtrl.MaxSilence);   // @@@ debug
   result = MTStartStreaming(curChannel, MT_STREAM_INPUT, &ioCtrl);

   if((result != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to record on port\n");
      PrintMTError();
   	return(IVR_FAIL);
   }

 	while(((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS) && (boardEvent.MTEvent.EventCode != EVT_STREAMIN_START))
   {
      // For debugging purposes
//  	   ivLog->Print("      WARNING: Consumed MT board event '%s'\n"
//       	"         User Status: %d\n         SubReason: %d\n"
//         "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//         boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
   }

   if(status != IVR_SUCCESS)
   {
      return(status);
   }

   // Update audioState
   audioState->StartStreaming();
   audioState->StartStreamers(format);
   audioState->RestartStreaming(true);
   audioState->StartRecording(filename, format);
   audioState->SetInterruptable(interruptable);
   audioState->ResetVoiceRecorded();

   status = maxRecTime == 0 ? AudioStopRecord() : IVR_SUCCESS;

   return(status);
}

/* -------------------------------------------------------------------------
   Name:        BoardMT::StartRecordBuffer                          (public)

   TO DO:       -  UNABLE TO MAKE StartRecordBuffer() WORK - TF Jun.10/04
                   -  ENCOUNTERED PROBLEMS WITH STREAMING
                   -  MAY NEED HELP FROM SB
                -  INSTEAD:
                   - OVERRODE BRD_GENERIC FUNCTION AudioRecordBuffer()
                   - SEE BoardMT::AudioRecordBuffer()
   ------------------------------------------------------------------------- */
/*
IV_STATUS BoardMT::StartRecordBuffer(char* buffer, unsigned int bufsize,
	bool interruptable, SOUND_FORMAT format, int volLevel, int maxSilence,
   bool disableBeep, bool automaticGainControl)
{
   IV_STATUS      status;
	MTBoardEvent   boardEvent;
	MT_IO_CONTROL  ioCtrl;
   MTCHAN_STATUS  channelStatus;
   MT_RESULT      result;
   char				ttDig;      		// used to peek touch-tone digits.
   // bool           loopAgain = true;
	// int            samples = 0;
   // unsigned long  channelStatusEvents;

	ivLog->Print("         StartRecordBuffer 0x%x, %i bytes, \n", buffer, bufsize);
	ivLog->Print("            interruptable = %i \n", interruptable);
	ivLog->Print("            format = %i (%i)\n", format, TF_WAV_8KHZ_16BIT);
	ivLog->Print("            volLevel = %i\n", volLevel);
	ivLog->Print("            maxSilence = %i\n", maxSilence);
	ivLog->Print("            disableBeep = %i\n", disableBeep);
	ivLog->Print("            automaticGainControl = %i\n", automaticGainControl);

	if ((buffer == NULL) || (bufsize < 1))
   {
      ivLog->Print("      ERROR: Invalid audio buffer\n");
   	return(IVR_FATAL);
   }

   if(audioState->IsRecording())
   {
      ivLog->Print("      ERROR: Already recording\n");
      return(IVR_FATAL);
   }

   // Only query for non-passive records
   if(lineOffHook)
   {
		ivLog->Print("         Line off hook\n");
      // Beep has to occur before interrupt check in case it is interrupted
      if(!disableBeep)
      {
			ivLog->Print("         Beep\n");
         status = AudioBeep(interruptable);

         // So that beep doesn't show up in recording
         Sleep(100);

         if(status != IVR_SUCCESS)
         {
            return(status);
         }
      }

      // Check for pending digit
      ivLog->Print("         Check digit\n");
	   QueryDigit(ttDig);

   	// Check for interruption
	   if(interruptable && (ttDig != '\0'))
      {
         // Skip record if interruptable and digit pending
  	      ivLog->Print("      Interrupted!\n");
       	return(IVR_SUCCESS);
      }
   }

   if (format != TF_WAV_8KHZ_16BIT)		// (format == TF_BY_EXTENSION)
   {
      ivLog->Print("      ERROR: Invalid recording format, only TF_WAV_8KHZ_16BIT supported\n");
   	return(IVR_NOTSUPPORTED);
   }

	// Get enabled event information
	unsigned long statusSize = sizeof(MTCHAN_STATUS);
   ivLog->Print("         MTGetChannelStatus\n");
   if(MTGetChannelStatus(curChannel, &channelStatus, &statusSize) != MT_RET_OK)
   {
      ivLog->Print("      ERROR: Failed to get port status\n");
      PrintMTError();
   	return(IVR_FATAL);
   }

   // channelStatusEvents = channelStatus.EventEnable;

	// Enable silence detection
   ivLog->Print("         MTSetEventFilters\n");
   MTSetEventFilters(curChannel, (unsigned short) (channelStatus.EventEnable | SE_SIL), 1);

	// Clear parameter
	memset(&ioCtrl, 0, sizeof(MT_IO_CONTROL));

   // Set maximum silence time
   ioCtrl.MaxSilence = maxSilence;

   // Terminate on digit: if interruptable, any; otherwise none
   ioCtrl.EndDigit = interruptable ? '@' : '\0';

   // Automatic Gain Control is engaged by default

   // Set file/audio format
   ioCtrl.FileType = GetEncoding(TF_WAV_8KHZ_16BIT);

   // Deliberately shallow copy, use the buffer created by calling module
	ioCtrl.Buffer = (unsigned char*) buffer;

   ioCtrl.Length = bufsize;

   // Enable/disable automatic gain control
   ivLog->Print("         MTAGCControl\n");
   if(MTAGCControl(curChannel, automaticGainControl) != MT_RET_OK)
   {
   	ivLog->Print("ERROR: Failed to set AGC\n");
      PrintMTError();
	}

   // Check for streaming
   ivLog->Print("         IsStreaming?\n");
   if(audioState->IsStreaming())
   {
	   ivLog->Print("         StopStreaming\n");
      status = StopStreaming();

      if(status != IVR_SUCCESS)
      {
         return(status);
      }
   }

   // IN BoardMT::StartRecordFile, BUT NOT IN BoardNMS::StartRecordBuffer()
   //		ivLog->Print("         RestartStreaming\n");
   //		audioState->RestartStreaming(true);

   // Record
   ivLog->Print("         MTRecBuffer\n");
   result = MTRecBuffer(curChannel, &ioCtrl);
   if((result != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to record on port\n");
      PrintMTError();
   	return(IVR_FAIL);
   }

	// Update audioState
   ivLog->Print("         Update audioState: \n");
   ivLog->Print("            StartStreaming \n");
   audioState->StartStreaming();			// FROM BRD_NMS
   ivLog->Print("            StartStreamers \n");
   audioState->StartStreamers(format);	// FROM BRD_NMS
   ivLog->Print("            StartRecording\n");
   audioState->StartRecording(NULL, format);	// no file name, recording to buffer
   ivLog->Print("            SetInterruptable\n");
   audioState->SetInterruptable(interruptable);
   ivLog->Print("            ResetVoiceRecorded\n");
   audioState->ResetVoiceRecorded();
   //
	// BoardNMS::StartRecordBuffer DOES THIS...
   //		// Update audioState
   //		audioState->StartStreaming();
   //		audioState->StartStreamers(format);
   //		audioState->RestartStreaming(true);
   //		audioState->StartRecording(NULL, format);
   //		audioState->SetInterruptable(interruptable);
   //

   return(IVR_SUCCESS);
}
*/ //BoardMT::StartRecordBuffer

IV_STATUS BoardMT::AudioWaitForRecordStop(void)
{
   MTBoardEvent boardEvent;
   IV_STATUS status;
   IV_STATUS status2 = IVR_SUCCESS;

   // @@@ debug
   ivLog->Print("      BoardMT::AudioWaitForRecordStop\n");
   // @@@ end debug
   if(!audioState->IsStreaming())
   {
      ivLog->Print("      (@@@) SUCCESS because AudioState thinks we're not streaming\n");
      return(IVR_SUCCESS);
   }

	// Wait for board to signal record completed
   while((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS)
   {
      // @@@ debug:
      ivLog->Print("      BoardMT::AudioWaitForRecordStop processing %s\n",
                   GetEventName(boardEvent.MTEvent.EventCode));
      // @@@ end debug.
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_MON_ACTIVITY:
         {
            audioState->SetVoiceRecorded();
            break;
         }

      	case EVT_EOF:
         case EVT_MAXBYTES:
         case EVT_MAXTIME:
         {
            status2 = IVR_TIMEOUT;
            ivLog->Print("      Maximum record time reached\n");
// Temporary system variable until TFDesigner can warn users about changes between versions
            pInfo->variables->SetSystem("@RECORD_RESULT", "MAX LENGTH");
            return(status2);
         }

         case EVT_MAX_SILENCE:
         case EVT_TERMSILENCE:
         {
            if(!audioState->IsVoiceRecorded())
            {
               ivLog->Print("      No voice detected\n");
// Removed until TFDesigner can warn users about changes between versions
//               remove(audioState.GetRecordFileName());
//               status2 = IVR_NO_VOICE;
// Temporary system variable until TFDesigner can warn users about changes between versions
               pInfo->variables->SetSystem("@RECORD_RESULT", "NO VOICE");
               status2 = IVR_SUCCESS;
            }
            else
            {
// Temporary system variable until TFDesigner can warn users about changes between versions
               pInfo->variables->SetSystem("@RECORD_RESULT", "SPEECH STOP");
            }

            // Continue to wait until we get the EVT_STREAMIN_STOP event
            break;
         }

         case EVT_STOP:
         {
// Temporary system variable until TFDesigner can warn users about changes between versions
            char ttDig;
            QueryDigit(ttDig);
            if(ttDig == '\0')
            {
               pInfo->variables->SetSystem("@RECORD_RESULT", "STOPPED");
            }

            return(status2);
         }

         case EVT_STREAMIN_STOP:
         case EVT_IO_COMPLETED:
         {
           	return(status2);
         }

         case EVT_TERMDIGIT:
         {
            // Interrupt recording
				if(audioState->IsInterruptable())
			   {
               fstream recordFile(audioState->GetRecordFileName(), std::ios::in | std::ios::out | std::ios::binary);

               // Temporary system variable until TFDesigner can warn users about changes between versions
               pInfo->variables->SetSystem("@RECORD_RESULT", "INTERRUPTED");

					// If recording to a buffer, record file name will be empty
               if((audioState->GetRecordFileName() != "") && (recordFile.is_open()))
               {
	               recordFile.close();
               }

               return(IVR_SUCCESS);
			   }

         	break;
         }

         default:
  	      {
            // For debugging purposes
            ivLog->Print("      WARNING: Consumed MT board event '%s'\n"
               "         User Status: %d\n         SubReason: %d\n"
               "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
               boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
         }
      }
   }

   AudioStopRecord();

   return(status);
}

IV_STATUS BoardMT::AudioRecognise(int, int)
{
   return(IVR_NOTSUPPORTED);
}

/*
IV_STATUS BoardMT::AudioRecognise(int noVoiceTimeout, int endOfVoiceTimeout)
{
   IV_STATUS status;
	MTBoardEvent boardEvent;
	MT_IO_CONTROL ioCtrl;
   MTCHAN_STATUS channelStatus;
   MT_RESULT result;
   bool loopAgain = true;
	unsigned int bytes = 0;
   unsigned long channelStatusEvents;
   char messageBuffer[1024];

   if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Port not off hook\n");
      return(IVR_FATAL);
   }

   if(pInfo->speechRecognition == NULL)
   {
      ivLog->Print("      ERROR: No speech recognition module loaded\n");
      ivLog->Print("         Make sure to use SR Acquire before SR steps\n");
      return(IVR_FATAL);
   }

	// Get enabled event information
	unsigned long statusSize = sizeof(MTCHAN_STATUS);
   if(MTGetChannelStatus(curChannel, &channelStatus, &statusSize) != MT_RET_OK)
   {
      ivLog->Print("      ERROR: Failed to get port status\n");
      PrintMTError();
   	return(IVR_FATAL);
   }

   channelStatusEvents = channelStatus.EventEnable;

	// Enable silence detection
   MTSetEventFilters(curChannel, (unsigned short) (channelStatus.EventEnable | SE_SIL), 1);

	// Clear parameter
	memset(&ioCtrl, 0, sizeof(MT_IO_CONTROL));

   // Set maximum silence time
   ioCtrl.MaxSilence = silenceTimeout;

	// Do not terminate on any digit
   ioCtrl.EndDigit = '\0';

   // Automatic Gain Control is engaged by default

   // Set file/audio format
   ioCtrl.FileType = GetEncoding(TF_WAV_8KHZ_16BIT);

  	// 8 samples/ms, 2 bytes/sample
   ioCtrl.Buffer = new unsigned char[8*2*timeout];

  	// 8 samples/ms, 2 bytes/sample
   ioCtrl.Length = 8*2*timeout;

   if(ioCtrl.Buffer == NULL)
   {
   	ivLog->Print("      ERROR: Insufficient memory.  AudioRecognise unable to allocate recording buffer\n");
      return(IVR_FAIL);
   }

   // Check for streaming
   if(audioState->IsStreaming())
   {
      status = StopStreaming();

      if(status != IVR_SUCCESS)
      {
         pInfo->speechRecognition->RecogEnd(messageBuffer);
         return(status);
      }
   }

   audioState->RestartStreaming(true);

   // Record
   if(((result = MTRecBuffer(curChannel, &ioCtrl)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to record on port\n");
      PrintMTError();
      pInfo->speechRecognition->RecogEnd(messageBuffer);
   	return(IVR_FATAL);
   }

   audioState->StartRecognising(pInfo->speechRecognition);

   boardEvent.MTEvent.EventCode = 0;

	// Wait for board to signal record completed
   while(loopAgain && ((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS))
   {
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_MAXBYTES:
         {
            bytes = ioCtrl.Length;
            loopAgain = false;
            break;
         }

         case EVT_TERMSILENCE:
         {
            loopAgain = false;
            break;
         }

			case EVT_MAX_SILENCE:
         {
         	// Get information from the event generated
            if(MTGetChannelStatus(curChannel, &channelStatus, &statusSize) != MT_RET_OK)
            {
               ivLog->Print("      ERROR: Failed to get port status\n");
               PrintMTError();
               pInfo->speechRecognition->RecogEnd(messageBuffer);
            	return(IVR_FATAL);
            }

            bytes = channelStatus.Bytes;

            break;
         }

         case EVT_IO_COMPLETED:
         {
            loopAgain = false;
         	break;
         }

         case EVT_DIGIT:
         {
            break;
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

   // Reset enabled events
   MTSetEventFilters(curChannel, (unsigned short) channelStatusEvents, 1);

   if(status != IVR_SUCCESS)
   {
      pInfo->speechRecognition->RecogEnd(messageBuffer);
   	return(status);
   }

   if(bytes == 0)
   {
      ivLog->Print("      ERROR: Failed to acquire audio samples\n");
      pInfo->speechRecognition->RecogEnd(messageBuffer);
      return(IVR_FAIL);
   }

   if(pInfo->speechRecognition->RecogProcessAudio((const char*) ioCtrl.Buffer, &bytes, messageBuffer) != IVR_SUCCESS)
   {
   	ivLog->Print("      ERROR: Failed to process audio buffer\n");
   	ivLog->Print("         %s\n", messageBuffer);
      return(IVR_FAIL);
   }

   return(IVR_SUCCESS);
}
*/

IV_STATUS BoardMT::StartPlayFile(const char* filename, bool interruptable,
	SOUND_FORMAT format, int volLevel, unsigned int playFromTime, unsigned int playForTime)
{
	MTBoardEvent	boardEvent;
	MT_IO_CONTROL	ioCtrl;
   MT_RESULT result;
   char tt;

   if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Port not off hook\n");
      return(IVR_FATAL);
   }

   if(access(filename, 00) != 0)
   {
		ivLog->Print("      File does not exist\n");
   	return(IVR_FAIL);
   }

   if(audioState->IsPlaying())
   {
      ivLog->Print("      ERROR: Already playing\n");
      return(IVR_FATAL);
   }

	QueryDigit(tt);
   if(interruptable && (tt != '\0'))
   {
      ivLog->Print("      Interrupted!\n");
      return(IVR_SUCCESS);
   }

	// Set volume
   if(MTSetVolume(curChannel, GetVolumeLevel(volLevel)) != MT_RET_OK)
   {
      ivLog->Print("      ERROR: Failed to set playback volume on port\n");
      PrintMTError();
   	return(IVR_FAIL);
   }

   format = (format == TF_BY_EXTENSION) ? GetExtensionFormat(filename) : format;

	// Clear parameter
	memset(&ioCtrl, 0, sizeof(MT_IO_CONTROL));

   // Length of time to play
   ioCtrl.MaxTime = playForTime == FOREVER ? 0 : playForTime;

   // Terminate on digit
   ioCtrl.EndDigit = interruptable ? '@' : '\0';

   // Set file/audio format
   ioCtrl.FileType = GetEncoding(format);

   // Time position to start playing from (in bytes)
   // Only 16 bit 8 kHz PCM is supported at this time
   if((ioCtrl.FileType == MT_PCM_Raw_16bit) && (playFromTime > 0))
   {
      fstream audioFile(filename, std::ios::in | std::ios::binary);
      unsigned char sizeBuffer[4];
      unsigned int audioLength;

      if(audioFile == NULL)
      {
         ivLog->Print("      ERROR: Could not open audio file\n");
         return(IVR_FAIL);
      }

      // Seek to RIFF size (4 chars from beginning of file)
      audioFile.seekg(4, std::ios::beg);

      // Get RIFF size
      audioFile.read((char*) sizeBuffer, 4);

      audioFile.close();

      // Translate small-endian 4 byte sequence into an integer
      audioLength = sizeBuffer[0] + (sizeBuffer[1] << 8) + (sizeBuffer[2] << 16) + (sizeBuffer[3] << 24);

      ioCtrl.FileIndex = playFromTime * 2 * 8;

      if(ioCtrl.FileIndex > audioLength)
      {
         ioCtrl.FileIndex = audioLength;
      }
   }

   // Without this flag the header is interpreted as sound which clicks
   ioCtrl.StartControl = WAVE_RIFFFMTDATA_CHUNK;

	// Play file
	if(((result = MTPlayFile(curChannel, filename, &ioCtrl)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to play file on port\n");
      PrintMTError();
   	return(IVR_FAIL);
   }

   audioState->StartPlaying(format);
   audioState->SetInterruptable(interruptable);

   if(playForTime == 0)
   {
      AudioStopPlay();
   }

   return(IVR_SUCCESS);
}

IV_STATUS BoardMT::StartPlayBuffer(char* audioBuffer, unsigned bufferSize, bool interruptable,
   SOUND_FORMAT format, int volLevel)
{
	MTBoardEvent boardEvent;
	MT_IO_CONTROL ioCtrl;
   MT_RESULT result;

   if(audioBuffer == NULL)
   {
      ivLog->Print("      ERROR: No audio buffer\n");
      return(IVR_FAIL);
   }

   if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Port not off hook\n");
      return(IVR_FATAL);
   }

   if(audioState->IsPlaying())
   {
      ivLog->Print("      ERROR: Already playing\n");
      return(IVR_FATAL);
   }

	// Set volume
   if(MTSetVolume(curChannel, GetVolumeLevel(volLevel)) != MT_RET_OK)
   {
      ivLog->Print("      ERROR: Failed to set playback volume on port\n");
      PrintMTError();
   	return(IVR_FAIL);
   }

	// Clear parameter
	memset(&ioCtrl, 0, sizeof(MT_IO_CONTROL));

   // Terminate on digit
   ioCtrl.EndDigit = interruptable ? '@' : '\0';

   // Set audio format
   ioCtrl.FileType = GetEncoding(format);

  	// Audio buffer
   ioCtrl.Buffer = (unsigned char*) audioBuffer;

  	// Size of audio buffer
   ioCtrl.Length = bufferSize;

	// Play file
	if(((result = MTPlayBuffer(curChannel, &ioCtrl)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to play buffer on port\n");
      PrintMTError();
   	return(IVR_FAIL);
   }

   audioState->StartPlaying(format);
   audioState->SetInterruptable(interruptable);

   return(IVR_SUCCESS);
}

IV_STATUS BoardMT::AudioStopPlay(void)
{
   MT_RESULT result;

   if(!audioState->IsPlaying())
   {
      return(IVR_SUCCESS);
   }

   // The spacing of MTStopCurrentFunction(curChannel,MT_PLAY) cannot be
   // because of the defines in brd_mtav_code.h
   if((result = MTStopCurrentFunction(curChannel,MT_PLAY)) != MT_RET_OK)
   {
      if(result == MT_RET_IDLE)
      {
         return(IVR_SUCCESS);
      }

      ivLog->Print("      ERROR: Failed to stop file playback on port\n");
      PrintMTError();
      return(IVR_FAIL);
   }

   return(AudioWaitForPlayStop());
}

IV_STATUS BoardMT::AudioWaitForPlayStop(void)
{
   MTBoardEvent boardEvent;
   IV_STATUS status;

   if(!audioState->IsPlaying())
   {
      return(IVR_SUCCESS);
   }

	// Wait for board to signal play completed
   while((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS)
   {
      switch(boardEvent.MTEvent.EventCode)
      {
      	case EVT_EOF:
         case EVT_MAXBYTES:
         case EVT_MAXTIME:
         case EVT_STOP:
         case EVT_IO_COMPLETED:
         {
            return(IVR_SUCCESS);
         }

         case EVT_DIGIT:
         {
				if(audioState->IsInterruptable())
			   {
               return(IVR_SUCCESS);
			   }

         	break;
         }

         default:
  	      {
            // For debugging purposes
//            if(!CallerIDEvent(boardEvent, callingDigits))
//            {
//     	      	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	"         User Status: %d\n         SubReason: %d\n"
//                  "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                  boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//            }
         }
      }
   }

   AudioStopPlay();

   return(status);
}

IV_STATUS BoardMT::AudioPlayEvalMessage(void)
{
	MTBoardEvent boardEvent;
   IV_STATUS status;
   HINSTANCE hLibrary = LoadLibrary("brd_mt.dll");
   HRSRC audioBufferHandle = FindResource(hLibrary, "EVAL_MESSAGE", "WAVE");

   status = StartPlayBuffer((char*) LoadResource(hLibrary, audioBufferHandle), SizeofResource(hLibrary, audioBufferHandle), false, TF_WAV_8KHZ_16BIT, 0);

   if(status != IVR_SUCCESS)
   {
      return(status);
   }

   return(AudioWaitForPlayStop());
}

IV_STATUS BoardMT::HTTPRequest(HTTPTransaction& http, int timeout, int connectionTimeout,
   int connectionRetries, const char* filename, SOUND_FORMAT format, int volLevel)
{
   LARGE_INTEGER liTimeout;
   IV_STATUS status;
   MTBoardEvent boardEvent;
   ESCTime startTime;
   ESCTime silenceStartTime;
   ESCTime currentTime;
   int ret;
   bool loopAgain;
   bool attemptConnect;
   int connectCounter = 0;

   if(!lineOffHook)
   {
      ivLog->Print("      ERROR: Port not off hook\n");
      return(IVR_FATAL);
   }

   // Start playing file here
   ivLog->Print("      Playing '%s'\n", filename);
   status = StartPlayFile(filename, false, format, volLevel);
   if(status != IVR_SUCCESS)
   {
      ivLog->Print("      ERROR: Failed to play file '%s'\n", filename);
      return(status);
   }

   startTime.Update();

	// Set connection timeout timer
   if(connectionTimeout != FOREVER)
   {
      liTimeout.QuadPart = -(LONGLONG) connectionTimeout*10000;
      SetWaitableTimer(hGenericTimer, &liTimeout, 0, 0, 0, false);
   }

   // Set Generic Event to socket handle
   SetGenericEvent(http.GetEventHandle());

   // Try up to iConnectionRetries times to connect
   loopAgain = true;
   attemptConnect = true;
   while(loopAgain)
   {
      // Connect to HTTP server
      if(attemptConnect)
      {
         ivLog->Print("      Connection attempt %d\n", ++connectCounter);

         if((ret = http.Connect()) != Socket::SUCCESS)
         {
            const char* errorMessage = Socket::WSAErrorString(ret);

            ivLog->Print("      ERROR: Unable to connect to host\n");
            ivLog->Print("         '%s'\n", errorMessage);
           	CancelWaitableTimer(hGenericTimer);
            ClearGenericEvent();
            http.Response()->SetText(errorMessage);
            return(IVR_FAIL);
         }
      }

      status = WaitForEvent(&boardEvent);

      switch(status)
      {
         // Board event
         case IVR_SUCCESS:
         {
         	switch(boardEvent.MTEvent.EventCode)
            {
            	case EVT_EOF:
               case EVT_MAXBYTES:
               case EVT_MAXTIME:
               case EVT_STOP:
               case EVT_IO_COMPLETED:
               {
                  // Reset timer if necessary
                  if(connectionTimeout != FOREVER)
                  {
                     silenceStartTime.Update();

                     if(timeout < connectCounter*connectionTimeout - (int) (1000*silenceStartTime.GetSecondsSince(startTime)))
                     {
                        liTimeout.QuadPart = -(LONGLONG) timeout * 10000;
                        SetWaitableTimer(hGenericTimer, &liTimeout, 0, 0, 0, false);
                     }
                  }
                  else if(timeout != FOREVER)
                  {
                     liTimeout.QuadPart = -(LONGLONG) timeout * 10000;
                     SetWaitableTimer(hGenericTimer, &liTimeout, 0, 0, 0, false);
                  }

                  break;
   		   	}

               case EVT_DIGIT:
               {
                  break;
               }

               default:
  	            {
                  // For debugging purposes
//                if(!CallerIDEvent(boardEvent, callingDigits))
//                {
//     	      	   ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	      "         User Status: %d\n         SubReason: %d\n"
//                      "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                      boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//                }
               }
            }

            break;
         }

         // Connection signalled
         case IVR_GENERIC_EVENT:
         {
            // Check status
            if((ret = http.GetStatus()) == Socket::SUCCESS)
            {
               ivLog->Print("      Connection successful\n");

               http.SelectEvents(FD_READ);

               loopAgain = false;

               if(silenceStartTime.IsValid())
               {
                  liTimeout.QuadPart = -(LONGLONG) (timeout  - (int) (1000*silenceStartTime.GetSecondsSince(startTime))) * 10000;
                  SetWaitableTimer(hGenericTimer, &liTimeout, 0, 0, 0, false);
               }
               else
               {
                 	CancelWaitableTimer(hGenericTimer);
               }
            }
            else
            {
               const char* errorMessage = Socket::WSAErrorString(ret);

               ivLog->Print("      Connection attempt failed\n");
               ivLog->Print("         '%s'\n", errorMessage);

               if(connectCounter > connectionRetries)
               {
                  http.Response()->SetText(errorMessage);
                  return(IVR_FAIL);
               }

               http.Close();
               attemptConnect = true;
            }

            break;
         }

         // Timeout
         case IVR_GENERIC_TIMEOUT:
         {
            int updatedTimeout;

             ivLog->Print("      ERROR: Timed out trying to connect to host\n");

            if((timeout != FOREVER) && silenceStartTime.IsValid())
            {
               currentTime.Update();
               updatedTimeout = timeout - (int) (1000*currentTime.GetSecondsSince(silenceStartTime));
            }
            else
            {
               updatedTimeout = connectionTimeout;
            }

            // Silence timeout or maximum connection attempts
            if((updatedTimeout <= 0) || (connectCounter > connectionRetries))
            {
              	CancelWaitableTimer(hGenericTimer);
               ClearGenericEvent();
               // Wait for file to stop playing
               if(audioState->IsPlaying() && ((status = AudioWaitForPlayStop()) != IVR_SUCCESS))
               {
                  return(status);
               }
               return(IVR_TIMEOUT);
            }
            // Reset timeout timer
            else
            {
               liTimeout.QuadPart = -(LONGLONG) (updatedTimeout < connectionTimeout ? updatedTimeout : connectionTimeout)*10000;
               SetWaitableTimer(hGenericTimer, &liTimeout, 0, 0, 0, false);
            }

            http.Close();
            attemptConnect = true;
            break;
         }

         default:
         {
            if((status != IVR_HANGUP) && (status != IVR_SHUTDOWN))
            {
               ivLog->Print("      ERROR: Unexpected problem waiting for connection event\n");
            }
           	CancelWaitableTimer(hGenericTimer);
            ClearGenericEvent();
            if(audioState->IsPlaying())
            {
               AudioStopPlay();
            }
            return(status);
         }
      }
   }

   ivLog->Print("      Sending request: '%s'\n", http.Request()->GetText());

   // Send the request
   while((ret = http.Send()) == WSAEWOULDBLOCK)
   {
      // Allow 10 ms for socket to clear
      Sleep(10);
   }

   if(ret != Socket::SUCCESS)
   {
      const char* errorMessage = Socket::WSAErrorString(ret);

      ivLog->Print("      ERROR: Failed to send request\n");
      ivLog->Print("         '%s'\n", errorMessage);
      CancelWaitableTimer(hGenericTimer);
      ClearGenericEvent();
      http.Response()->SetText(errorMessage);
      if(audioState->IsPlaying() && ((status = AudioWaitForPlayStop()) != IVR_SUCCESS))
      {
         return(status);
      }
      return(IVR_FAIL);
   }

   ivLog->Print("      Request successful, waiting for response\n");

   loopAgain = true;
   while(loopAgain)
   {
      status = WaitForEvent(&boardEvent);

      switch(status)
      {
         // Board event
         case IVR_SUCCESS:
         {
         	switch(boardEvent.MTEvent.EventCode)
            {
            	case EVT_EOF:
               case EVT_MAXBYTES:
               case EVT_MAXTIME:
               case EVT_STOP:
               case EVT_IO_COMPLETED:
               {
                	// Set timeout timer
                  if(timeout != FOREVER)
                  {
                     liTimeout.QuadPart = -(LONGLONG) timeout * 10000;
                     SetWaitableTimer(hGenericTimer, &liTimeout, 0, 0, 0, false);
                  }

                  break;
   		   	}

               case EVT_DIGIT:
               {
                  break;
               }

               default:
  	            {
                  // For debugging purposes
//                if(!CallerIDEvent(boardEvent, callingDigits))
//                {
//     	      	   ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//               	      "         User Status: %d\n         SubReason: %d\n"
//                      "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                      boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//                }
               }
            }

            break;
         }

         // Socket signalled
         case IVR_GENERIC_EVENT:
         {
            if((ret = http.Receive()) == Socket::SUCCESS)
            {
               loopAgain = false;
            }
            else if(ret == WSAEWOULDBLOCK)
            {
               // Allow 10 ms for socket to clear
               Sleep(10);
            }
            else
            {
               const char* errorMessage = Socket::WSAErrorString(ret);

               ivLog->Print("      ERROR: Failed to receive response\n");
               ivLog->Print("         '%s'\n", errorMessage);
              	CancelWaitableTimer(hGenericTimer);
               ClearGenericEvent();
               http.Response()->SetText(errorMessage);
               if(audioState->IsPlaying() && ((status = AudioWaitForPlayStop()) != IVR_SUCCESS))
               {
                  return(status);
               }
               return(IVR_FAIL);
            }

            break;
         }

         // Timed out
         case IVR_GENERIC_TIMEOUT:
         {
            ivLog->Print("      ERROR: Timed out waiting for response\n");
           	CancelWaitableTimer(hGenericTimer);
            ClearGenericEvent();
            return(IVR_TIMEOUT);
         }

         default:
         {
            if((status != IVR_HANGUP) && (status != IVR_SHUTDOWN))
            {
               ivLog->Print("      ERROR: Unexpected problem waiting for initial response event\n");
            }
           	CancelWaitableTimer(hGenericTimer);
            ClearGenericEvent();
            if(audioState->IsPlaying())
            {
               AudioStopPlay();
            }
            return(status);
         }
      }
   }

   // Check that response was OK (20x)
   // Bad return code: not 200 to 209
   if((http.Response()->GetStatusCode() < 200) || (http.Response()->GetStatusCode() > 209))
   {
      ivLog->Print("      ERROR: '%s'\n", http.Response()->GetReasonPhrase());
      CancelWaitableTimer(hGenericTimer);
      ClearGenericEvent();
      if(audioState->IsPlaying() && ((status = AudioWaitForPlayStop()) != IVR_SUCCESS))
      {
         return(status);
      }
      return(IVR_FAIL);
   }

   // Don't have whole message yet
   while(!http.Response()->HaveCompleteMessage())
   {
      status = WaitForEvent(&boardEvent);

      switch(status)
      {
         // Board event
         case IVR_SUCCESS:
         {
         	switch(boardEvent.MTEvent.EventCode)
            {
            	case EVT_EOF:
               case EVT_MAXBYTES:
               case EVT_MAXTIME:
               case EVT_STOP:
               case EVT_IO_COMPLETED:
               {
                	// Set timeout timer
                  if(timeout != FOREVER)
                  {
                     liTimeout.QuadPart = -(LONGLONG) timeout * 10000;
                     SetWaitableTimer(hGenericTimer, &liTimeout, 0, 0, 0, false);
                  }

                  break;
   		   	}

               case EVT_DIGIT:
               {
                  break;
               }

               default:
  	            {
                  // For debugging purposes
//                if(!CallerIDEvent(boardEvent, callingDigits))
//                {
//     	         	ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
//                     	"         User Status: %d\n         SubReason: %d\n"
//                      "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
//                      boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
//                }
               }
            }

            break;
         }

         // Socket signalled
         case IVR_GENERIC_EVENT:
         {
            if((ret = http.Receive()) == WSAEWOULDBLOCK)
            {
               // Allow 10 ms for socket to clear
               Sleep(10);
            }
            else if(ret != Socket::SUCCESS)
            {
               const char* errorMessage = Socket::WSAErrorString(ret);

               ivLog->Print("      ERROR: Failed to receive remainder of response\n");
               ivLog->Print("         '%s'\n", errorMessage);
              	CancelWaitableTimer(hGenericTimer);
               ClearGenericEvent();
               http.Response()->SetText(errorMessage);
               if(audioState->IsPlaying() && ((status = AudioWaitForPlayStop()) != IVR_SUCCESS))
               {
                  return(status);
               }
               return(IVR_FAIL);
            }

            break;
         }

         // Timeout
         case IVR_GENERIC_TIMEOUT:
         {
            ivLog->Print("      ERROR: Timed out waiting for remainder of response\n");
           	CancelWaitableTimer(hGenericTimer);
            ClearGenericEvent();
            return(IVR_TIMEOUT);
         }

         default:
         {
            if((status != IVR_HANGUP) && (status != IVR_SHUTDOWN))
            {
               ivLog->Print("      ERROR: Unexpected problem waiting for complete response event\n");
            }
           	CancelWaitableTimer(hGenericTimer);
            ClearGenericEvent();
            if(audioState->IsPlaying())
            {
               AudioStopPlay();
            }
            return(status);
         }
      }
   }

   ivLog->Print("      Response received\n");

  	CancelWaitableTimer(hGenericTimer);
   ClearGenericEvent();

   // Wait for file to stop playing
   if(audioState->IsPlaying() && ((status = AudioWaitForPlayStop()) != IVR_SUCCESS))
   {
      return(status);
   }

   return(IVR_SUCCESS);
}

/* -------------------------------------------------------------------------
   Name:        BoardMT::AudioRecordBuffer                          (public)

   Purpose:     Record to an audio buffer - TF Feb.4, 2004

   TO DO:       -  UNABLE TO MAKE StartRecordBuffer() WORK - TF Jun.10/04
                   -  ENCOUNTERED PROBLEMS WITH STREAMING
                   -  MAY NEED HELP FROM SB
                   -  OVERRIDE BRD_GENERIC FUNCTION WITH THIS ONE INSTEAD

   Notes:       -  Audio buffer must be created by the calling module.
                -  Maximum silence time is maxSilence or until buffer full,
                   which ever comes first.
                -  There is no maximum record time, as this is determined by
                   the size of the buffer.
   ------------------------------------------------------------------------- */
IV_STATUS BoardMT::AudioRecordBuffer(char* buffer, unsigned int bufsize,
	bool interruptable, SOUND_FORMAT format, int /*volLevel*/,
	int maxSilence, bool disableBeep, bool /*automaticGainControl*/)
{
   IV_STATUS      status;
	MTBoardEvent   boardEvent;
	MT_IO_CONTROL  ioCtrl;
   MTCHAN_STATUS  channelStatus;
   MT_RESULT      result;
   char				ttDig;      		// used to peek touch-tone digits.
   bool           loopAgain = true;
	int            samples = 0;
   // int         counter;
   unsigned long  channelStatusEvents;

   ivLog->Print("         AudioRecordBuffer, buffer: 0x%x, size: %i \n", buffer, bufsize);

	if ((buffer == NULL) || (bufsize < 1))
   {
      ivLog->Print("      ERROR: Invalid audio buffer\n");
   	return(IVR_FATAL);
   }

   if(audioState->IsRecording())
   {
      ivLog->Print("      Warning: Already recording, stopping previous record\n");
      status = AudioStopRecord();
      if (status != IVR_SUCCESS)
      {
      	return(status);
      }
   }

   // Only query for non-passive records
   if(lineOffHook)
   {
      // Beep has to occur before interrupt check in case it is interrupted
      if(!disableBeep)
      {
         status = AudioBeep(interruptable);

         // So that beep doesn't show up in recording
         Sleep(100);

         if(status != IVR_SUCCESS)
         {
            return(status);
         }
      }

	   if(interruptable)
      {
	      // Check for pending digit
		   QueryDigit(ttDig);

	   	// Check for interruption
		   if(ttDig != '\0')
	      {
	         // Skip record if interruptable and digit pending
	  	      ivLog->Print("      Interrupted!\n");
	       	return(IVR_SUCCESS);
         }
      }
   }

   if (format != TF_WAV_8KHZ_16BIT)		// (format == TF_BY_EXTENSION)
   {
      ivLog->Print("      ERROR: Invalid recording format, only TF_WAV_8KHZ_16BIT supported\n");
   	return(IVR_NOTSUPPORTED);
   }

	// Get enabled event information
	unsigned long statusSize = sizeof(MTCHAN_STATUS);
   if(MTGetChannelStatus(curChannel, &channelStatus, &statusSize) != MT_RET_OK)
   {
      ivLog->Print("      ERROR: Failed to get port status\n");
      PrintMTError();
   	return(IVR_FATAL);
   }

   channelStatusEvents = channelStatus.EventEnable;

	// Enable silence detection
   MTSetEventFilters(curChannel, (unsigned short) (channelStatus.EventEnable | SE_SIL), 1);

	// Clear parameter
	memset(&ioCtrl, 0, sizeof(MT_IO_CONTROL));

   // Set maximum silence time
   ioCtrl.MaxSilence = maxSilence;

   // Terminate on digit: if interruptable, any; otherwise none
   ioCtrl.EndDigit = interruptable ? '@' : '\0';

   // Automatic Gain Control is engaged by default

   // Set file/audio format
   ioCtrl.FileType = GetEncoding(TF_WAV_8KHZ_16BIT);

   // Deliberately shallow copy, use buffer created by calling module
	ioCtrl.Buffer = (unsigned char*) buffer;

   ioCtrl.Length = bufsize;

   // Check for streaming
   if(audioState->IsStreaming())
   {
      status = StopStreaming();

      if(status != IVR_SUCCESS)
      {
         return(status);
      }
   }

   audioState->RestartStreaming(true);

   // Record
   if(((result = MTRecBuffer(curChannel, &ioCtrl)) != MT_RET_OK) && (result != MT_RET_IO_PENDING))
   {
      ivLog->Print("      ERROR: Failed to record on port\n");
      PrintMTError();
   	return(IVR_FATAL);
   }

   boardEvent.MTEvent.EventCode = 0;

	// Wait for board to signal record completed
   while(loopAgain && ((status = WaitForEvent(&boardEvent)) == IVR_SUCCESS))
   {
      switch(boardEvent.MTEvent.EventCode)
      {
         case EVT_MAXBYTES:
         {
            samples = ioCtrl.Length / 2;
            loopAgain = false;
            break;
         }

			case EVT_MAX_SILENCE:
         case EVT_TERMSILENCE:
         {
         	// Get information from the event generated
            if(MTGetChannelStatus(curChannel, &channelStatus, &statusSize) != MT_RET_OK)
            {
               ivLog->Print("      ERROR: Failed to get port status\n");
               PrintMTError();
            	return(IVR_FATAL);
            }

            samples = channelStatus.Bytes / 2;
            loopAgain = false;
            break;
         }

         case EVT_IO_COMPLETED:
         {
            loopAgain = false;
         	break;
         }

         case EVT_DIGIT:
         {
            break;
         }

         default:
  	      {
            //	// For debugging purposes
            //	if(!CallerIDEvent(boardEvent, callingDigits))
            //	{
            //		ivLog->Print("      ERROR: Unexpected MT board event '%s'\n"
            //						 "         User Status: %d\n         SubReason: %d\n"
            //						 "         Extra Information: %d\n", GetEventName(boardEvent.MTEvent.EventCode),
            //		boardEvent.MTEvent.UserStatus, boardEvent.MTEvent.SubReason, boardEvent.MTEvent.XtraInfo);
            //	}
         }
      }
   }

   // Reset enabled events
   MTSetEventFilters(curChannel, (unsigned short) channelStatusEvents, 1);

   if(status != IVR_SUCCESS)
   {
   	return(status);
   }

	// QUESTION: Does this function really need to worry about samples?
   if(samples == 0)
   {
      ivLog->Print("      ERROR: Failed to acquire audio samples\n");
      return(IVR_FAIL);
   }

   return(status);
}
/*BoardMT::AudioRecordBuffer*/

IV_STATUS BoardMT::SetActivityMode(T_ACTIVITY_MODE requestedSetting)
{
   if (requestedSetting == activityMode)
   {
      return(IVR_SUCCESS);
   }
   else if(activityMode == ACTIVE_ONLY)
   {
      return(IVR_FAIL);
   }
   else
   {
      activityMode = requestedSetting;
      return(IVR_SUCCESS);
   }
}
//BoardMT::SetActivityMode

BoardMT::T_ACTIVITY_MODE BoardMT::CurrentActivityMode(void)
{
    return(activityMode);
}
//BoardMT::CurrentActivityMode
