#include <algorithm>
using std::transform;
#include <deque>
using std::deque;

#include <math.h>    // fabs

#include "iv_math2.h"
#include "..\escutil\StringSupport.h"
#include "..\escutil\iv_str.h"        // CharStr2HexStr, HexStr2CharStr
#include "..\escutil\ESCTime.h"       // ESCTime class
#include "..\BoardDLLs\brd_generic.h"
#include "..\TFEngine\TFEngine_communication.h"
#include "step_RunBASICScript.h"

Script::Script(string newScript)
{
   script = newScript;
   parseIndex = script.begin();
   currentTokenType = NONE;
   previousExpressionTokenType = NONE;
}

Script::~Script(void)
{
}

string Script::GetCurrentToken(void)
{
   return(currentToken);
}

Script::TOKEN_TYPE Script::GetCurrentTokenType(void)
{
   return(currentTokenType);
}

string Script::GetFirstToken(void)
{
   parseIndex = script.begin();
   return(GetNextToken());
}

string Script::GetNextToken(void)
{
   string::iterator nextTokenStart;

   // Update previousToken
   previousExpressionToken = currentToken;
   previousExpressionTokenType = currentTokenType;

   // Find beginning of next token (ignore white space)
   while(parseIndex != script.end() && ((*parseIndex == ' ') || (*parseIndex == '\t')))
   {
      parseIndex++;
   }

   // Check for end of script
   if(parseIndex == script.end())
   {
      currentToken = "";
      currentTokenType = NONE;
      return(currentToken);
   }

   // Find end of token
   nextTokenStart = parseIndex++;

   // Operators
   if(strchr("+-*/%<>=!^", *nextTokenStart) != NULL)
   {
      currentTokenType = OPERATOR;

      // ! is only used for !=
      if((*nextTokenStart == '!') && (*parseIndex != '='))
      {
         currentTokenType = INVALID;
      }
      // Two character symbols: !=, <=, <>, and >=
      else if((((*nextTokenStart == '!') || (*nextTokenStart == '>')) && (*parseIndex == '='))
         || ((*nextTokenStart == '<') && ((*parseIndex == '=') || (*parseIndex == '>'))))
      {
         parseIndex++;
      }
   }
   // Separators
   else if(strchr("@(,)\"\n:", *nextTokenStart) != NULL)
   {
      currentTokenType = SEPARATOR;
   }
   else if(strchr("'", *nextTokenStart) != NULL)
   {
      // Comments following
   }
   else
   {
      bool decimalFound = false;

      // Number
      if(isdigit(*nextTokenStart) || (*nextTokenStart == '-'))
      {
         currentTokenType = NUMBER;
      }
      // Invalid token
      else if(!isdigit(*nextTokenStart) && !isalpha(*nextTokenStart) && (*parseIndex != '_'))
      {
         currentTokenType = INVALID;
      }
      // Keyword or Label
      else
      {
         currentTokenType = NAME;
      }

      while(
         // Not end of script
         (parseIndex != script.end())
         // Not white space
         && (*parseIndex != ' ') && (*parseIndex != '\t')
         && (
            // Number
            ((currentTokenType == NUMBER) && (isdigit(*parseIndex) || ((*parseIndex == '.') && !decimalFound)))
            ||
            // Name
            (isalpha(*parseIndex) || isdigit(*parseIndex) || (*parseIndex == '_'))
            )
         )
      {
         if((*parseIndex == '.') && !decimalFound)
         {
            decimalFound = true;
         }
         parseIndex++;
      }
   }

   // Get token substring
   currentToken = script.substr(nextTokenStart - script.begin(), parseIndex - nextTokenStart);

   // Convert token to upper case
   transform(currentToken.begin(), currentToken.end(), currentToken.begin(), toupper);

   // Check for operators that are non-symbolic
   if((currentToken == "AND") || (currentToken == "BAND") || (currentToken == "OR")
      || (currentToken == "BOR") || (currentToken == "XOR") || (currentToken == "BXOR")
      || (currentToken == "NOT") || (currentToken == "BNOT") || (currentToken == "MOD"))
   {
      currentTokenType = OPERATOR;
   }

   // Check for functions that accept parameters
   if((currentToken == "ABS") || (currentToken == "ASC") || (currentToken == "CHAR")
      || (currentToken == "CHAR2HEX") || (currentToken == "CONCAT")
      || (currentToken == "DATEPART") || (currentToken == "HEX2CHAR")
      || (currentToken == "LEFT") || (currentToken == "LEN") || (currentToken == "LOWERCASE")
      || (currentToken == "POS") || (currentToken == "RIGHT") || (currentToken == "ROUND")
      || (currentToken == "SGN") || (currentToken == "SUBSTR") || (currentToken == "TIMEPART")
      || (currentToken == "TRIM") || (currentToken == "UPPERCASE"))
   {
      currentTokenType = FUNCTION;
   }

   if((currentToken == "REM") || (currentToken == "'"))
   {
      while(
         // Not end of script
         (parseIndex != script.end())
         // Not white space
         && (*parseIndex != '\n'))
      {
         parseIndex++;
      }
      return(GetNextToken());
   }
   
   return(currentToken);
}

string Script::GetPreviousExpressionToken(void)
{
   return(previousExpressionToken);
}

Script::TOKEN_TYPE Script::GetPreviousExpressionTokenType(void)
{
   return(previousExpressionTokenType);
}

// Returns a string with all text until the next quote, end of line, or end of script
string Script::GetUntilEndQuote(void)
{
   string quotedText;

   // Get text until quote
   while((parseIndex != script.end()) && (*parseIndex != '\"'))
   {
      quotedText += *parseIndex;
      parseIndex++;
   }

   return(quotedText);
}

string Script::StartNewExpression(void)
{
   string token = GetNextToken();
   previousExpressionToken.clear();
   previousExpressionTokenType = NONE;
   return(token);
}

Expression::Expression(int priority) : priority(priority)
{
   left = NULL;
   right = NULL;
}

Expression::~Expression(void)
{
   delete left;
   delete right;
}

Expression* Expression::Add(Expression* expression)
{
   // Don't process NULL expressions
   if(expression != NULL)
   {
      // A higher precedence expression should be added to a lower precedence
      // expression so that lower precedence expressions are always higher in the
      // expression tree than higher precedence expressions
      // This constructs a preorder parse tree
      if(expression->GetPriority() < GetPriority())
      {
         return(expression->Add(this));
      }
      // Fill from left to right
      else if(left == NULL)
      {
         left = expression;
      }
      else if(right == NULL)
      {
         right = expression;
      }
      // Expand to the right
      else
      {
         right = right->Add(expression);
      }
   }

   return(this);
}

void Expression::Dump(Log* ivLog)
{
   char number[14];
   string childExpression;
   string thisExpression = Name();
   thisExpression += " (";
   sprintf(number, "%X", this);
   thisExpression += number;
   thisExpression += ")";

   ivLog->Print("%s\n", thisExpression.c_str());

   if(left == NULL)
   {
      childExpression = "(NULL)";
   }
   else
   {
      childExpression = left->Name() + " (";
      sprintf(number, "%X", left);
      childExpression += number;
      childExpression += ")";
   }

   ivLog->Print("%s\n", (thisExpression + " Left: " + childExpression).c_str());

   if(left != NULL)
   {
      left->Dump(ivLog);
   }

   if(right == NULL)
   {
      childExpression = "(NULL)";
   }
   else
   {
      childExpression = right->Name() + " (";
      sprintf(number, "%X", right);
      childExpression += number;
      childExpression += ")";
   }

   ivLog->Print("%s\n", (thisExpression + " Right: " + childExpression).c_str());

   if(right != NULL)
   {
      right->Dump(ivLog);
   }
}

string Expression::EvaluateCompare(PIVRINFO pInfo, CMP_FN compareFunction, bool& invalid)
{
   string operand1;
   string operand2;
   IV_STATUS status;
   bool result;
   string returnValue;

   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   operand1 = left->Evaluate(pInfo, invalid);
   operand2 = right->Evaluate(pInfo, invalid);

   status = VarCmp(pInfo, operand1.c_str(), operand2.c_str(), compareFunction, result, true);

   invalid |= status != IVR_SUCCESS;

   return(result ? "1" : "0");
}

string Expression::EvaluateMath(PIVRINFO pInfo, MATH_FN mathFunction, bool& invalid)
{
   string operand1;
   string operand2;
   IV_STATUS status;
   char* value;
   string returnValue;

   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   operand1 = left->Evaluate(pInfo, invalid);
   operand2 = right->Evaluate(pInfo, invalid);

   status = VarMath(pInfo, value, operand1.c_str(), operand2.c_str(), NULL, mathFunction);

   invalid |= status != IVR_SUCCESS;
   returnValue = value == NULL ? "" : value;
   delete[] value;

   return(returnValue);
}

int Expression::GetPriority(void)
{
   return(priority);
}

string Expression::GetString(void) const
{
   return((left != NULL ? left->GetString() : (string) "") + (string) " " + Symbol()
       + (string) " " + (right != NULL ? right->GetString() : (string) ""));
}

string Expression::Symbol(void) const
{
   return("");
}

ExpressionTreeStack::ExpressionTreeStack(void)
{
}

ExpressionTreeStack::~ExpressionTreeStack(void)
{
   // Make sure any expression trees left on the stack get deleted
   for(deque<Expression*>::iterator tree = begin(); tree != end(); tree++)
   {
      delete *tree;
   }
}

Function::Function(unsigned int paramsCount) : Expression(13), expectedParamsCount(paramsCount)
{
}

Function::~Function(void)
{
   for(unsigned int curParam = 0; curParam < params.size(); curParam++)
   {
      delete params[curParam];
   }
}

void Function::AddParameter(Expression* expression)
{
   params.push_back(expression);
}

unsigned int Function::GetExpectedParamsCount(void)
{
   return(expectedParamsCount);
}

bool Function::EvaluateParameters(PIVRINFO pInfo, bool& invalid)
{
   unsigned int paramsSize;

   evaluatedParams.clear();
   invalid = expectedParamsCount == 0;
   paramsSize = params.size();

   for(unsigned int index = 0; (index < paramsSize) && !invalid; index++)
   {
      string valueStr = params[index]->Evaluate(pInfo, invalid);

      if(!invalid)
      {
         evaluatedParams.push_back(valueStr);
      }
   }

   invalid |= (evaluatedParams.size() != expectedParamsCount);

   return(!invalid);
}

string Function::GetString(void) const
{
   string functionString = Symbol() + (string) "(";

   for(unsigned int index = 0; index < expectedParamsCount; index++)
   {
      functionString += ((index < params.size()) && (params[index] != NULL) ? params[index]->GetString() : (string) "");
      functionString += index < expectedParamsCount - 1 ? (string) ", " : (string) ")";
   }

   return(functionString);
}

UnaryOperator::UnaryOperator(void) : Expression(12)
{
}

UnaryOperator::~UnaryOperator(void)
{
}

Expression* UnaryOperator::Add(Expression* expression)
{
   // Don't process NULL expressions
   if(expression != NULL)
   {
      // A higher precedence expression should be added to a lower precedence
      // expression so that lower precedence expressions are always higher in the
      // expression tree than higher precedence expressions
      // This constructs a preorder parse tree
      if(expression->GetPriority() < GetPriority())
      {
         return(expression->Add(this));
      }
      // Fill to right only
      else if(right == NULL)
      {
         right = expression;
      }
      // Expand to the right
      else
      {
         right = right->Add(expression);
      }
   }

   return(this);
}

string UnaryOperator::GetString(void) const
{
   return(Symbol() + (right != NULL ? right->GetString() : (string) ""));
}

AbsoluteValue::AbsoluteValue(void) : Function(1)
{
}

AbsoluteValue::~AbsoluteValue(void)
{
}

string AbsoluteValue::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   double value;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   value = fabs(atof(evaluatedParams[0].c_str()));

   return(FormattedString("%f", value));
}

string AbsoluteValue::Symbol(void) const
{
   return("ABS");
}

string AbsoluteValue::Name(void)
{
   return("Absolute Value");
}

Add::Add(void) : Expression(9)
{
}

Add::~Add(void)
{
}

string Add::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateMath(pInfo, FN_ADD, invalid));
}

string Add::Symbol(void) const
{
   return("+");
}

string Add::Name(void)
{
   return("Add");
}

BitwiseAnd::BitwiseAnd(void) : Expression(6)
{
}

BitwiseAnd::~BitwiseAnd(void)
{
}

string BitwiseAnd::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char result[20];

   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   sprintf(result, "%d", atoi(left->Evaluate(pInfo, invalid).c_str()) & atoi(right->Evaluate(pInfo, invalid).c_str()));

   return((string) result);
}

string BitwiseAnd::Symbol(void) const
{
   return("BAND");
}

string BitwiseAnd::Name(void)
{
   return("Bitwise And");
}

LogicalAnd::LogicalAnd(void) : Expression(3)
{
}

LogicalAnd::~LogicalAnd(void)
{
}

string LogicalAnd::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   return(atoi(left->Evaluate(pInfo, invalid).c_str()) && atoi(right->Evaluate(pInfo, invalid).c_str()) ? "1" : "0");
}

string LogicalAnd::Symbol(void) const
{
   return("AND");
}

string LogicalAnd::Name(void)
{
   return("Logical And");
}

Asc::Asc(void) : Function(1)
{
}

Asc::~Asc(void)
{
}

string Asc::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   if(evaluatedParams[0].empty())
   {
      invalid = true;
      return("");
   }

   return(FormattedString("%d", evaluatedParams[0][0]));
}

string Asc::Symbol(void) const
{
   return("ASC");
}

string Asc::Name(void)
{
   return("Character");
}

Char::Char(void) : Function(1)
{
}

Char::~Char(void)
{
}

string Char::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char value;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   if((value = (char) atoi(evaluatedParams[0].c_str())) == '\0')
   {
      invalid = true;
      return("");
   }

   return(FormattedString("%c", value));
}

string Char::Symbol(void) const
{
   return("CHAR");
}

string Char::Name(void)
{
   return("Character");
}

Char2Hex::Char2Hex(void) : Function(1)
{
}

Char2Hex::~Char2Hex(void)
{
}

string Char2Hex::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char* hexChars;
   int parameterLength;
   string hexString;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   if(evaluatedParams[0].empty())
   {
      return("");
   }

   parameterLength = evaluatedParams[0].length();
   hexChars = new char[2*parameterLength + 1];
   CharStr2HexStr((const unsigned char*) evaluatedParams[0].c_str(), hexChars, parameterLength, true);
   hexChars[2*parameterLength] = '\0';
   hexString = hexChars;
   delete[] hexChars;

   return(hexString);
}

string Char2Hex::Symbol(void) const
{
   return("CHAR2HEX");
}

string Char2Hex::Name(void)
{
   return("Character to Hexidecimal");
}

Concat::Concat(void) : Function(2)
{
}

Concat::~Concat(void)
{
}

string Concat::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(EvaluateParameters(pInfo, invalid) ? evaluatedParams[0] + evaluatedParams[1] : (string) "");
}

string Concat::Symbol(void) const
{
   return("CONCAT");
}

string Concat::Name(void)
{
   return("Concatenate");
}

Constant::Constant(string value) : Expression(13), value(value)
{
}

Constant::~Constant(void)
{
}

string Constant::Evaluate(PIVRINFO, bool&)
{
   return(value);
}

string Constant::GetString(void) const
{
   return(value);
}

string Constant::Name(void)
{
   return("Constant");
}

Divide::Divide(void) : Expression(10)
{
}

Divide::~Divide(void)
{
}

string Divide::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateMath(pInfo, FN_DIV, invalid));
}

string Divide::Symbol(void) const
{
   return("/");
}

string Divide::Name(void)
{
   return("Divide");
}

EqualTo::EqualTo(void) : Expression(7)
{
}

EqualTo::~EqualTo(void)
{
}

string EqualTo::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateCompare(pInfo, FN_EQ, invalid));
}

string EqualTo::Symbol(void) const
{
   return("=");
}

string EqualTo::Name(void)
{
   return("EqualTo");
}

Exponentiate::Exponentiate(void) : Expression(11)
{
}

Exponentiate::~Exponentiate(void)
{
}

string Exponentiate::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateMath(pInfo, FN_POW, invalid));
}

string Exponentiate::Symbol(void) const
{
   return("^");
}

string Exponentiate::Name(void)
{
   return("Exponentiate");
}

GreaterThan::GreaterThan(void) : Expression(8)
{
}

GreaterThan::~GreaterThan(void)
{
}

string GreaterThan::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateCompare(pInfo, FN_GT, invalid));
}

string GreaterThan::Symbol(void) const
{
   return(">");
}

string GreaterThan::Name(void)
{
   return("GreaterThan");
}

GreaterThanOrEqualTo::GreaterThanOrEqualTo(void) : Expression(8)
{
}

GreaterThanOrEqualTo::~GreaterThanOrEqualTo(void)
{
}

string GreaterThanOrEqualTo::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateCompare(pInfo, FN_GE, invalid));
}

string GreaterThanOrEqualTo::Symbol(void) const
{
   return(">=");
}

string GreaterThanOrEqualTo::Name(void)
{
   return("GreaterThanOrEqualTo");
}

Hex2Char::Hex2Char(void) : Function(1)
{
}

Hex2Char::~Hex2Char(void)
{
}

string Hex2Char::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char* chars;
   int resultLength;
   string charString;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   if(evaluatedParams[0].empty())
   {
      return("");
   }

   if(evaluatedParams[0].length() % 2 == 1)
   {
      invalid = true;
      return("");
   }

   resultLength = evaluatedParams[0].length()/2;
   chars = new char[resultLength + 1];
   HexStr2CharStr(evaluatedParams[0].c_str(), (unsigned char*) chars, resultLength);
   chars[resultLength] = '\0';
   charString = chars;
   delete[] chars;

   return(charString);
}

string Hex2Char::Symbol(void) const
{
   return("HEX2CHAR");
}

string Hex2Char::Name(void)
{
   return("Hexidecimal to Character");
}

Left::Left(void) : Function(2)
{
}

Left::~Left(void)
{
}

string Left::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(EvaluateParameters(pInfo, invalid) ? evaluatedParams[0].substr(0, atoi(evaluatedParams[1].c_str())) : (string) "");
}

string Left::Symbol(void) const
{
   return("LEFT");
}

string Left::Name(void)
{
   return("Left");
}

Len::Len(void) : Function(1)
{
}

Len::~Len(void)
{
}

string Len::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(EvaluateParameters(pInfo, invalid) ? FormattedString("%d", evaluatedParams[0].length()) : "");
}

string Len::Symbol(void) const
{
   return("LEN");
}

string Len::Name(void)
{
   return("Length");
}

LessThan::LessThan(void) : Expression(8)
{
}

LessThan::~LessThan(void)
{
}

string LessThan::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateCompare(pInfo, FN_LT, invalid));
}

string LessThan::Symbol(void) const
{
   return("<");
}

string LessThan::Name(void)
{
   return("LessThan");
}

LessThanOrEqualTo::LessThanOrEqualTo(void) : Expression(8)
{
}

LessThanOrEqualTo::~LessThanOrEqualTo(void)
{
}

string LessThanOrEqualTo::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateCompare(pInfo, FN_LE, invalid));
}

string LessThanOrEqualTo::Symbol(void) const
{
   return("<=");
}

string LessThanOrEqualTo::Name(void)
{
   return("LessThanOrEqualTo");
}

Lowercase::Lowercase(void) : Function(1)
{
}

Lowercase::~Lowercase(void)
{
}

string Lowercase::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   transform(evaluatedParams[0].begin(), evaluatedParams[0].end(), evaluatedParams[0].begin(), tolower);

   return(evaluatedParams[0]);
}

string Lowercase::Symbol(void) const
{
   return("LOWERCASE");
}

string Lowercase::Name(void)
{
   return("Lower Case");
}

Mod::Mod(void) : Expression(10)
{
}

Mod::~Mod(void)
{
}

string Mod::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateMath(pInfo, FN_MOD, invalid));
}

string Mod::Symbol(void) const
{
   return("MOD");
}

string Mod::Name(void)
{
   return("Mod");
}

Multiply::Multiply(void) : Expression(10)
{
}

Multiply::~Multiply(void)
{
}

string Multiply::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateMath(pInfo, FN_MULT, invalid));
}

string Multiply::Symbol(void) const
{
   return("*");
}

string Multiply::Name(void)
{
   return("Multiply");
}

LogicalNot::LogicalNot(void) : UnaryOperator()
{
}

LogicalNot::~LogicalNot(void)
{
}

string LogicalNot::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   if(right == NULL)
   {
      invalid = true;
      return("");
   }

   return(atoi(right->Evaluate(pInfo, invalid).c_str()) == 0 ? "1" : "0");
}

string LogicalNot::Symbol(void) const
{
   return("NOT");
}

string LogicalNot::Name(void)
{
   return("Logical Not");
}

BitwiseNot::BitwiseNot(void) : UnaryOperator()
{
}

BitwiseNot::~BitwiseNot(void)
{
}

string BitwiseNot::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char result[20];

   if(right == NULL)
   {
      invalid = true;
      return("");
   }

   sprintf(result, "%d", ~atoi(right->Evaluate(pInfo, invalid).c_str()));

   return((string) result);
}

string BitwiseNot::Symbol(void) const
{
   return("BNOT");
}

string BitwiseNot::Name(void)
{
   return("Bitwise Not");
}

Negate::Negate(void) : UnaryOperator()
{
}

Negate::~Negate(void)
{
}

string Negate::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char result[20];

   if(right == NULL)
   {
      invalid = true;
      return("");
   }

   sprintf(result, "%d", -atoi(right->Evaluate(pInfo, invalid).c_str()));

   return((string) result);
}

string Negate::Symbol(void) const
{
   return("-");
}

string Negate::Name(void)
{
   return("Negate");
}

NotEqualTo::NotEqualTo(void) : Expression(7)
{
}

NotEqualTo::~NotEqualTo(void)
{
}

string NotEqualTo::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateCompare(pInfo, FN_NE, invalid));
}

string NotEqualTo::Symbol(void) const
{
   return("!=");
}

string NotEqualTo::Name(void)
{
   return("NotEqualTo");
}

BitwiseExclusiveOr::BitwiseExclusiveOr(void) : Expression(5)
{
}

BitwiseExclusiveOr::~BitwiseExclusiveOr(void)
{
}

string BitwiseExclusiveOr::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char result[20];

   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   sprintf(result, "%d", atoi(left->Evaluate(pInfo, invalid).c_str()) ^ atoi(right->Evaluate(pInfo, invalid).c_str()));

   return((string) result);
}

string BitwiseExclusiveOr::Symbol(void) const
{
   return("BXOR");
}

string BitwiseExclusiveOr::Name(void)
{
   return("Bitwise Exclusive Or");
}

LogicalExclusiveOr::LogicalExclusiveOr(void) : Expression(2)
{
}

LogicalExclusiveOr::~LogicalExclusiveOr(void)
{
}

string LogicalExclusiveOr::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   int leftValue;
   int rightValue;

   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   leftValue = atoi(left->Evaluate(pInfo, invalid).c_str());
   rightValue = atoi(right->Evaluate(pInfo, invalid).c_str());

   return((leftValue && !rightValue) || (!leftValue && rightValue) ? "1" : "0");
}

string LogicalExclusiveOr::Symbol(void) const
{
   return("XOR");
}

string LogicalExclusiveOr::Name(void)
{
   return("Logical Exclusive Or");
}

BitwiseOr::BitwiseOr(void) : Expression(4)
{
}

BitwiseOr::~BitwiseOr(void)
{
}

string BitwiseOr::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   char result[20];

   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   sprintf(result, "%d", atoi(left->Evaluate(pInfo, invalid).c_str()) | atoi(right->Evaluate(pInfo, invalid).c_str()));

   return((string) result);
}

string BitwiseOr::Symbol(void) const
{
   return("BOR");
}

string BitwiseOr::Name(void)
{
   return("Bitwise Or");
}

LogicalOr::LogicalOr(void) : Expression(1)
{
}

LogicalOr::~LogicalOr(void)
{
}

string LogicalOr::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   if((left == NULL) || (right == NULL))
   {
      invalid = true;
      return("");
   }

   return(atoi(left->Evaluate(pInfo, invalid).c_str()) || atoi(right->Evaluate(pInfo, invalid).c_str()) ? "1" : "0");
}

string LogicalOr::Symbol(void) const
{
   return("OR");
}

string LogicalOr::Name(void)
{
   return("Logical Or");
}

Parenthesised::Parenthesised(Expression* expression) : Expression(14)
{
   left = expression;
}

Parenthesised::~Parenthesised(void)
{
}

string Parenthesised::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   if(left == NULL)
   {
      invalid = true;
      return("");
   }

   return(left->Evaluate(pInfo, invalid));
}

string Parenthesised::GetString(void) const
{
   return((string) "(" + (left != NULL ? left->GetString() : (string) "") + (string) ")");
}

string Parenthesised::Name(void)
{
   return("Parenthesised");
}

Pos::Pos(void) : Function(2)
{
}

Pos::~Pos(void)
{
}

string Pos::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(EvaluateParameters(pInfo, invalid) ? FormattedString("%d", evaluatedParams[0].find(evaluatedParams[1])) : "");
}

string Pos::Symbol(void) const
{
   return("POS");
}

string Pos::Name(void)
{
   return("Position");
}

Right::Right(void) : Function(2)
{
}

Right::~Right(void)
{
}

string Right::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   int numchars;
   int totalchars;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }
   totalchars = evaluatedParams[0].length();
   numchars = atoi(evaluatedParams[1].c_str());

   if((totalchars - numchars) < 0)
   {
      // Fewer characters than requested, return the whole string
      return(evaluatedParams[0].c_str());
   }

   return(evaluatedParams[0].substr(totalchars - numchars, numchars));
}

string Right::Symbol(void) const
{
   return("RIGHT");
}

string Right::Name(void)
{
   return("Right");
}

Round::Round(void) : Function(1)
{
}

Round::~Round(void)
{
}

string Round::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   float value;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   value = atof(evaluatedParams[0].c_str());

   // Round up 1/2 way between positive integers, and round down 1/2 way between negative integers
   return(FormattedString("%d", value >= 0 ? (int) (value + 0.5) : (int) (value - 0.5)));
}

string Round::Symbol(void) const
{
   return("ROUND");
}

string Round::Name(void)
{
   return("Round");
}

Sign::Sign(void) : Function(1)
{
}

Sign::~Sign(void)
{
}

string Sign::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   float value;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   value = atof(evaluatedParams[0].c_str());

   if(value > 0)
   {
      return("1");
   }
   else if(value == 0)
   {
      return("0");
   }
   else
   {
      return("-1");
   }
}

string Sign::Symbol(void) const
{
   return("SIGN");
}

string Sign::Name(void)
{
   return("Sign");
}

SubStr::SubStr(void) : Function(3)
{
}

SubStr::~SubStr(void)
{
}

string SubStr::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   int startpos;
   int numchars;
   int totalchars;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   totalchars = evaluatedParams[0].length();
   startpos = atoi(evaluatedParams[1].c_str());
   numchars = atoi(evaluatedParams[2].c_str());

   pInfo->ivLog->Print("      WARNING: TeleFlow version 2007.1123 and later, changed SUBSTR to 1-based string position\n");
   // @@@ Should remove the above line after a year or so

   if(startpos < 1)
   {
      // Versions older than 2009.03.03 would have given a warning and returned empty string
      pInfo->ivLog->Print("      ERROR: Invalid start position (%i) in SUBSTR function\n", startpos);
      invalid = true;
      return("");
   }

   if(startpos > totalchars)
   {
      // Position past end-of-string, return empty string
      return("");
   }

   // adjust 1-based TeleFlow counting to 0-based offset
   startpos = startpos - 1;

   return(evaluatedParams[0].substr(startpos, numchars));
}

string SubStr::Symbol(void) const
{
   return("SUBSTR");
}

string SubStr::Name(void)
{
   return("Substring");
}

Subtract::Subtract(void) : Expression(9)
{
}

Subtract::~Subtract(void)
{
}

string Subtract::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   return(Expression::EvaluateMath(pInfo, FN_SUB, invalid));
}

string Subtract::Symbol(void) const
{
   return("-");
}

string Subtract::Name(void)
{
   return("Subtract");
}

Trim::Trim(void) : Function(1)
{
}

Trim::~Trim(void)
{
}

string Trim::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   string::size_type beginIndex;
   string::size_type endIndex;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   if((beginIndex = evaluatedParams[0].find_first_not_of(' ')) == string::npos)
   {
      return("");
   }

   if((endIndex = evaluatedParams[0].find_last_not_of(' ')) == string::npos)
   {
      return(evaluatedParams[0].substr(beginIndex));
   }

   return(evaluatedParams[0].substr(beginIndex, endIndex - beginIndex + 1));
}

string Trim::Symbol(void) const
{
   return("TRIM");
}

string Trim::Name(void)
{
   return("Trim");
}

Uppercase::Uppercase(void) : Function(1)
{
}

Uppercase::~Uppercase(void)
{
}

string Uppercase::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   transform(evaluatedParams[0].begin(), evaluatedParams[0].end(), evaluatedParams[0].begin(), toupper);

   return(evaluatedParams[0]);
}

string Uppercase::Symbol(void) const
{
   return("UPPERCASE");
}

string Uppercase::Name(void)
{
   return("Upper Case");
}

DatePart::DatePart(void) : Function(1)
{
}

DatePart::~DatePart(void)
{
}

string DatePart::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   string datetime;
   string result;
   char sepChar[2];

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   datetime = evaluatedParams[0];
   if(DetermineType(datetime.c_str()) != VAR_DATETIME)
   {
      pInfo->ivLog->Print("      ERROR: Invalid Date/Time value, '%s'\n", datetime.c_str());
      invalid = true;
      return("");
   }

   sepChar[0] = pInfo->variables->Get("@DATE_SEPARATOR")[0]; // only the first character
   if(sepChar[0] != '-')      // only '-' and '/' are valid
      sepChar[0] = '/';       // default is '/'
   sepChar[1] = '\0';

   pInfo->ivLog->Print("DatePart of datetime '%s'\n", datetime.c_str());
   result = datetime.substr(0,10);
   pInfo->ivLog->Print("      result string '%s'\n", result.c_str());
   pInfo->ivLog->Print("      set sep char to '%s'\n", sepChar);
   result.replace(4, 1, sepChar, 1);
   result.replace(7, 1, sepChar, 1);
   pInfo->ivLog->Print("      new result string '%s'\n", result.c_str());

   return(result);
}

string DatePart::Symbol(void) const
{
   return("DATEPART");
}

string DatePart::Name(void)
{
   return("DatePart");
}

TimePart::TimePart(void) : Function(1)
{
}

TimePart::~TimePart(void)
{
}

string TimePart::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   string datetime;
   string result;

   if(!EvaluateParameters(pInfo, invalid))
   {
      return("");
   }

   datetime = evaluatedParams[0];
   if(DetermineType(datetime.c_str()) != VAR_DATETIME)
   {
      pInfo->ivLog->Print("      ERROR: Invalid Date/Time value, '%s'\n", datetime.c_str());
      invalid = true;
      return("");
   }

   pInfo->ivLog->Print("TimePart of datetime string '%s'\n", datetime.c_str());
   result = datetime.substr(11,8);
   pInfo->ivLog->Print("      result string '%s'\n", result.c_str());
   return(result);
}

string TimePart::Symbol(void) const
{
   return("TIMEPART");
}

string TimePart::Name(void)
{
   return("TimePart");
}

Variable::Variable(string variableName) : Expression(13), variableName(variableName)
{
}

Variable::~Variable(void)
{
}

string Variable::Evaluate(PIVRINFO pInfo, bool& invalid)
{
   const char* value = pInfo->variables->Get(variableName.c_str());
   // If value is NULL than an error occurred
   invalid = value == NULL;
   return(value == NULL ? "" : value);
}

string Variable::GetString(void) const
{
   return(variableName);
}

string Variable::Name(void)
{
   return("Variable");
}

Command::Command(void)
{
   next = NULL;
}

Command::~Command(void)
{
}

void Command::Dump(Log* ivLog)
{
   char number[14];
   string childExpression;
   string thisExpression = Name();
   thisExpression += " (";
   sprintf(number, "%X", this);
   thisExpression += number;
   thisExpression += ")";

   if(next == NULL)
   {
      childExpression = "(NULL)";
   }
   else
   {
      childExpression = next->Name() + " (";
      sprintf(number, "%X", next);
      childExpression += number;
      childExpression += ")";
   }

   ivLog->Print("%s\n", (thisExpression + " Next: " + childExpression).c_str());

   if(next != NULL)
   {
      next->Dump(ivLog);
   }
}

Command* Command::GetNextCommand(void)
{
   return(next);
}

void Command::SetNextCommand(Command* nextCommand)
{
   next = nextCommand;
}

Branch::Branch(BRANCH_TYPE branchType)
{
   elseBranch = NULL;
   ifBranch = NULL;
   this->branchType = branchType;
}

Branch::~Branch(void)
{
}

Branch::BRANCH_TYPE Branch::GetBranchType(void)
{
   return(branchType);
}

Command* Branch::GetElseBranch(void)
{
  return(elseBranch);
}

Command* Branch::GetIfBranch(void)
{
   return ifBranch;
}

void Branch::SetElseBranch(Command* command)
{
   elseBranch = command;
}

void Branch::SetNextCommand(Command* command)
{
  ifBranch = command;
}

Assignment::Assignment(string variableName, Expression* expressionTree)
   : variableName(variableName), expressionTree(expressionTree)
{
}

Assignment::~Assignment(void)
{
   delete expressionTree;
}

bool Assignment::Execute(PIVRINFO pInfo)
{
   bool invalid = false;
   string assignmentValue = expressionTree->Evaluate(pInfo, invalid);

   pInfo->ivLog->Print("      %s = '%s'\n", variableName.c_str(), expressionTree->GetString().c_str());

   // Check for evaluation error
   if(invalid)
   {
      return(false);
   }

   if(!pInfo->variables->Set(variableName.c_str(), assignmentValue.c_str()))
   {
      // Error setting variable
      return(false);
   }

   // Assignments may be the last command in the execution path
   // therefore, lack of a next step is acceptable
   return(true);
}

string Assignment::Name(void)
{
   return("Assignment");
}

ScopedAssignment::ScopedAssignment(string variableName, Expression* expressionTree, bool global)
   : Assignment(variableName, expressionTree), global(global)
{
}

ScopedAssignment::~ScopedAssignment(void)
{
}

bool ScopedAssignment::Execute(PIVRINFO pInfo)
{
   bool invalid = false;
   string assignmentValue = expressionTree->Evaluate(pInfo, invalid);

   pInfo->ivLog->Print("      %s %s = '%s'\n", global ? "GLOBAL" : "LOCAL", variableName.c_str(), expressionTree->GetString().c_str());

   // Check for evaluation error
   if(invalid)
   {
      return(false);
   }

   if(global)
   {
      if(!pInfo->variables->SetGlobal(variableName.c_str(), assignmentValue.c_str()))
      {
         // Error setting variable
         return(false);
      }
   }
   else if(!pInfo->variables->SetLocal(variableName.c_str(), assignmentValue.c_str()))
   {
      // Error setting variable
      return(false);
   }

   // Assignments may be the last command in the execution path
   // therefore, lack of a next step is acceptable
   return(true);
}

Break::Break(void)
{
}

Break::~Break(void)
{
}

bool Break::Execute(PIVRINFO)
{
   return(true);
}

string Break::Name(void)
{
   return("Break");
}

Case::Case(Expression* argument, Expression* condition, Case* linkedCase)
   : Branch(CASE), argument(argument), lowerBoundCondition(condition),
   upperBoundCondition(NULL), previousCaseLinked(linkedCase)
{
}

Case::~Case(void)
{
   delete lowerBoundCondition;
   delete upperBoundCondition;
}

void Case::SetNextCommand(Command* nextCommand)
{
   Branch::SetNextCommand(nextCommand);
   // If there is a Case command that is linked to this one (meaning that the
   // case was on a script line like "CASE 1, 2, 4, 5 TO 8", then all linked
   // case commands need to have their nextCommand updated.
   if(previousCaseLinked != NULL )
   {
      previousCaseLinked->SetNextCommand(nextCommand);
   }
}

bool Case::Execute(PIVRINFO pInfo)
{
   bool argumentInvalid = false;
   bool lowerConditionInvalid = false;
   bool upperConditionInvalid = false;
   string argumentValue;
   string lowerConditionValue;
   string upperConditionValue;
   bool result;

   // Select and CaseElse variants
   if(lowerBoundCondition == NULL)
   {
      next = ifBranch;
      return(next != NULL);
   }

   if(argument == NULL)
   {
      return(false);
   }

   argumentValue = argument->Evaluate(pInfo, argumentInvalid);

   if(argumentInvalid)
   {
      pInfo->ivLog->Print("      Case argument '%s' is invalid", argument->GetString().c_str());
   }
   else
   {
      pInfo->ivLog->Print("      Case argument '%s' evaluates to '%s'", argument->GetString().c_str(), argumentValue.c_str());
   }

   lowerConditionValue = lowerBoundCondition->Evaluate(pInfo, lowerConditionInvalid);

   // The Case is a simple one like CASE 2
   if(upperBoundCondition == NULL)
   {
      if(lowerConditionInvalid)
      {
         pInfo->ivLog->Print(" and condition '%s' is invalid\n", lowerBoundCondition->GetString().c_str());
      }
      else
      {
         pInfo->ivLog->Print(" and condition '%s' evaluates to '%s'\n", lowerBoundCondition->GetString().c_str(), lowerConditionValue.c_str());
      }

      if(argumentInvalid || lowerConditionInvalid)
      {
         return(false);
      }

      // Check for evaluation error
      if(VarCmp(pInfo, argumentValue.c_str(), lowerConditionValue.c_str(), FN_EQ, result, true) != IVR_SUCCESS)
      {
         return(false);
      }

   }
   // The Case is a range like CASE 2 TO 5
   else
   {
      upperConditionValue = upperBoundCondition->Evaluate(pInfo, upperConditionInvalid);

      if(lowerConditionInvalid)
      {
         pInfo->ivLog->Print(", lower condition '%s' is invalid", lowerBoundCondition->GetString().c_str());
      }
      else
      {
         pInfo->ivLog->Print(", lower condition '%s' evaluates to '%s'", lowerBoundCondition->GetString().c_str(), lowerConditionValue.c_str());
      }

      if(upperConditionInvalid)
      {
         pInfo->ivLog->Print(", and upper condition '%s' is invalid\n", upperBoundCondition->GetString().c_str());
      }
      else
      {
         pInfo->ivLog->Print(", and upper condition '%s' evaluates to '%s'\n", upperBoundCondition->GetString().c_str(), upperConditionValue.c_str());
      }

      if(argumentInvalid || lowerConditionInvalid || upperConditionInvalid)
      {
         return(false);
      }

      // Check for evaluation error
      if(VarCmp(pInfo, argumentValue.c_str(), lowerConditionValue.c_str(), FN_GE, result, true) != IVR_SUCCESS)
      {
         return(false);
      }

      if(result)
      {
         // Check for evaluation error
         if(VarCmp(pInfo, argumentValue.c_str(), upperConditionValue.c_str(), FN_LE, result, true) != IVR_SUCCESS)
         {
            return(false);
         }
      }
   }

   next = result ? ifBranch : elseBranch;

   return(next != NULL);
}

Expression* Case::GetArgument(void)
{
   return(argument);
}

string Case::Name(void)
{
   return("Case");
}

void Case::SetUpperBoundCondition(Expression* condition)
{
   upperBoundCondition = condition;
}

CaseElse::CaseElse(void) : Case(NULL, NULL)
{
}

CaseElse::~CaseElse(void)
{
}

string CaseElse::Name(void)
{
   return("Case Else");
}

Do::Do(void) : Branch(DO)
{
}

Do::~Do(void)
{
}

bool Do::Execute(PIVRINFO)
{
   next = ifBranch;
   return(next != NULL);
}

string Do::Name(void)
{
   return("Do");
}

Else::Else(void) : Branch(IF)
{
}

Else::~Else(void)
{
}

bool Else::Execute(PIVRINFO)
{
   next = ifBranch;
   return(next != NULL);
}

string Else::Name(void)
{
   return("Else");
}

End::End(void)
{
}

End::~End(void)
{
}

bool End::Execute(PIVRINFO)
{
   // Ends may be the last command in the execution path
   // therefore, lack of a next step is acceptable
   return(true);
}

string End::Name(void)
{
   return("End");
}

For::For(string variableName, Expression* start, Expression* end, Expression* step)
   : Branch(FOR), variableName(variableName), start(start), end(end), step(step)
{
}

For::~For(void)
{
   delete start;
   delete end;
   delete step;
}

bool For::Execute(PIVRINFO pInfo)
{
   bool invalid = false;
   int startValue = atoi(start->Evaluate(pInfo, invalid).c_str());
   int endValue = atoi(end->Evaluate(pInfo, invalid).c_str());
   int stepValue = step != NULL ? atoi(step->Evaluate(pInfo, invalid).c_str()) : 1;
   Next* forNext = (Next*) elseBranch;

   // Check for evaluation error
   if(invalid)
   {
      return(false);
   }

   if(stepValue == 0)
   {
      pInfo->ivLog->Print("      ERROR: STEP value cannot be 0\n");
      return(false);
   }

   if(!pInfo->variables->Set(variableName.c_str(), FormattedString("%d", startValue)))
   {
      // Error setting variable
      return(false);
   }

   // Prepare NEXT
   forNext->SetVariableName(variableName);
   forNext->SetEndValue(endValue);
   forNext->SetStartValue(startValue);
   forNext->SetStepValue(stepValue);

   next = ifBranch;

   // Branch constructs should always be able to provide an execution path
   // therefore, lack of a next step is an error
   return(next != NULL);
}

string For::Name(void)
{
   return("For");
}

GoSub::GoSub(deque<Command*>& returnStack) : returnStack(returnStack)
{
   returnNext = NULL;
}

GoSub::~GoSub(void)
{
}

void GoSub::Dump(Log* ivLog)
{
   char number[14];
   string childExpression;
   string thisExpression = Name();
   thisExpression += " (";
   sprintf(number, "%X", this);
   thisExpression += number;
   thisExpression += ")";

   if(next == NULL)
   {
      childExpression = "(NULL)";
   }
   else
   {
      childExpression = next->Name() + " (";
      sprintf(number, "%X", next);
      childExpression += number;
      childExpression += ")";
   }

   ivLog->Print("%s\n", (thisExpression + " Next: " + childExpression).c_str());

   if(next != NULL)
   {
      next->Dump(ivLog);
   }

   if(returnNext == NULL)
   {
      childExpression = "(NULL)";
   }
   else
   {
      childExpression = next->Name() + " (";
      sprintf(number, "%X", returnNext);
      childExpression += number;
      childExpression += ")";
   }

   ivLog->Print("%s\n", (thisExpression + " GoSub: " + childExpression).c_str());

   if(returnNext != NULL)
   {
      returnNext->Dump(ivLog);
   }
}

bool GoSub::Execute(PIVRINFO)
{
   // Place step to return to (follows GoSub) on returnStack
   returnStack.push_front(returnNext);

   // GoSubs should always have a destination, therefore, lack of a next step is an error
   return(next != NULL);
}

string GoSub::Name(void)
{
   return("GoSub");
}

void GoSub::SetNextCommand(Command* nextCommand)
{
   returnNext = next;
   next = nextCommand;
}

GoTo::GoTo(void)
{
}

GoTo::~GoTo(void)
{
}

bool GoTo::Execute(PIVRINFO)
{
   // GoSubs should always have a destination, therefore, lack of a next step is an error
   return(next != NULL);
}

string GoTo::Name(void)
{
   return("GoTo");
}

If::If(Expression* condition) : Branch(IF), condition(condition)
{
}

If::~If(void)
{
   delete condition;
}

void If::Dump(Log* ivLog)
{
   char number[14];
   string childExpression;
   string thisExpression = Name();
   thisExpression += " (";
   sprintf(number, "%X", this);
   thisExpression += number;
   thisExpression += ")";

   if(next == NULL)
   {
      childExpression = "(NULL)";
   }
   else
   {
      childExpression = next->Name() + " (";
      sprintf(number, "%X", next);
      childExpression += number;
      childExpression += ")";
   }

   ivLog->Print("%s\n", (thisExpression + " If: " + childExpression).c_str());

   if(next != NULL)
   {
      next->Dump(ivLog);
   }

   if(elseBranch == NULL)
   {
      childExpression = "(NULL)";
   }
   else
   {
      childExpression = next->Name() + " (";
      sprintf(number, "%X", elseBranch);
      childExpression += number;
      childExpression += ")";
   }

   ivLog->Print("%s\n", (thisExpression + " Else: " + childExpression).c_str());

   if(elseBranch != NULL)
   {
      elseBranch->Dump(ivLog);
   }
}

bool If::Execute(PIVRINFO pInfo)
{
   bool invalid = false;
   string conditionValue = condition->Evaluate(pInfo, invalid);
   bool result;

   // Check for evaluation error
   if(invalid)
   {
      pInfo->ivLog->Print("      %s condition '%s' is invalid\n", Name().c_str(), condition->GetString().c_str());
      return(false);
   }

   result = atoi(conditionValue.c_str()) != 0;
   next = result ? ifBranch : elseBranch;

   pInfo->ivLog->Print("      %s condition '%s' evaluates %s\n", Name().c_str(), condition->GetString().c_str(), result ? "TRUE" : "FALSE");

   return(next != NULL);
}

Expression* If::GetCondition(void)
{
   return(condition);
}

string If::Name(void)
{
   return("If");
}

void If::SetCondition(Expression* condition)
{
   this->condition = condition;
}

ElseIf::ElseIf(Expression* condition) : If(condition)
{
}

ElseIf::~ElseIf(void)
{
}

string ElseIf::Name(void)
{
   return("ElseIf");
}

Label::Label(void)
{
}

Label::~Label(void)
{
}

bool Label::Execute(PIVRINFO)
{
   // Labels may be the last command in the execution path
   // therefore, lack of a next step is acceptable
   return(true);
}

string Label::Name(void)
{
   return("Label");
}

Next::Next(void) : Branch(FOR)
{
}

Next::~Next(void)
{
}

bool Next::Execute(PIVRINFO pInfo)
{
   int currentValue = atoi(pInfo->variables->Get(variableName.c_str()));

   // Increment the variable
   currentValue += stepValue;

   if(!pInfo->variables->Set(variableName.c_str(), FormattedString("%d", currentValue)))
   {
      // Error setting variable
      return(false);
   }

   // Variable value must stay between values set by For
   if(((currentValue >= startValue) && (currentValue <= endValue))
      || ((currentValue <= startValue) && (currentValue >= endValue)))
   {
      // Nexts should always have a loop Label, therefore, lack of a loopCommand step is an error
      next = ifBranch;
   }
   else
   {
      next = elseBranch;
   }

   return(next != NULL);
}

string Next::Name(void)
{
   return("Next");
}

void Next::SetEndValue(int value)
{
   endValue = value;
}

void Next::SetStartValue(int value)
{
   startValue = value;
}

void Next::SetStepValue(int value)
{
   stepValue = value;
}

void Next::SetVariableName(string name)
{
   variableName = name;
}

Print::Print(Constant* printString) : printString(printString)
{
}

Print::~Print(void)
{
   delete printString;
}

bool Print::Execute(PIVRINFO pInfo)
{
   bool invalid = false;
   char* text;

   if(printString == NULL)
   {
      return(false);
   }

   text = pInfo->variables->Parse(printString->Evaluate(pInfo, invalid).c_str());

   // Check for error evaluating or parsing
   if((text == NULL) || invalid)
   {
      return(false);
   }
   else
   {
      pInfo->ivLog->Print("      %s\n", text);
      delete[] text;
      // Prints may be the last command in the execution path,
      // therefore, lack of a next step is acceptable
      return(true);
   }
}

string Print::Name(void)
{
   return("Print");
}

Return::Return(deque<Command*>& returnStack) : returnStack(returnStack)
{
}

Return::~Return(void)
{
}

bool Return::Execute(PIVRINFO)
{
   if(returnStack.size() == 0)
   {
      next = NULL;
   }
   else
   {
      next = returnStack.front();
      returnStack.pop_front();
   }

   // Gosubs may be the last command in the execution path,
   // therefore, lack of a next step is acceptable
   return(true);
}

string Return::Name(void)
{
   return("Return");
}

Select::Select(Expression* argument) : Case(argument, NULL)
{
}

Select::~Select(void)
{
   delete argument;
}

string Select::Name(void)
{
   return("Select");
}

While::While(Expression* condition) : If(condition)
{
   branchType = WHILE;
}

While::~While(void)
{
}

string While::Name(void)
{
   return("While");
}

Until::Until(Expression* condition) : While(condition)
{
   this->condition = new LogicalNot();
   this->condition->Add(new Parenthesised(condition));
}

Until::~Until(void)
{
}

string Until::Name(void)
{
   return("Until");
}

StepRunBASICScript::StepRunBASICScript(STEPREC& stepRecord) : Step(stepRecord), BASICScript(stepRecord.s1.arr)
{
   // Initialise flag so that script is only parsed once
   parsedScript = false;

   // Add keywords to list so that they are not used as labels etc.
   keywords.insert("ABS");
   keywords.insert("AND");
   keywords.insert("ASC");
   keywords.insert("BAND");
   keywords.insert("BNOT");
   keywords.insert("BOR");
   keywords.insert("BXOR");
   keywords.insert("CASE");
   keywords.insert("CHAR");
   keywords.insert("CHAR2HEX");
   keywords.insert("CONCAT");
   keywords.insert("DATEPART");
   keywords.insert("DO");
   keywords.insert("ELSE");
   keywords.insert("ELSEIF");
   keywords.insert("END");
   keywords.insert("EXIT");
   keywords.insert("FOR");
   keywords.insert("GLOBAL");
   keywords.insert("GOSUB");
   keywords.insert("GOTO");
   keywords.insert("HEX2CHAR");
   keywords.insert("IF");
   keywords.insert("LEFT");
   keywords.insert("LEN");
   keywords.insert("LOCAL");
   keywords.insert("LOOP");
   keywords.insert("LOWERCASE");
   keywords.insert("MOD");
   keywords.insert("NEXT");
   keywords.insert("NOT");
   keywords.insert("OR");
   keywords.insert("POS");
   keywords.insert("PRINT");
   keywords.insert("REM");
   keywords.insert("RETURN");
   keywords.insert("RIGHT");
   keywords.insert("ROUND");
   keywords.insert("SELECT");
   keywords.insert("SGN");
   keywords.insert("STEP");
   keywords.insert("SUBSTR");
   keywords.insert("THEN");
   keywords.insert("TIMEPART");
   keywords.insert("TO");
   keywords.insert("TRIM");
   keywords.insert("UNTIL");
   keywords.insert("UPPERCASE");
   keywords.insert("WHILE");
   keywords.insert("XOR");
}

StepRunBASICScript::~StepRunBASICScript(void)
{
   // Clean up command list
   while(!commands.empty())
   {
      delete commands.front();
      commands.pop_front();
   }
}

const char* StepRunBASICScript::GetStepName(void) const
{
	return("Run BASIC Script");
}

IV_STATUS StepRunBASICScript::Execute(PIVRINFO pInfo)
{
   IV_STATUS status;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   // Only parse it the first time
   if(!parsedScript)
   {
      if(!ParseCommands(pInfo->ivLog))
      {
         pInfo->ivLog->Print("      ERROR: Could not parse BASIC Script\n");
         return(IVR_FATAL);
      }

      parsedScript = true;
   }

   for(Command* currentCommand = commands.front(); currentCommand != NULL;
      currentCommand = currentCommand->GetNextCommand())
   {
      // Process any system events
      while(WaitForSingleObject(pInfo->messages->GetMessageSignal(), 0) == WAIT_OBJECT_0)
      {
         if((status = pInfo->ivBoard->ProcessSystemEvent()) != IVR_SUCCESS)
         {
            return(status);
         }
      }

      if(!currentCommand->Execute(pInfo))
      {
         pInfo->ivLog->Print("      ERROR: Failed to run BASIC Script\n");
         return(IVR_FATAL);
      }
   }

   return(IVR_SUCCESS);
}

Expression* StepRunBASICScript::ParseExpression(Log* ivLog, bool defensiveParenthesisReturn)
{
   Expression* expressionTree = NULL;
   // Initial condition to aid in parse checking
   Expression* newNode;
   // Stack of expression tree isolated by parentheses
   ExpressionTreeStack parenthesisStack;
   Function* curFunction = NULL;
   Script::TOKEN_TYPE previousTokenType;

   // Continue parsing while it's variables, numbers, operators, and parentheses
   for(string token = BASICScript.StartNewExpression();
      // Variable
      (token == "@")
      // Constant
      || (BASICScript.GetCurrentTokenType() == Script::NUMBER)
      || (BASICScript.GetCurrentToken() == "\"")
      // Operator
      || (BASICScript.GetCurrentTokenType() == Script::OPERATOR)
      // Function
      || (BASICScript.GetCurrentTokenType() == Script::FUNCTION)
      // Open parenthesis
      || (BASICScript.GetCurrentToken() == "(")
      // Close parenthesis
      || (BASICScript.GetCurrentToken() == ")");
      token = BASICScript.GetNextToken())
   {
      newNode = NULL;
      previousTokenType = BASICScript.GetPreviousExpressionTokenType();

      // Variable
      if(token == "@")
      {
         // Get remainder of variable name
         token += BASICScript.GetNextToken();

         if(BASICScript.GetCurrentTokenType() != Script::NAME)
         {
            ivLog->Print("      ERROR: Expected variable name and got '%s'\n", token.c_str());
            delete expressionTree;
            return(NULL);
         }

         if((previousTokenType != Script::SEPARATOR) && (previousTokenType != Script::OPERATOR)
            && (previousTokenType != Script::NONE))
         {
            ivLog->Print("      ERROR: Invalid syntax at '%s', expected an operator\n", token.c_str());
            delete expressionTree;
            return(NULL);
         }

         newNode = new Variable(token);
      }
      // Number or string
      else if((BASICScript.GetCurrentTokenType() == Script::NUMBER) || (token == "\""))
      {
         if((previousTokenType != Script::SEPARATOR) && (previousTokenType != Script::OPERATOR)
            && (previousTokenType != Script::NONE))
         {
            ivLog->Print("      ERROR: Invalid syntax at '%s', expected an operator\n", token.c_str());
            delete expressionTree;
            return(NULL);
         }

         // String
         if(token == "\"")
         {
            newNode = new Constant(BASICScript.GetUntilEndQuote());

            if(BASICScript.GetNextToken() != "\"")
            {
               ivLog->Print("      ERROR: Expected end quote for expression and got '%s'\n", BASICScript.GetCurrentToken().c_str());
               delete expressionTree;
               return(NULL);
            }
         }
         // Number
         else
         {
            newNode = new Constant(token);
         }
      }
      // Operator
      else if(BASICScript.GetCurrentTokenType() == Script::OPERATOR)
      {
         if(token == "+")
         {
            string previousToken = BASICScript.GetPreviousExpressionToken();

            if(((previousTokenType == Script::SEPARATOR) && (previousToken == ")"))
               || (previousTokenType == Script::NUMBER) || (previousTokenType == Script::NAME))
            {
               newNode = new Add();
            }
            else if(((previousTokenType == Script::SEPARATOR) && (previousToken != ")"))
               || (previousTokenType == Script::NONE) || (previousTokenType == Script::OPERATOR))
            {
               // Do nothing
            }
            else
            {
               ivLog->Print("      ERROR: Invalid syntax at '%s', expected to follow a value or separator but found '%s'\n", token.c_str(), previousToken.c_str());
               delete expressionTree;
               return(NULL);
            }
         }
         else if(token == "-")
         {
            string previousToken = BASICScript.GetPreviousExpressionToken();

            if(((previousTokenType == Script::SEPARATOR) && (previousToken == ")"))
               || (previousTokenType == Script::NUMBER) || (previousTokenType == Script::NAME))
            {
               newNode = new Subtract();
            }
            else if(((previousTokenType == Script::SEPARATOR) && (previousToken != ")"))
               || (previousTokenType == Script::NONE) || (previousTokenType == Script::OPERATOR))
            {
               newNode = new Negate();
            }
            else
            {
               ivLog->Print("      ERROR: Invalid syntax at '%s', expected to follow a value or separator but found '%s'\n", token.c_str(), previousToken.c_str());
               delete expressionTree;
               return(NULL);
            }
         }
         else if(token == "*")
         {
            newNode = new Multiply();
         }
         else if(token == "/")
         {
            newNode = new Divide();
         }
         else if(token == "MOD")
         {
            newNode = new Mod();
         }
         else if(token == "BNOT")
         {
            newNode = new BitwiseNot();
         }
         else if(token == "NOT")
         {
            newNode = new LogicalNot();
         }
         else if(token == "BAND")
         {
            newNode = new BitwiseAnd();
         }
         else if(token == "AND")
         {
            newNode = new LogicalAnd();
         }
         else if(token == "BXOR")
         {
            newNode = new BitwiseExclusiveOr();
         }
         else if(token == "XOR")
         {
            newNode = new LogicalExclusiveOr();
         }
         else if(token == "BOR")
         {
            newNode = new BitwiseOr();
         }
         else if(token == "OR")
         {
            newNode = new LogicalOr();
         }
         else if(token == "<")
         {
            newNode = new LessThan();
         }
         else if(token == "<=")
         {
            newNode = new LessThanOrEqualTo();
         }
         else if(token == ">")
         {
            newNode = new GreaterThan();
         }
         else if(token == ">=")
         {
            newNode = new GreaterThanOrEqualTo();
         }
         else if(token == "=")
         {
            newNode = new EqualTo();
         }
         else if((token == "!=") || (token == "<>"))
         {
            newNode = new NotEqualTo();
         }
         else if(token == "^")
         {
            newNode = new Exponentiate();
         }
      }
      // Functions
      else if(BASICScript.GetCurrentTokenType() == Script::FUNCTION)
      {
         if((previousTokenType != Script::SEPARATOR) && (previousTokenType != Script::NUMBER)
            && (previousTokenType != Script::OPERATOR) && (previousTokenType != Script::NONE))
         {
            ivLog->Print("      ERROR: Invalid syntax at '%s', expected a value or operator\n", token.c_str());
            delete expressionTree;
            return(NULL);
         }
         else if(token == "ABS")
         {
            curFunction = new Left();
         }
         else if(token == "ASC")
         {
            curFunction = new Asc();
         }
         else if(token == "CHAR")
         {
            curFunction = new Char();
         }
         else if(token == "CHAR2HEX")
         {
            curFunction = new Char2Hex();
         }
         else if(token == "CONCAT")
         {
            curFunction = new Concat();
         }
         else if(token == "DATEPART")
         {
            curFunction = new DatePart();
         }
         else if(token == "HEX2CHAR")
         {
            curFunction = new Hex2Char();
         }
         else if(token == "LEFT")
         {
            curFunction = new Left();
         }
         else if(token == "LEN")
         {
            curFunction = new Len();
         }
         else if(token == "LOWERCASE")
         {
            curFunction = new Lowercase();
         }
         else if(token == "POS")
         {
            curFunction = new Pos();
         }
         else if(token == "RIGHT")
         {
            curFunction = new Right();
         }
         else if(token == "ROUND")
         {
            curFunction = new Round();
         }
         else if(token == "SGN")
         {
            curFunction = new Round();
         }
         else if(token == "SUBSTR")
         {
            curFunction = new SubStr();
         }
         else if(token == "TIMEPART")
         {
            curFunction = new TimePart();
         }
         else if(token == "TRIM")
         {
            curFunction = new Trim();
         }
         else if(token == "UPPERCASE")
         {
            curFunction = new Uppercase();
         }

         newNode = curFunction;
      }
      else
      {
         if(token == "(")
         {
            if(previousTokenType != Script::FUNCTION)
            {
               // Push current expression tree on stack
               parenthesisStack.push_front(expressionTree);

               // Reset expression Tree
               expressionTree = NULL;
            }
            else
            {
               if(curFunction != NULL)
               {
                  for(unsigned int curParam = 0; curParam < curFunction->GetExpectedParamsCount(); curParam++)
                  {
                     Expression* value = ParseExpression(ivLog, (curParam == (curFunction->GetExpectedParamsCount() - 1)));

                     if(value == NULL)
                     {
                        ivLog->Print(FormattedString( "      ERROR: Invalid parameter %d in function '%s'\n", curParam + 1, curFunction->Name().c_str()));
                        delete expressionTree;
                        return(NULL);
                     }

                     curFunction->AddParameter(value);
                  }
               }
               else
               {
                  ivLog->Print("      ERROR: Parser missing function\n");
                  delete expressionTree;
                  return(NULL);
               }
            }
         }
         else if(token == ")")
         {
            // Check that there is something on the stack
            if(parenthesisStack.size() == 0)
            {
               if(defensiveParenthesisReturn)
               {
                  // How else do we tell that we are evaluating an expression
                  // that will end with either a comma, or a close parenthesis?
                  break;
               }
               else
               {
                  ivLog->Print("      ERROR: Close parenthesis without open parenthesis\n");
                  delete expressionTree;
                  return(NULL);
               }
            }

            // Wrap current expression tree with Parenthesised
            newNode = new Parenthesised(expressionTree);

            // Set expressionTree to top expression tree on stack
            expressionTree = parenthesisStack.front();

            // Pop expression tree off stack
            parenthesisStack.pop_front();
         }
         else
         {
            ivLog->Print("      ERROR: Invalid separator '%s'\n", token.c_str());
            delete expressionTree;
            return(NULL);
         }
      }

      if(newNode != NULL)
      {
         // Add new expression to expressionTree
         expressionTree = expressionTree == NULL ? newNode : expressionTree->Add(newNode);
      }
   }

   // Check for left over parentheses
   if(!parenthesisStack.empty())
   {
      ivLog->Print("      ERROR: Open parenthesis without close parenthesis\n");
      delete expressionTree;
      return(NULL);
   }

   return(expressionTree);
}

// Structure for connecting branches
typedef struct BranchInfo
{
   Branch* lastBranchStart;
   list<Command*> branchEnds;
} BranchInfo;

bool StepRunBASICScript::ParseCommands(Log* ivLog)
{
   // Since loops and conditional statements can be nested, this stack provides
   // a location to work with the current loop or conditional
   deque<BranchInfo> nestStack;

   // List of Labels by name
   map<string, Label*> labels;

   // Lists of Commands for each Label, listed by Label name
   map<string, list<Command*> > labelsNeeded;

   // Most recently created Command
   Command* currentCommand = NULL;

   // Command created in this iteration
   Command* newCommand;

   // Iterate through the script
   for(BASICScript.GetFirstToken();
      (BASICScript.GetCurrentTokenType() != Script::NONE) && (BASICScript.GetCurrentTokenType() != Script::INVALID);
      BASICScript.GetNextToken())
   {
      newCommand = NULL;

      // Keyword or label
      if(BASICScript.GetCurrentTokenType() == Script::NAME)
      {
         string name = BASICScript.GetCurrentToken();

         // Keyword
         if(name == "IF")
         {
            // Get conditional expression for IF
            Expression* condition = ParseExpression(ivLog);
            // Get token that follows expression
            string tokenAfterExpression = BASICScript.GetCurrentToken();

            if(condition == NULL)
            {
               ivLog->Print("      ERROR: Expression required for IF condition\n");
               return(false);
            }

            if(tokenAfterExpression != "THEN")
            {
               ivLog->Print("      ERROR: Expected THEN after IF condition and got '%s'\n", tokenAfterExpression.c_str());
               delete condition;
               return(false);
            }

            nestStack.push_front(BranchInfo());
            newCommand = nestStack.front().lastBranchStart = new If(condition);
         }
         else if(name == "ELSEIF")
         {
            // Get conditional expression for ELSEIF
            Expression* condition = ParseExpression(ivLog);
            // Get token that follows expression
            string tokenAfterExpression = BASICScript.GetCurrentToken();
            ElseIf* newBranch;

            if(condition == NULL)
            {
               ivLog->Print("      ERROR: Expression required for ELSEIF condition\n");
               return(false);
            }

            if(tokenAfterExpression != "THEN")
            {
               ivLog->Print("      ERROR: Expected THEN after ELSEIF condition and got '%s'\n", tokenAfterExpression.c_str());
               delete condition;
               return(false);
            }

            if(nestStack.empty())
            {
               ivLog->Print("      ERROR: ELSEIF requires a preceding IF\n");
               delete condition;
               return(false);
            }
            newCommand = newBranch = new ElseIf(condition);
            nestStack.front().lastBranchStart->SetElseBranch(newBranch);
            nestStack.front().lastBranchStart = newBranch;
            nestStack.front().branchEnds.push_back(currentCommand);
         }
         else if(name == "ELSE")
         {
            Else* newBranch;

            if(nestStack.empty())
            {
               ivLog->Print("      ERROR: ELSE requires a preceding IF\n");
               return(false);
            }

            newCommand = newBranch = new Else();
            nestStack.front().lastBranchStart->SetElseBranch(newBranch);
            nestStack.front().lastBranchStart = newBranch;
            nestStack.front().branchEnds.push_back(currentCommand);
         }
         else if(name == "END")
         {
            // We will need to verify that we are ending the correct structure
            // i.e. END IF, END SELECT, or simply END
            // This can be done later.
            string token = BASICScript.GetNextToken();

            if(nestStack.empty())
            {
               ivLog->Print("      ERROR: END requires a control block to terminate\n");
               return(false);
            }

            if(token == "IF")
            {
               if(nestStack.front().lastBranchStart->GetBranchType() != Branch::IF)
               {
                  ivLog->Print("      ERROR: END IF must match an IF\n");
                  return(false);
               }
            }
            else if(token == "SELECT")
            {
               if(nestStack.front().lastBranchStart->GetBranchType() != Branch::CASE)
               {
                  ivLog->Print("      ERROR: END SELECT must match a SELECT\n");
                  return(false);
               }
            }

            newCommand = new End();
            nestStack.front().lastBranchStart->SetElseBranch(newCommand);

            // Update next Command of the last Command for all the conditional clauses
            for(list<Command*>::iterator branchEnd = nestStack.front().branchEnds.begin();
               branchEnd != nestStack.front().branchEnds.end(); branchEnd++)
            {
               (*branchEnd)->SetNextCommand(newCommand);
            }

            // Pop the BranchStack
            nestStack.pop_front();
         }
         else if(name == "THEN")
         {
            ivLog->Print("      ERROR: IF required before THEN\n");
            return(false);
         }
         else if(name == "FOR")
         {
            string variableName = BASICScript.GetNextToken();
            Expression* start;
            Expression* end;
            Expression* step = NULL;
            string tokenAfterExpression;

            if(variableName != "@")
            {
               ivLog->Print("      ERROR: Expected variable name for FOR and got '%s'\n", variableName.c_str());
               return(false);
            }

            // Get remainder of variable name
            variableName += BASICScript.GetNextToken();

            if(BASICScript.GetCurrentTokenType() != Script::NAME)
            {
               ivLog->Print("      ERROR: Expected variable name and got '%s'\n", variableName.c_str());
               return(false);
            }

            tokenAfterExpression = BASICScript.GetNextToken();

            if(tokenAfterExpression != "=")
            {
               ivLog->Print("      ERROR: Expected assignment operator '=' after variable name and got '%s'\n", tokenAfterExpression.c_str());
               return(false);
            }

            start = ParseExpression(ivLog);

            if(start == NULL)
            {
               ivLog->Print("      ERROR: Expression required for FOR start\n");
               return(false);
            }

            tokenAfterExpression = BASICScript.GetCurrentToken();

            if(tokenAfterExpression != "TO")
            {
               ivLog->Print("      ERROR: Expected TO after FOR start expression and got '%s'\n", tokenAfterExpression.c_str());
               delete start;
               return(false);
            }

            end = ParseExpression(ivLog);

            if(end == NULL)
            {
               ivLog->Print("      ERROR: Expression required for FOR end\n");
               delete start;
               return(false);
            }

            tokenAfterExpression = BASICScript.GetCurrentToken();

            if(tokenAfterExpression != "\n")
            {
               if(tokenAfterExpression != "STEP")
               {
                  ivLog->Print("      ERROR: Expected end of line or STEP after FOR end and got '%s'\n", tokenAfterExpression.c_str());
                  delete start;
                  delete end;
                  return(false);
               }

               step = ParseExpression(ivLog);

               if(step == NULL)
               {
                  ivLog->Print("      ERROR: Expression required for FOR step\n");
                  delete start;
                  delete end;
                  return(false);
               }

               tokenAfterExpression = BASICScript.GetCurrentToken();
            }

            if(tokenAfterExpression != "\n")
            {
               ivLog->Print("      ERROR: Expected end of line after FOR statement and got '%s'\n", tokenAfterExpression.c_str());
               delete start;
               delete end;
               delete step;
               return(false);
            }

            nestStack.push_front(BranchInfo());
            newCommand = nestStack.front().lastBranchStart = new For(variableName, start, end, step);
         }
         else if(name == "TO")
         {
            ivLog->Print("      ERROR: FOR required before TO\n");
            return(false);
         }
         else if(name == "STEP")
         {
            ivLog->Print("      ERROR: STEP must follow FOR and TO on same line\n");
            return(false);
         }
         else if(name == "NEXT")
         {
            Next* newBranch;

            if((nestStack.size() == 0) || (nestStack.front().lastBranchStart->GetBranchType() != Branch::FOR))
            {
               ivLog->Print("      ERROR: Invalid use of NEXT, cannot find FOR\n");
               return(false);
            }

            newBranch = new Next();
            nestStack.front().lastBranchStart->SetElseBranch(newBranch);
            newBranch->SetNextCommand(nestStack.front().lastBranchStart->GetIfBranch());
            commands.push_back(newBranch);
            currentCommand->SetNextCommand(newBranch);
            currentCommand = new End();
            newBranch->SetElseBranch(currentCommand);
            commands.push_back(currentCommand);

            // Update next Command of the last Command for all the conditional clauses
            for(list<Command*>::iterator branchEnd = nestStack.front().branchEnds.begin();
               branchEnd != nestStack.front().branchEnds.end(); branchEnd++)
            {
               (*branchEnd)->SetNextCommand(currentCommand);
            }

            // Pop the BranchStack
            nestStack.pop_front();
         }
         else if(name == "DO")
         {
            // Peek at the next token
            string token = BASICScript.GetNextToken();

            if(token == "WHILE")
            {
               Expression* condition;

               if((condition = ParseExpression(ivLog)) == NULL)
               {
                 ivLog->Print("      ERROR: Expression required for WHILE\n");
                 return(false);
               }

               nestStack.push_front(BranchInfo());
               newCommand = nestStack.front().lastBranchStart = new While(condition);
            }
            else if(token == "UNTIL")
            {
               Expression* condition;

               if((condition = ParseExpression(ivLog)) == NULL)
               {
                 ivLog->Print("      ERROR: Expression required for UNTIL\n");
                 return(false);
               }

               nestStack.push_front(BranchInfo());
               newCommand = nestStack.front().lastBranchStart = new Until(condition);
            }
            else if(token == "\n")
            {
               nestStack.push_front(BranchInfo());
               newCommand = nestStack.front().lastBranchStart = new Do();
            }
            else
            {
               ivLog->Print("      ERROR: Expected optional WHILE or UNTIL after DO and got '%s'\n", token.c_str());
               return(false);
            }
         }
         else if(name == "WHILE")
         {
            ivLog->Print("      ERROR: WHILE must follow DO or LOOP\n");
            return(false);
         }
         else if(name == "UNTIL")
         {
            ivLog->Print("      ERROR: UNTIL must follow DO or LOOP\n");
            return(false);
         }
         else if(name == "LOOP")
         {
            string token = BASICScript.GetNextToken();

            if(nestStack.size() == 0)
            {
               ivLog->Print("      ERROR: LOOP must be used with a preceding DO\n");
               return(false);
            }

            if(token == "WHILE")
            {
               While* newWhile;
               Expression* condition;

               if(nestStack.front().lastBranchStart->GetBranchType() == Branch::WHILE)
               {
                  ivLog->Print("      ERROR: Only one WHILE or UNTIL is permitted per LOOP\n");
                  return(false);
               }
               else if(nestStack.front().lastBranchStart->GetBranchType() != Branch::DO)
               {
                  ivLog->Print("      ERROR: LOOP must be used with a preceding DO\n");
                  return(false);
               }

               if((condition = ParseExpression(ivLog)) == NULL)
               {
                 ivLog->Print("      ERROR: Expression required for WHILE\n");
                 return(false);
               }

               newWhile = new While(condition);
               currentCommand->SetNextCommand(newWhile);
               newWhile->SetNextCommand(nestStack.front().lastBranchStart);
               commands.push_back(newWhile);
               currentCommand = newWhile;
               newCommand = new End();
               newWhile->SetElseBranch(newCommand);
            }
            else if(token == "UNTIL")
            {
               Until* newUntil;
               Expression* condition;

               if(nestStack.front().lastBranchStart->GetBranchType() == Branch::WHILE)
               {
                  ivLog->Print("      ERROR: Only one WHILE or UNTIL is permitted per LOOP\n");
                  return(false);
               }
               else if(nestStack.front().lastBranchStart->GetBranchType() != Branch::DO)
               {
                  ivLog->Print("      ERROR: LOOP must be used with a preceding DO\n");
                  return(false);
               }

               if((condition = ParseExpression(ivLog)) == NULL)
               {
                 ivLog->Print("      ERROR: Expression required for WHILE\n");
                 return(false);
               }

               newUntil = new Until(condition);
               currentCommand->SetNextCommand(newUntil);
               newUntil->SetNextCommand(nestStack.front().lastBranchStart);
               commands.push_back(newUntil);
               currentCommand = new End();
               newUntil->SetElseBranch(currentCommand);
               commands.push_back(currentCommand);
            }
            else if((BASICScript.GetCurrentTokenType() == Script::NONE) || (token == "\n"))
            {
               currentCommand->SetNextCommand(nestStack.front().lastBranchStart);
               currentCommand = new End();
               nestStack.front().lastBranchStart->SetElseBranch(currentCommand);
               commands.push_back(currentCommand);
            }
            else
            {
               ivLog->Print("      ERROR: Expected optional WHILE or UNTIL after LOOP and got '%s'\n", token.c_str());
               return(false);
            }

            // Update next Command of the last Command for all the conditional clauses
            for(list<Command*>::iterator branchEnd = nestStack.front().branchEnds.begin();
               branchEnd != nestStack.front().branchEnds.end(); branchEnd++)
            {
               (*branchEnd)->SetNextCommand(currentCommand);
            }

            // Pop the BranchStack
            nestStack.pop_front();
         }
         else if(name == "SELECT")
         {
            string token = BASICScript.GetNextToken();

            if(token == "CASE")
            {
               Expression* argument = ParseExpression(ivLog);

               if(argument == NULL)
               {
                  ivLog->Print("      ERROR: Cannot parse argument expression for SELECT\n");
                  return(false);
               }

               nestStack.push_front(BranchInfo());
               newCommand = nestStack.front().lastBranchStart = new Select(argument);
            }
            else
            {
               ivLog->Print("      ERROR: SELECT must be followed by CASE\n");
               return(false);
            }
         }
         else if(name == "CASE")
         {
            Expression* condition;
            Case* newCase;
            Case* curCaseOnScriptLine = NULL;
            vector<Command*> newCommands;

            if((nestStack.size() == 0) || (nestStack.front().lastBranchStart->GetBranchType() != Branch::CASE))
            {
               ivLog->Print("      ERROR: CASE must be used in a SELECT\n");
               return(false);
            }

            condition = ParseExpression(ivLog);

            // Not first case after select
            if(nestStack.front().lastBranchStart != currentCommand)
            {
              nestStack.front().branchEnds.push_back(currentCommand);
            }

            if(condition == NULL)
            {
               if(BASICScript.GetCurrentToken() != "ELSE")
               {
                  ivLog->Print("      ERROR: Cannot parse condition expression for CASE\n");
                  return(false);
               }

               newCase = new CaseElse();
               newCommands.push_back(newCase);
            }
            else
            {
               newCase = new Case(((Case*) nestStack.front().lastBranchStart)->GetArgument(), condition);
               newCommands.push_back(newCase);
               curCaseOnScriptLine = newCase;
            }

            nestStack.front().lastBranchStart->SetElseBranch(newCase);
            nestStack.front().lastBranchStart = newCase;

            do
            {
               if(BASICScript.GetCurrentToken() == "TO")
               {
                  condition = ParseExpression(ivLog);
                  if(condition == NULL)
                  {
                     ivLog->Print("      ERROR: Cannot parse condition expression for CASE after TO\n");
                     return(false);
                  }
                  newCase->SetUpperBoundCondition(condition);
               }
               newCase = NULL;

               if ((condition != NULL) && (BASICScript.GetCurrentToken() == ","))
               {
                  condition = ParseExpression(ivLog);
                  if(condition == NULL)
                  {
                     ivLog->Print("      ERROR: Cannot parse condition expression for CASE\n");
                     return(false);
                  }
                  newCase = new Case(((Case*) nestStack.front().lastBranchStart)->GetArgument(), condition, curCaseOnScriptLine);
                  newCommands.push_back(newCase);
                  curCaseOnScriptLine = newCase;
                  nestStack.front().lastBranchStart->SetElseBranch(newCase);
                  nestStack.front().lastBranchStart = newCase;
               }
            }
            while((BASICScript.GetCurrentTokenType() != Script::NONE) && (BASICScript.GetCurrentToken() != "\n") && (condition != NULL));

            if(newCommands.size() > 0)
            {
               // Add new Commands to clean up list
               for(unsigned int curCommandIndex = 0; curCommandIndex < newCommands.size(); curCommandIndex++)
               {
                  commands.push_back(newCommands[curCommandIndex]);
                  // Ensure all new commands follow each other
                  if(curCommandIndex > 0)
                  {
                     newCommands[curCommandIndex]->SetNextCommand(newCommands[curCommandIndex - 1]);
                  }
               }

               if(currentCommand != NULL)
               {
                  // Update previous Command with new Command
                  currentCommand->SetNextCommand(newCommands[0]);
               }

               // Reset current Command
               currentCommand = newCommands[newCommands.size() - 1];
            }
         }
         else if(name == "BREAK")
         {
            deque<BranchInfo>::iterator pos;

            for(pos = nestStack.begin();
               (pos != nestStack.end()) && (pos->lastBranchStart->GetBranchType() == Branch::IF);
               pos++);

            if(pos == nestStack.end())
            {
              ivLog->Print("      ERROR: BREAK must be used within a breakable branch\n");
              return(false);
            }

            newCommand = new Break();
            pos->branchEnds.push_back(newCommand);
         }
         else if(name == "EXIT")
         {
            string nextToken = BASICScript.GetNextToken();

            if((BASICScript.GetCurrentTokenType() == Script::NONE) || (nextToken == "GOSUB") || (nextToken == "\n"))
            {
               newCommand = new Return(returnStack);
            }
            else if((nextToken == "DO") || (nextToken == "WHILE") || (nextToken == "UNTIL") || (nextToken == "FOR"))
            {
               Branch::BRANCH_TYPE matchBranch = nextToken == "FOR" ? Branch::FOR : Branch::WHILE;
               deque<BranchInfo>::iterator pos;

               for(pos = nestStack.begin();
                  (pos != nestStack.end()) && (pos->lastBranchStart->GetBranchType() != matchBranch);
                  pos++);

               if(pos == nestStack.end())
               {
                   ivLog->Print("      ERROR: EXIT does not have a matching %s branch\n", nextToken.c_str());
                   return(false);
               }

               newCommand = new Break;
               pos->branchEnds.push_back(newCommand);
            }
            else
            {
                ivLog->Print("      ERROR: EXIT type unrecognised\n", nextToken.c_str());
                return(false);
            }
         }
         else if(name == "GOTO")
         {
            string tokenAfterName = BASICScript.GetNextToken();

            if(BASICScript.GetCurrentTokenType() != Script::NAME)
            {
               ivLog->Print("      ERROR: Expected label for GOTO and got '%s'\n", tokenAfterName.c_str());
               return(false);
            }

            if(keywords.count(tokenAfterName) != 0)
            {
               ivLog->Print("      ERROR: Keyword '%s' cannot be used as a label\n", tokenAfterName.c_str());
               return(false);
            }

            if((BASICScript.GetCurrentTokenType() != Script::NONE) && (BASICScript.GetNextToken() != "\n"))
            {
               ivLog->Print("      ERROR: Expected end of line after GOTO %s: and got '%s'\n", tokenAfterName.c_str(), BASICScript.GetCurrentToken().c_str());
               return(false);
            }

            newCommand = new GoTo;

            // Add Command to list of Commands to be updated with this Label
            labelsNeeded[tokenAfterName].push_back(newCommand);
         }
         else if(name == "GOSUB")
         {
            string tokenAfterName = BASICScript.GetNextToken();

            if(BASICScript.GetCurrentTokenType() != Script::NAME)
            {
               ivLog->Print("      ERROR: Expected label for GOSUB and got '%s'\n", tokenAfterName.c_str());
               return(false);
            }

            if(keywords.count(tokenAfterName) != 0)
            {
               ivLog->Print("      ERROR: Keyword '%s' cannot be used as a label\n", tokenAfterName.c_str());
               return(false);
            }

            if((BASICScript.GetCurrentTokenType() != Script::NONE) && (BASICScript.GetNextToken() != "\n"))
            {
               ivLog->Print("      ERROR: Expected end of line after GOSUB %s: and got '%s'\n", tokenAfterName.c_str(), BASICScript.GetCurrentToken().c_str());
               return(false);
            }

            newCommand = new GoSub(returnStack);

            // Add Command to list of Commands to be updated with this Label
            labelsNeeded[tokenAfterName].push_back(newCommand);
         }
         else if(name == "RETURN")
         {
            if((BASICScript.GetCurrentTokenType() != Script::NONE) && (BASICScript.GetNextToken() != "\n"))
            {
               ivLog->Print("      ERROR: Expected end of line after RETURN: and got '%s'\n", BASICScript.GetCurrentToken().c_str());
               return(false);
            }

            newCommand = new Return(returnStack);
         }
         else if((name == "GLOBAL") || (name == "LOCAL"))
         {
            bool global = name == "GLOBAL"; 
            string variableName = BASICScript.GetNextToken();
            string nextToken;
            Expression* value;

            if(variableName != "@")
            {
               ivLog->Print("      ERROR: Expected variable name and got '%s'\n", variableName.c_str());
               return(false);
            }

            variableName += BASICScript.GetNextToken();

            if(BASICScript.GetCurrentTokenType() != Script::NAME)
            {
               ivLog->Print("      ERROR: Expected variable name and got '%s'\n", variableName.c_str());
               return(false);
            }

            nextToken = BASICScript.GetNextToken();

            if(nextToken == "\n")
            {
               value = new Constant("");
            }
            else if(nextToken == "=")
            {
               value = ParseExpression(ivLog);
            }
            else
            {
               ivLog->Print("      ERROR: Expected assignment operator '=' after variable name and got '%s'\n", BASICScript.GetCurrentToken().c_str());
               return(false);
            }

            if(value == NULL)
            {
               ivLog->Print("      ERROR: Cannot parse assignment expression for '%s'\n", variableName.c_str());
               return(false);
            }

            newCommand = new ScopedAssignment(variableName, value, global);
         }
         else if(name == "PRINT")
         {
            string tokenAfterName = BASICScript.GetNextToken();
            string printString;

            if(tokenAfterName == "\"")
            {
               printString = BASICScript.GetUntilEndQuote();
            }
            else if((BASICScript.GetCurrentTokenType() != Script::NONE) && (tokenAfterName != "\n"))
            {
               ivLog->Print("      ERROR: Expected quote for PRINT and got '%s'\n", tokenAfterName.c_str());
               return(false);
            }

            if((tokenAfterName == "\"") && (BASICScript.GetNextToken() != "\""))
            {
               ivLog->Print("      ERROR: Expected end quote for PRINT and got '%s'\n", BASICScript.GetCurrentToken().c_str());
               return(false);
            }

            newCommand = new Print(new Constant(printString));
         }
         // Label (all names followed by a colon are interpreted as Labels)
         else if(BASICScript.GetNextToken() == ":")
         {
            string tokenAfterName = BASICScript.GetCurrentToken();

            if(keywords.count(tokenAfterName) != 0)
            {
               ivLog->Print("      ERROR: Keyword '%s' cannot be used as a label\n", tokenAfterName.c_str());
               return(false);
            }

            if(labels.count(name) != 0)
            {
               ivLog->Print("      ERROR: Label name '%s' already exists\n", name.c_str());
               return(false);
            }

            newCommand = labels[name] = new Label();
         }
         else
         {
            ivLog->Print("      ERROR: Expected keyword or label and got '%s'\n", name.c_str());
            return(false);
         }
      }
      // Variable
      else if(BASICScript.GetCurrentToken() == "@")
      {
         string variableName = "@" + BASICScript.GetNextToken();
         string nextToken;
         Expression* value;

         if(BASICScript.GetCurrentTokenType() != Script::NAME)
         {
            ivLog->Print("      ERROR: Expected variable name and got '%s'\n", variableName.c_str());
            return(false);
         }

         nextToken = BASICScript.GetNextToken();

         if(nextToken == "\n")
         {
            value = new Constant("");
         }
         else if(nextToken == "=")
         {
            value = ParseExpression(ivLog);
         }
         else
         {
            ivLog->Print("      ERROR: Expected assignment operator '=' after variable name and got '%s'\n", BASICScript.GetCurrentToken().c_str());
            return(false);
         }

         if(value == NULL)
         {
            ivLog->Print("      ERROR: Cannot parse assignment expression for '%s'\n", variableName.c_str());
            return(false);
         }

         newCommand = new Assignment(variableName, value);
      }
      // Blank line
      else if(BASICScript.GetCurrentToken() == "\n")
      {
         // Do nothing
      }
      // Invalid starting token
      else
      {
         ivLog->Print("      ERROR: Expected keyword or label and got '%s'\n", BASICScript.GetCurrentToken().c_str());
         return(false);
      }

      if(currentCommand != NULL)
      {
         // Update previous Command with new Command
         currentCommand->SetNextCommand(newCommand);
      }

      if(newCommand != NULL)
      {
         commands.push_back(newCommand);
         currentCommand = newCommand;
      }
   }

   if(!nestStack.empty())
   {
      ivLog->Print("      ERROR: IF/ELSEIF/ELSE clauses require an END, %d %s missing\n", nestStack.size(), nestStack.size() == 1 ? "is" : "are");
      return(false);
   }

   // Iterate through labels
   for(map<string, list<Command*> >::iterator label = labelsNeeded.begin();
      label != labelsNeeded.end(); label++)
   {
      if(labels.count(label->first) == 0)
      {
         ivLog->Print("      ERROR: Label '%s' not found\n", label->first.c_str());
         return(false);
      }

      // Iterate through list of commands waiting for label
      for(list<Command*>::iterator command = label->second.begin();
         command != label->second.end(); command++)
      {
         // Set label
         (*command)->SetNextCommand(labels[label->first]);
      }
   }

   return(true);
}
