#include "..\inet\Socket.h"
#include "..\escutil\iv_str.h"          // HexStr2CharStr
#include "..\BoardDLLs\brd_generic.h"   // ProcessSystemEvent()
#include "TFEngine_communication.h"
#include "step_UDPIPReceive.h"

StepUDPIPReceive::StepUDPIPReceive(STEPREC& stepRecord) : Step(stepRecord)
{
   socketHandle = stepRecord.s1.arr;
   peerNameVariable = stepRecord.s7.arr;
   peerPortVariable = stepRecord.s8.arr;
   responseVariable = stepRecord.s2.arr;
   timeout = stepRecord.s3.arr;
   includeTerminationString = (bool) stepRecord.u1;
   hexString = (bool) stepRecord.u2;
   uppercase = (bool) stepRecord.u3;
}

StepUDPIPReceive::~StepUDPIPReceive(void)
{
}

const char* StepUDPIPReceive::GetStepName(void) const
{
	return("UDP Receive");
}

IV_STATUS StepUDPIPReceive::Execute(PIVRINFO pInfo)
{
   char* value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   int iTimeout;
   LARGE_INTEGER liTimeout;
   HANDLE waitObjs[3];
   HANDLE hTimer;
   UDPSocket socket(pInfo->ivLog);
   ACE_INET_Addr peer;
   string returnMessage;
   int ret;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

   if((value = pInfo->variables->ParseParameter("Socket handle", socketHandle.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      ACE_HANDLE handle = (ACE_HANDLE) atoi(value);
      delete[] value;
      socket.set_handle(handle);
   }

   if((value = pInfo->variables->ParseParameter("Timeout", timeout.c_str(), "@TIMEOUT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iTimeout = atoi(value);
      delete[] value;

      if((iTimeout < 0) && (iTimeout != FOREVER))
      {
		   pInfo->ivLog->Print("      ERROR: Silence TIMEOUT '%s' must be >= 0\n",
            pInfo->variables->IsParsedEmpty(timeout.c_str()) ? "@TIMEOUT" : timeout.c_str());
         bParmsValid = false;
      }
   }

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }

   // Initialise value to empty string
   if(!pInfo->variables->Set(responseVariable.c_str(), ""))
   {
      pInfo->ivLog->Print("      ERROR: Failed to clear response variable '%s'\n", responseVariable.c_str());
      return(IVR_FATAL);
   }

   if((ret = socket.SelectEvents(FD_READ)) != Socket::SUCCESS)
   {
      pInfo->ivLog->Print("      ERROR: Failed to set socket events to FD_READ\n");
      pInfo->ivLog->Print("         %s\n", Socket::WSAErrorString(ret));
      return(IVR_FAIL);
   }

   // Set timeout timer
   hTimer = CreateWaitableTimer(0, false, 0);
   if(iTimeout != FOREVER)
   {
      liTimeout.QuadPart = -(LONGLONG) iTimeout * 10000;
      SetWaitableTimer(hTimer, &liTimeout, 0, 0, 0, false);
   }

   // Set up events array
   waitObjs[0] = pInfo->messages->GetMessageSignal();  // system event such as shutdown
   waitObjs[1] = hTimer;                               // timer for the timeout value
   waitObjs[2] = socket.GetEventHandle();              // signal from socket

   while(returnMessage.length() == 0)
   {
      switch(WaitForMultipleObjects(3, waitObjs, false, INFINITE))
      {
         // System event
         case WAIT_OBJECT_0:
         {
            IV_STATUS status = pInfo->ivBoard->ProcessSystemEvent();

            if(status != IVR_SUCCESS)
            {
               CloseHandle(hTimer);
               return(status);
            }

            break;
         }

         // Timed out
         case WAIT_OBJECT_0 + 1:
         {
            pInfo->ivLog->Print("      Timed out waiting for complete response\n");
            // Store as much of the message as we have received in the result variable
            pInfo->variables->Set(responseVariable.c_str(), "");
            CloseHandle(hTimer);
            return(IVR_TIMEOUT);
         }

         // Socket signalled
         case WAIT_OBJECT_0 + 2:
         {
            const char* receivedBytes;
            int bytesReceived;

            if((ret = socket.ReceiveBytes(receivedBytes, bytesReceived, peer)) == Socket::SUCCESS)
            {
               if(receivedBytes != NULL)
               {
                  if(hexString)
                  {
                     char* temp = new char[2*bytesReceived + 1];
                     CharStr2HexStr((const unsigned char*) receivedBytes, temp, bytesReceived, uppercase);
                     temp[2*bytesReceived] = '\0';
                     returnMessage = temp;
                     delete[] temp;
                  }
                  else
                  {
                     char* temp = new char[bytesReceived + 1];
                     strncpy(temp, receivedBytes, bytesReceived);
                     temp[bytesReceived] = '\0';
                     returnMessage = temp;
                     delete[] temp;
                  }

                  if(returnMessage.length() < (unsigned int) (hexString ? 2 : 1)*bytesReceived)
                  {
                     pInfo->ivLog->Print("      WARNING: Receive terminated on NULL character\n");
                  }
               }
            }
            else if(ret == WSAEWOULDBLOCK)
            {
               // Allow 10 ms for socket to clear
               Sleep(10);
            }
            else
            {
               const char* errorMessage = Socket::WSAErrorString(ret);

               if(ret == WSAENOTCONN)
               {
                  pInfo->ivLog->Print("      ERROR: Socket not connected\n");
               }
               else
               {
                  pInfo->ivLog->Print("      ERROR: Failed to receive\n");
                  pInfo->ivLog->Print("         '%s'\n", errorMessage);
               }

               pInfo->variables->Set(responseVariable.c_str(), errorMessage);
               CloseHandle(hTimer);
               return(IVR_FAIL);
            }

            break;
         }

         // WaitForMultipleObjects failed
         case WAIT_FAILED:
         {
            DWORD errorID = GetLastError();
            char MsgBuf[1024];

            FormatMessage(
               FORMAT_MESSAGE_FROM_SYSTEM,
               0,
               errorID,
               MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
               MsgBuf,
               1024,
               0
            );

            pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
            pInfo->variables->Set(responseVariable.c_str(), MsgBuf);
            CloseHandle(hTimer);
            return(IVR_FAIL);
         }
      }
   }

   CloseHandle(hTimer);

   if(!pInfo->variables->Set(responseVariable.c_str(), returnMessage.c_str()))
   {
      pInfo->ivLog->Print("      ERROR: Failed to set variable to receive message '%s'\n", responseVariable.c_str());
      return(IVR_FATAL);
   }

   if(!pInfo->variables->IsEmpty(peerNameVariable.c_str()) && !pInfo->variables->Set(peerNameVariable.c_str(), peer.get_host_name()))
   {
      pInfo->ivLog->Print("      ERROR: Failed to set peer name variable '%s'\n", responseVariable.c_str());
      return(IVR_FATAL);
   }

   if(!pInfo->variables->IsEmpty(peerPortVariable.c_str()))
   {
      char number[14];

      sprintf(number, "%d", peer.get_port_number());

      if(!pInfo->variables->Set(peerPortVariable.c_str(), number))
      {
         pInfo->ivLog->Print("      ERROR: Failed to set peer port variable '%s'\n", peerPortVariable.c_str());
         return(IVR_FATAL);
      }
   }

   return(IVR_SUCCESS);
 }
