#include <io.h>
#include <exception>
using std::exception;

#include "..\escutil\iv_str.h"
#include "..\BoardDLLs\brd_generic.h"
#include "TFEngine_communication.h"
#include "step_sendemail.h"

StepSendEmail::StepSendEmail(STEPREC& stepRecord) : Step(stepRecord)
{
   server = stepRecord.s1.arr;
   account = stepRecord.s2.arr;
   password = stepRecord.s3.arr;
   authenticationType = (StepSendEmail::AUTHENTICATION_TYPE) stepRecord.u1;
   fromAddress = stepRecord.s4.arr;
   recipients = stepRecord.s5.arr;
   subject = stepRecord.s6.arr;
   body = stepRecord.s7.arr;
   attachments = stepRecord.s8.arr;
   port = stepRecord.s9.arr;
   securityType = (StepSendEmail::SECURITY_TYPE) stepRecord.u2;
}

StepSendEmail::~StepSendEmail(void)
{
}

const char* StepSendEmail::GetStepName(void) const
{
	return("Send E-Mail");
}

string StepSendEmail::EncodeQuotedPrintable(string text)
{
   const char HexTable[] = "0123456789ABCDEF";
   string encoding;

   for(string::iterator current = text.begin(); current < text.end(); current++)
   {
      if((*current < ' ') || (*current == '=') || (*current == '?') || (*current > '~'))
      {
         encoding += "=";
         encoding += HexTable[*current/16];
         encoding += HexTable[*current%16];
      }
      else
      {
         encoding += *current;
      }
   }

   return(encoding);
}

bool StepSendEmail::NeedsMIME(string& text)
{
   bool needsMIME = false;

   // Look for characters requiring MIME
   for(string::iterator current = text.begin(); current < text.end(); current++)
   {
      if((*current < ' ') || (*current == '=') || (*current == '?') || (*current > '~'))
      {
         needsMIME = true;
      }
   }

   if(needsMIME)
   {
      text = "=?ISO-8859-1?Q?" + EncodeQuotedPrintable(text) + "?=";
   }

    return(needsMIME);
}

string StepSendEmail::RemoveComments(string list)
{
   string::iterator current;
   string newList;

   for(current = list.begin(); current < list.end(); current++)
   {
      switch(*current)
      {
         // Start parenthesis
         case '(':
         {
            // Skip comments to end parenthesis
            while((++current < list.end()) && (*current != ')'));
            break;
         }

         // Start quote
         case '\"':
         {
            newList += '\"';

            // Retain everything in quotes
            while((++current < list.end()) && (*current != '\"'))
            {
               newList += *current;
            }

            newList += '\"';

            break;
         }

         default:
         {
            newList += *current;
         }
      }
   }

   return(newList);
}

string StepSendEmail::CleanUpAddressList(string list)
{
   string::iterator current;
   string newList;

   // Eliminate comments in parentheses or quotes, and white space
   // Replace semi-colons with commas
   // Use addresses in angle brackets
   for(current = list.begin(); current < list.end(); current++)
   {
      switch(*current)
      {
         // Start parenthesis
         case '(':
         {
            // Skip comments to end parenthesis
            while((++current < list.end()) && (*current != ')'));
            break;
         }

         // Start quote
         case '\"':
         {
            // Skip comments to end quote
            while((++current < list.end()) && (*current != '\"'));
            break;
         }

         // Start angle bracket
         case '<':
         {
            // Find start of this address
            string::size_type addressStart = newList.rfind(',');

            if(addressStart == string::npos)
            {
               newList.clear();
            }
            else
            {
               newList = newList.substr(0, addressStart + 1);
            }

            // Copy address in angle brackets
            while((++current < list.end()) && (*current != '>'))
            {
               newList += *current;
            }

            // Find end of address
            while((++current < list.end()) && (*current != ',') && (*current != ';'));

            if((current < list.end()) && ((*current == ',') || (*current == ';')))
            {
               newList += ',';
            }

            break;
         }

         // Replace semi-colon with comma
         case ';':
         {
            newList += ',';
            break;
         }

         // Skip white space
         case ' ':
         case '\t':
         case '\n':
         {
            break;
         }

         default:
         {
            newList += *current;
         }
      }
   }

   // Find last (if any) trailing separator (',')
   for(current = newList.end() - 1; (current >= newList.begin()) && (*current == ','); current--);

   if(current < newList.begin())
   {
      return("");
   }
   else
   {
      return(newList.substr(0, current - newList.begin() + 1));
   }
}

string StepSendEmail::EncodeBase64(const char* text, unsigned int length)
{
   char Base64table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
   string encoding = "\r\n";
   const char* current;
   unsigned int convCount = 0;

   for(current = text; current < text + length - 2; current += 3)
   {
      encoding += Base64table[(*current >> 2) & 0x3F];
      encoding += Base64table[((*current << 4) & 0x30) | ((*(current + 1) >> 4) & 0x0F)];
      encoding += Base64table[((*(current + 1) << 2) & 0x3C) | ((*(current + 2) >> 6) & 0x03)];
      encoding += Base64table[*(current + 2) & 0x3F];

      convCount = convCount + 4;

      if(convCount == 72)
      {
         encoding += "\r\n";
         convCount = 0;
      }
   }

   if(current == text + length - 1)
   {
      encoding += Base64table[(*current >> 2) & 0x3F];
      encoding += Base64table[(*current << 4) & 0x30];
      encoding += "==";
   }
   else if(current == text + length - 2)
   {
      encoding += Base64table[(*current >> 2) & 0x3F];
      encoding += Base64table[((*current << 4) & 0x30) | ((*(current + 1) >> 4) & 0x0F)];
      encoding += Base64table[((*(current + 1) << 2) & 0x3C) & 0x3F];
      encoding += "=";
   }

   return(encoding.substr(2));
}

string StepSendEmail::EncodeBase64(FILE* file)
{
   char Base64table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
   string encoding = "";
   char buffer[3];
   unsigned int length;
   unsigned int convCount = 0;

   fseek(file, 0, SEEK_SET);

   for(length = fread(buffer, 1, 3, file); length == 3; length = fread(buffer, 1, 3, file))
   {
      encoding += Base64table[(buffer[0] >> 2) & 0x3F];
      encoding += Base64table[((buffer[0] << 4) & 0x30) | ((buffer[1] >> 4) & 0x0F)];
      encoding += Base64table[((buffer[1] << 2) & 0x3C) | ((buffer[2] >> 6) & 0x03)];
      encoding += Base64table[buffer[2] & 0x3F];

      convCount = convCount + 4;

      if(convCount == 72)
      {
         encoding += "\r\n";
         convCount = 0;
      }
   }

   if(length == 1)
   {
      encoding += Base64table[(buffer[0] >> 2) & 0x3F];
      encoding += Base64table[(buffer[0] << 4) & 0x30];
      encoding += "==";
   }
   else if(length == 2)
   {
      encoding += Base64table[(buffer[0] >> 2) & 0x3F];
      encoding += Base64table[((buffer[0] << 4) & 0x30) | ((buffer[1] >> 4) & 0x0F)];
      encoding += Base64table[((buffer[1] << 2) & 0x3C) & 0x3F];
      encoding += "=";
   }

   return(encoding);
}

string StepSendEmail::GenerateBoundary(void)
{
   char AlphaNumerics[63] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
   string boundary;

   // Generate a random 20 character message boundary identifier
   // Seed random number generator with current time
   srand((unsigned int) time(NULL));

   for(unsigned int counter = 0; counter < 20; counter++)
   {
      // Add a random character
      boundary += AlphaNumerics[rand()%62];
   }

   return(boundary);
}

IV_STATUS StepSendEmail::GetSMPTPReply(PIVRINFO pInfo)
{
   IV_STATUS status;
   int ret;

   smtp->Response()->ClearText();

   while(!smtp->Response()->HaveCompleteMessage())
   {
      switch(WaitForMultipleObjects(2, waitObjs, false, INFINITE))
      {
         // System event
         case WAIT_OBJECT_0:
         {
            status = pInfo->ivBoard->ProcessSystemEvent();

            if(status != IVR_SUCCESS)
            {
               return(status);
            }

            break;
         }

         // Socket signalled
         case WAIT_OBJECT_0 + 1:
         {
            if((ret = smtp->Receive()) == WSAEWOULDBLOCK)
            {
               // Allow 10 ms for socket to clear
               Sleep(10);
            }
            else if(ret != Socket::SUCCESS)
            {
               pInfo->ivLog->Print("      ERROR: Failed to receive response\n");
               pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
               return(IVR_FAIL);
            }

            break;
         }

         // WaitForMultipleObjects failed
         case WAIT_FAILED:
         {
            DWORD errorID = GetLastError();
            char MsgBuf[1024];

            FormatMessage(
               FORMAT_MESSAGE_FROM_SYSTEM,
               0,
               errorID,
               MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
               MsgBuf,
               1024,
               0
            );

            pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
            return(IVR_FAIL);
         }
      }
   }

   pInfo->ivLog->Print("      '%s'\n", smtp->Response()->GetPlainText());

   return(IVR_SUCCESS);
}

IV_STATUS StepSendEmail::SendSMPTPMessage(PIVRINFO pInfo)
{
   IV_STATUS status;
   int ret;

   while((ret = smtp->Send()) == WSAEWOULDBLOCK)
   {
      switch(WaitForMultipleObjects(1, waitObjs, false, 10))
      {
         // System event
         case WAIT_OBJECT_0:
         {
            status = pInfo->ivBoard->ProcessSystemEvent();

            if(status != IVR_SUCCESS)
            {
               return(status);
            }

            break;
         }

         case WAIT_TIMEOUT:
         {
            break;
         }

         // WaitForMultipleObjects failed
         case WAIT_FAILED:
         {
            DWORD errorID = GetLastError();
            char MsgBuf[1024];

            FormatMessage(
               FORMAT_MESSAGE_FROM_SYSTEM,
               0,
               errorID,
               MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
               MsgBuf,
               1024,
               0
            );

            pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
            return(IVR_FAIL);
         }
      }
   }

   return(IVR_SUCCESS);
}

IV_STATUS StepSendEmail::Execute(PIVRINFO pInfo)
{
   char* value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   bool mime = false;
   string parsedServer;
   unsigned short parsedPort;
   string message;
   string parsedAccount;
   string parsedPassword;
   string parsedFromAddress;
   string boundary;
   string parsedRecipients;
   string parsedBody;
   string parsedAttachments;
   ESCTime curtime;
   TIME_ZONE_INFORMATION tzinfo;
   DWORD tz = GetTimeZoneInformation(&tzinfo);
   LONG hours = (tzinfo.Bias + (tz == TIME_ZONE_ID_STANDARD ? tzinfo.StandardBias : tzinfo.DaylightBias))/60;
   LONG minutes = (tzinfo.Bias + (tz == TIME_ZONE_ID_STANDARD ? tzinfo.StandardBias : tzinfo.DaylightBias))%60;
   char temp[128];
   IV_STATUS status;
   unsigned int connectionRetries = 2;
   unsigned int connectCounter = 0;
   bool loopAgain;
   bool attemptConnect;
   int ret;
   SMTPTransaction currentSMTP(pInfo->ivLog);

   smtp = &currentSMTP;

   // Server
   if((value = pInfo->variables->ParseParameter("Mail server", server.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedServer = value;
      delete[] value;
   }

   // Port
   if((value = pInfo->variables->ParseParameter("Mail server port", port.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedPort = (unsigned short) atoi(value);
      delete[] value;
      if (parsedPort == 0)
      {
         // Default values for port:  25 for insecure, 465 for SSL/TLS
         parsedPort = (securityType == SECURITY_SSL) ? 465 : 25;
         pInfo->ivLog->Print("      Defaulting mail server port to %u\n", parsedPort);
      }
   }

	// Date: specified in RFC 2822
   // e.g. "Date: Fri, 26 Nov 2004 11:38:11 -0800\r\n"
   curtime.Update();
   message = (string) "Date: " + (string) curtime.GetShortDayOfWeekName();
   sprintf(temp, ", %.2d ", curtime.GetDay());
   message += temp;
   message += (string) curtime.GetShortMonthName();
   sprintf(temp, " %.4d %.2d:%.2d:%.2d %+03d%02d", curtime.GetYear(), curtime.GetHour(),
      curtime.GetMinute(), curtime.GetSecond(), -hours, -minutes);
   message += temp;
   message += "\r\n";

   // Sender: specified in RFC 2822
   if((value = pInfo->variables->ParseParameter("Sender's e-mail address", account.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedAccount = value;
      delete[] value;

      if(!parsedAccount.empty())
      {
         mime |= NeedsMIME(parsedAccount);
         message += "Sender: " + parsedAccount + "\r\n";
      }
   }
   
   if((authenticationType == AUTHENTICATION_LOGIN) || (authenticationType == AUTHENTICATION_PLAIN))
   {
      if(parsedAccount.empty())
      {
         pInfo->ivLog->Print("      ERROR: Sender's password '%s' cannot be empty when using authenticated SMTP\n", password.c_str());
         bParmsValid = false;
      }

      // Password
      if((value = pInfo->variables->ParseParameter("Sender's password", password.c_str(), "")) == NULL)
      {
         bParmsValid = false;
      }
      else
      {
         parsedPassword = value;
         delete[] value;

         if(parsedPassword.empty())
         {
            pInfo->ivLog->Print("      ERROR: Sender's password '%s' cannot be empty when using authenticated SMTP\n", password.c_str());
            bParmsValid = false;
         }
      }
   }

   // From: specified in RFC 2822
   // Reply-To: specified in RFC 2822
   if((value = pInfo->variables->ParseParameter("From address", fromAddress.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedFromAddress = value;
      delete[] value;

      parsedFromAddress = RemoveComments(parsedFromAddress);
      mime |= NeedsMIME(parsedFromAddress);
      message += "From: " + parsedFromAddress + "\r\n";
      message += "Reply-To: " + parsedFromAddress + "\r\n";

      parsedFromAddress = CleanUpAddressList(parsedFromAddress);

      if((value = FindDelimitedElement(parsedFromAddress.c_str(), 0, 2, ",", true, true)) != NULL)
      {
         delete[] value;
         pInfo->ivLog->Print("      ERROR: From address '%s' cannot contain more than 1 address\n", fromAddress.c_str());
         bParmsValid = false;
      }
   }

   // To: specified in RFC 2822
   if((value = pInfo->variables->ParseParameter("Recipient list", recipients.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedRecipients = value;
      delete[] value;

      parsedRecipients = RemoveComments(parsedRecipients);
      mime |= NeedsMIME(parsedRecipients);
      message += "To: " + parsedRecipients + "\r\n";
   }

   // Subject: specified in RFC 2822
   if((value = pInfo->variables->ParseParameter("Subject line", subject.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      string parsedSubject = value;
      delete[] value;

      mime |= NeedsMIME(parsedSubject);
      message += "Subject: " + parsedSubject + "\r\n";
   }

   message += "X-Mailer: enGenic TeleFlow Send E-Mail step, http://www.enGenic.com\r\n";

   // Body
   if((value = pInfo->variables->ParseParameter("Body", body.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      string::size_type index1;
      string::size_type index2;

      parsedBody = value;
      delete[] value;

      // Wrap lines longer than 78 characters
      for(index1 = 0, index2 = parsedBody.find("\r\n"); index2 != string::npos; index2 = parsedBody.find("\r\n", index1))
      {
         // Line is too long
         if(index2 - index1 > 78)
         {
            // Insert break
            parsedBody.insert(index1 + 78, "\r\n");
            // Go to new end of line
            index1 += 80;
         }
         // Line is short enough
         else
         {
            // Go to end of line
            index1 = index2 + 2;
         }
      }

      // Replace any "\r\n.\r\n" sequence with "\r\n..\r\n"
      for(index1 = parsedBody.find("\r\n.\r\n"); index1 != string::npos; index1 = parsedBody.find("\r\n.\r\n", index1 + 4))
      {
            parsedBody.insert(index1 + 2, ".");
      }
   }

   // Create body header
   if((value = pInfo->variables->ParseParameter("Attachments", attachments.c_str(), "")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedAttachments = value;
      delete[] value;
   }

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }

   if(!parsedAttachments.empty())
   {
      // Indicate MIME version and type
      boundary =  "enGenic_TeleFlow_Send_E-mail_Step-" + GenerateBoundary();
      message +=  "MIME-Version: 1.0\r\n"
                  "Content-Type: Multipart/Mixed; boundary=\"" + boundary + "\"\r\n"
                  "--" + boundary + "\r\n"
                  "Content-Type: text/plain; charset=US-ASCII\r\n"
                  "Content-Transfer-Encoding: " + (mime ? "quoted-printable" : "7BIT\r\n") +
                  "Content-description: Mail message body\r\n"
                  "\r\n"
                  + parsedBody + "\r\n";
   }
   else if(mime)
   {
      // Indicate MIME version and type
      message +=  "MIME-Version: 1.0\r\n"
                  "Content-Type: text/plain; charset=US-ASCII\r\n"
                  "Content-Transfer-Encoding: quoted-printable\r\n"
                  "Content-description: Mail message body\r\n"
                  "\r\n"
                  + parsedBody + "\r\n";
   }
   // No body header needed
   else
   {
      message += parsedBody + "\r\n";
   }

   // Add any attachments
   if(!parsedAttachments.empty())
   {
      parsedAttachments = strrepl(parsedAttachments.c_str(), ";", ",");

      message += "\r\n";

      for(unsigned int pos = 1; (value = FindDelimitedElement(parsedAttachments.c_str(), 0, pos, ",", true, true)) != NULL; pos++)
      {
         string attachFile;
         string attachFilePath;
         string FileNameOnly;
         FILE* file;

         strtrim_left(value);
         strtrim_right(value);
         attachFile = value;
         delete[] value;
         attachFilePath = value = pInfo->variables->FilePath(attachFile.c_str());
         delete[] value;

         file = fopen(attachFilePath.c_str(), "rb");

         if(file == NULL)
         {
            pInfo->ivLog->Print("      ERROR: Attachment '%s' cannot be opened\n", attachFilePath.c_str());
            return(IVR_FAIL);
         }

         // Write header
         FileNameOnly = attachFilePath.substr(attachFilePath.find_last_of('\\') + 1);
         message +=  "--" + boundary + "\r\n"
                     "Content-Type: " + DetermineFileType(FileNameOnly) + "; name=" + FileNameOnly + "\r\n"
                     "Content-Disposition: attachment; filename=\"" + FileNameOnly + "\"\r\n"
                     "Content-Transfer-Encoding: BASE64\r\n"
                     "\r\n";

         message += EncodeBase64(file) + "\r\n\r\n";

         fclose(file);
      }
   }

   // End of message token
   message += "\r\n.\r\n";

// SMTP transmission sequence from RFC2821
// S: SUCCESS
// E: ERROR
// I: INCOMPLETE
// CONNECTION ESTABLISHMENT
//    S: 220
//    E: 554
// EHLO or HELO
//    S: 250
//    E: 504, 550
// MAIL
//    S: 250
//    E: 552, 451, 452, 550, 553, 503
// RCPT
//    S: 250, 251 (but see section 3.4 for discussion of 251 and 551)
//    E: 550, 551, 552, 553, 450, 451, 452, 503, 550
// DATA
//    I: 354 -> data -> S: 250
//                      E: 552, 554, 451, 452
//    E: 451, 554, 503
// QUIT
//    S: 221

   // Set SMTP Server, port, and security for connection
   smtp->SetPeerPort(parsedPort);
   smtp->SetPeerName(parsedServer.c_str());
   if (securityType == SECURITY_SSL)
   {
      smtp->EnableSSL();
   }

   // Set up events array
   waitObjs[0] = pInfo->messages->GetMessageSignal();  // system event such as shutdown
   waitObjs[1] = smtp->GetEventHandle();                // signal from socket

   // Try up to connectionRetries times to connect
   loopAgain = true;
   attemptConnect = true;
   while(loopAgain)
   {
      // Connect to SMTP server
      if(attemptConnect)
      {
         pInfo->ivLog->Print("      Connection attempt %d\n", ++connectCounter);

         if((ret = smtp->Connect()) != Socket::SUCCESS)
         {
            pInfo->ivLog->Print("      ERROR: Failed to connect to host\n");
            pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
            return(IVR_FAIL);
         }
      }

      switch(WaitForMultipleObjects(2, waitObjs, false, INFINITE))
      {
         // System event
         case WAIT_OBJECT_0:
         {
            status = pInfo->ivBoard->ProcessSystemEvent();

            if(status != IVR_SUCCESS)
            {
               return(status);
            }

            attemptConnect = false;
            break;
         }

         // Connection signalled
         case WAIT_OBJECT_0 + 1:
         {
            // Check status
            if((ret = smtp->GetStatus()) == Socket::SUCCESS)
            {
               pInfo->ivLog->Print("      Connection successful\n");

               // Enable receipt of all possible events
               smtp->SelectEvents(FD_READ|FD_WRITE|FD_OOB|FD_ACCEPT|FD_CONNECT|FD_CLOSE|FD_QOS|FD_GROUP_QOS);

               loopAgain = false;
            }
            else
            {
               pInfo->ivLog->Print("      Connection attempt failed\n");
               pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));

               if(connectCounter > connectionRetries)
               {
                  return(IVR_FAIL);
               }

               smtp->Close();
               attemptConnect = true;
            }

            break;
         }

         // WaitForMultipleObjects failed
         case WAIT_FAILED:
         {
            DWORD errorID = GetLastError();
            char MsgBuf[1024];

            FormatMessage(
               FORMAT_MESSAGE_FROM_SYSTEM,
               0,
               errorID,
               MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
               MsgBuf,
               1024,
               0
            );

            pInfo->ivLog->Print("      ERROR: WaitForMultipleObjects returned: %x: %s", errorID, MsgBuf);
            return(IVR_FAIL);
         }
      }
   }

   if(smtp->IsSSLEnabled())
   {
      if((ret = smtp->SSLHandshake()) == Socket::SUCCESS)
      {
         pInfo->ivLog->Print("      SSL Handshake successful\n");
      }
      else
      {
         const char* errorMessage = Socket::WSAErrorString(ret);

         pInfo->ivLog->Print("      ERROR: Unable to complete SSL handshake\n");
         pInfo->ivLog->Print("         '%s'\n", errorMessage);
         return(IVR_FAIL);
      }
   }
      
   if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
   {
      return(status);
   }
   else if(smtp->Response()->GetStatusCode() != 220)
   {
      pInfo->ivLog->Print("      ERROR: Mail server refused connection\n");
      pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
      return(IVR_FAIL);
   }

   // Handshake with SMTP server
   smtp->Request()->SetCommand("EHLO");
   smtp->Request()->SetArguments(Socket::GetHostName());

   if((ret = smtp->Send()) != Socket::SUCCESS)
   {
      pInfo->ivLog->Print("      ERROR: Failed to handshake with mail server\n");
      pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
      return(IVR_FAIL);
   }

   if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
   {
      return(status);
   }
   else if(smtp->Response()->GetStatusCode() != 250)
   {
      // Alternate handshake with SMTP server
      smtp->Request()->SetCommand("HELO");

      if((ret = smtp->Send()) != Socket::SUCCESS)
      {
         pInfo->ivLog->Print("      ERROR: Failed to handshake with mail server\n");
         pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
         return(IVR_FAIL);
      }

      if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
      {
         return(status);
      }
      else if(smtp->Response()->GetStatusCode() != 250)
      {
         pInfo->ivLog->Print("      ERROR: Mail server refused handshake\n");
         pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
         return(IVR_FAIL);
      }
   }

   // Login to SMTP server
   switch(authenticationType)
   {
      case AUTHENTICATION_LOGIN:
      {
         pInfo->ivLog->Print("      Logging in to mail server with LOGIN authentication\n");

         smtp->Request()->SetCommand("AUTH");
         smtp->Request()->SetArguments("LOGIN");

         if((ret = smtp->Send()) != Socket::SUCCESS)
         {
            pInfo->ivLog->Print("      ERROR: Failed to login to mail server\n");
            pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
            return(IVR_FAIL);
         }

         if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
         {
            return(status);
         }
         else if(smtp->Response()->GetStatusCode() != 334)
         {
            pInfo->ivLog->Print("      ERROR: Mail server refused login request\n");
            pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
            return(IVR_FAIL);
         }

         smtp->Request()->SetText((EncodeBase64(parsedAccount.c_str(), parsedAccount.length()) + (string) "\r\n").c_str());

         if((ret = smtp->Send()) != Socket::SUCCESS)
         {
            pInfo->ivLog->Print("      ERROR: Failed to login to mail server\n");
            pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
            return(IVR_FAIL);
         }

         if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
         {
            return(status);
         }
         else if(smtp->Response()->GetStatusCode() != 334)
         {
            pInfo->ivLog->Print("      ERROR: Mail server refused login account name\n");
            pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
            return(IVR_FAIL);
         }

         smtp->Request()->SetText((EncodeBase64(parsedPassword.c_str(), parsedPassword.length()) + (string) "\r\n").c_str());

         if((ret = smtp->Send()) != Socket::SUCCESS)
         {
            pInfo->ivLog->Print("      ERROR: Failed to login to mail server\n");
            pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
            return(IVR_FAIL);
         }

         if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
         {
            return(status);
         }
         else if(smtp->Response()->GetStatusCode() / 100 != 2)
         {
            pInfo->ivLog->Print("      ERROR: Mail server refused account password\n");
            pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
            return(IVR_FAIL);
         }

         break;
      }

      case AUTHENTICATION_PLAIN:
      {
         char* buffer = new char[parsedAccount.length() + parsedPassword.length() + 3];

         pInfo->ivLog->Print("      Logging in to mail server with PLAIN authentication\n");

         buffer[0] = '\0';
         sprintf(buffer + 1, "%s", parsedAccount.c_str());
         sprintf(buffer + parsedAccount.length() + 2, parsedPassword.c_str());

         smtp->Request()->SetCommand("AUTH");
         smtp->Request()->SetArguments(((string) "PLAIN " + EncodeBase64(buffer, parsedAccount.length() + parsedPassword.length() + 3)).c_str());
         delete[] buffer;

         if((ret = smtp->Send()) != Socket::SUCCESS)
         {
            pInfo->ivLog->Print("      ERROR: Failed to login to mail server\n");
            pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
            return(IVR_FAIL);
         }

         if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
         {
            return(status);
         }
         else if(smtp->Response()->GetStatusCode() / 100 != 2)
         {
            pInfo->ivLog->Print("      ERROR: Mail server refused account name or password\n");
            pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
            return(IVR_FAIL);
         }

         break;
      }

      case AUTHENTICATION_NONE:
      {
         // No action required
         break;
      }
   }

   // Set mail sender
   smtp->Request()->SetCommand("MAIL");
   smtp->Request()->SetArguments(((string) "FROM:<" + parsedFromAddress + (string) ">").c_str());

   if((ret = smtp->Send()) != Socket::SUCCESS)
   {
      pInfo->ivLog->Print("      ERROR: Failed to send MAIL FROM to mail server\n");
      pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
      return(IVR_FAIL);
   }

   if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
   {
      return(status);
   }
   else if(smtp->Response()->GetStatusCode() != 250)
   {
      pInfo->ivLog->Print("      ERROR: Mail server refused MAIL FROM\n");
      pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
      return(IVR_FAIL);
   }

   // Initiate mail recipient list
   parsedRecipients = CleanUpAddressList(parsedRecipients);

   smtp->Request()->SetCommand("RCPT");
   // Specify mail recipients
   for(unsigned int pos = 1; (value = FindDelimitedElement(parsedRecipients.c_str(), 0, pos, ",", true, true)) != NULL; pos++)
   {
      smtp->Request()->SetArguments(((string) "TO:<" + value + (string) ">").c_str());
      delete[] value;

      if((ret = smtp->Send()) != Socket::SUCCESS)
      {
         pInfo->ivLog->Print("      ERROR: Failed to send RCPT TO to mail server\n");
         pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
         return(IVR_FAIL);
      }

      if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
      {
         return(status);
      }
      else if(smtp->Response()->GetStatusCode() / 100 != 2)
      {
         pInfo->ivLog->Print("      ERROR: Mail server refused RCPT TO\n");
         pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
         return(IVR_FAIL);
      }
   }

   // Initiate message data
   smtp->Request()->SetCommand("DATA");
   smtp->Request()->SetArguments("");

   if((ret = smtp->Send()) != Socket::SUCCESS)
   {
      pInfo->ivLog->Print("      ERROR: Failed to send DATA to mail server\n");
      pInfo->ivLog->Print("         '%s'\n", Socket::WSAErrorString(ret));
      return(IVR_FAIL);
   }

   if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
   {
      return(status);
   }
   else if(smtp->Response()->GetStatusCode() != 354)
   {
      pInfo->ivLog->Print("      ERROR: Mail server refused DATA\n");
      pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
      return(IVR_FAIL);
   }

   // Send message data
   smtp->Request()->SetText(message.c_str());

   // Sending the data may take a long time, so we keep checking to make sure the message data is fully sent before waiting for a reply
   if((status = SendSMPTPMessage(pInfo)) != IVR_SUCCESS)
   {
      return(status);
   }

   if((status = GetSMPTPReply(pInfo)) != IVR_SUCCESS)
   {
      return(status);
   }
   else if(smtp->Response()->GetStatusCode() != 250)
   {
      pInfo->ivLog->Print("      ERROR: Mail server refused to send e-mail\n");
      pInfo->ivLog->Print("         '%s'\n", smtp->Response()->GetText());
      return(IVR_FAIL);
   }

   return(IVR_SUCCESS);
}

string StepSendEmail::DetermineFileType(string FileName)
{
   char value [40];
   unsigned long lTypeSize;
   LONG lResult;
   HKEY key=NULL;
   unsigned long lStringType = REG_SZ;
   string extension;
   string type = "application/octet-stream";    // default
   size_t dotfound;

   dotfound = FileName.find_last_of('.');
   if(dotfound!=string::npos)
   {
      extension = FileName.substr(dotfound);
      if(!extension.empty())
      {
         lResult = RegOpenKeyEx(HKEY_CLASSES_ROOT, (LPCTSTR)(extension.c_str()), 0, KEY_READ, &key);
         if(lResult == ERROR_SUCCESS)
         {
            lTypeSize = (long) sizeof(value);
            lResult = RegQueryValueEx(key, "Content Type", NULL, &lStringType, (unsigned char*)value, &lTypeSize);
            RegCloseKey (key);
         }
         if(lResult == ERROR_SUCCESS)
         {
            type = value;
         }
         // Registry is missing information, try a few hard-coded settings
         else if(stricmp(extension.c_str(), ".pdf") == 0)
         {
            type = "application/pdf";
         }
         else if(stricmp(extension.c_str(), ".wav") == 0)
         {
            type = "audio/wav";
         }
         else if(stricmp(extension.c_str(), ".tap") == 0)
         {
            type = "application/TeleFlow";
         }
         else if(stricmp(extension.c_str(), ".tam") == 0)
         {
            type = "application/TeleFlow";
         }
      }
   }
   return(type);
}

