#include <math.h>             // Needed for "ceil".
#include "step_vocpassreg.h"
#include "..\escutil\iv_str.h"
#include "..\BoardDLLs\brd_generic.h"
#include "..\Voice\Voice.h"
#include "VoiceCreator.h"

StepVocPassReg::StepVocPassReg(STEPREC& stepRecord) : Step(stepRecord)
{
	userID = stepRecord.s1.arr;
	playFile1 = stepRecord.s2.arr;
	playFile2 = stepRecord.s3.arr;
	playFile3 = stepRecord.s4.arr;
	regPercent = stepRecord.s5.arr;
   maxTime = stepRecord.s6.arr;
   disableBeep = (bool) stepRecord.u1;
}

StepVocPassReg::~StepVocPassReg(void)
{
}

const char* StepVocPassReg::GetStepName(void) const
{
	return("Voice Pass Register");
}

// -------------------------------------------------------------------------
//   Name:        StepVocPassReg::Execute
//
//   Notes:       QUESTION: Should we have just one audio buffer and re-use
//                  for each utterance? (Note: VoiceCop's sample app uses
//                  three)
//                ANSWER: Keep the three-buffer method, because if one of the
//                  buffers is deemed to be inconsistent (unacceptable), it
//                  can be re-played and/or re-recorded without losing the
//                  data collected in the other two utterances.
// -------------------------------------------------------------------------
IV_STATUS StepVocPassReg::Execute(PIVRINFO pInfo)
{
   IV_STATUS 	 status; // status returned from this module

   // To evaluate parameters from the Step
   string     	 parsedUserID;					  // parsed version of user ID
   char* value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   char* scriptCodeFileName;
   int       	 regpercent;			  // registration tolerance percent 1 = tightest, 100 = loosest
   int		 	 iMaxTime;             // target maximum record length in milliseconds at step-level
                                      //  - this is the value supplied by the TeleFlow Designer
                                      //  - will get rounded to whole seconds before
                                      //    getting passed to board-level because
                                      //    VoiceCop expects value in seconds only
   SOUND_FORMAT format;

   // To perform voice print registration
   char* audioBuf[3];  				  // Buffer currently being processed
   string playfile[3];        		  // File prompt currently being played (shallow copy)
	unsigned int		bufsize;					     // size of audio buffer
   char*	cfgPath = pInfo->variables->FilePath("cfg\\eng\\");	// path to VoiceCop configuration files
   char* modelPath = pInfo->variables->FilePath("model\\");    // path to VoiceCop model files

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

	if(PortRequired(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FATAL);
   }

   // VoiceVerify will only function on licensed TeleFlow
   if(pInfo->portLicense == UNLICENSED)
   {
      pInfo->ivLog->Print ("      THIS PORT IS UNLICENSED\n");
      return(IVR_NOTSUPPORTED);
   }

	// Parse User ID field
   if((value = pInfo->variables->ParseParameter("User ID", userID.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedUserID = value;
      delete[] value;
   }

	// Process first script code
   if((value = pInfo->variables->ParseParameter("First script code", playFile1.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      status = GetScriptCodeFileNameAndFormat(pInfo, value, scriptCodeFileName, format);
      delete[] value;

      if(status != IVR_SUCCESS)
      {
         bParmsValid = false;
      }
      else
      {
         playfile[0] = value = pInfo->variables->FilePath(scriptCodeFileName);
         delete[] value;
         delete[] scriptCodeFileName;
         pInfo->ivLog->Print("      First script code path is '%s'\n", playfile[0].c_str());
      }
   }

	// Process second script code
   if((value = pInfo->variables->ParseParameter("Second script code", playFile2.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      status = GetScriptCodeFileNameAndFormat(pInfo, value, scriptCodeFileName, format);
      delete[] value;

      if(status != IVR_SUCCESS)
      {
         bParmsValid = false;
      }
      else
      {
         playfile[1] = value = pInfo->variables->FilePath(scriptCodeFileName);
         delete[] value;
         delete[] scriptCodeFileName;
         pInfo->ivLog->Print("      Second script code path is '%s'\n", playfile[1].c_str());
      }
   }

	// Process Third script code
   if((value = pInfo->variables->ParseParameter("Third script code", playFile3.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      status = GetScriptCodeFileNameAndFormat(pInfo, value, scriptCodeFileName, format);
      delete[] value;

      if(status != IVR_SUCCESS)
      {
         bParmsValid = false;
      }
      else
      {
         playfile[2] = value = pInfo->variables->FilePath(scriptCodeFileName);
         delete[] value;
         delete[] scriptCodeFileName;
         pInfo->ivLog->Print("      Third script code path is '%s'\n", playfile[2].c_str());
      }
   }

	// Parse registration factor
   if((value = pInfo->variables->ParseParameter("Registration Factor", regPercent.c_str(), "50")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
	   regpercent = atoi(value);
      delete[] value;

      if(regpercent < 1)
      {
	      pInfo->ivLog->Print("      ERROR: Registration factor below minimum of 1\n");
         bParmsValid = false;
      }
      else if(regpercent > 100)
      {
	      pInfo->ivLog->Print("      ERROR: Registration factor above maximum of 100\n");
         bParmsValid = false;
      }
   }

   // Parse maximum record time
   if((value = pInfo->variables->ParseParameter("Maximum record time", maxTime.c_str(), "4000")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iMaxTime = atoi(value);
      delete[] value;

      iMaxTime = (iMaxTime/1000 + (iMaxTime % 1000 > 0) ? 1 : 0)*1000;   // Rounded to the nearest second for VoiceCop
      pInfo->ivLog->Print("      Record time rounded to nearest second: %i ms\n", iMaxTime);
   }

   if(!bParmsValid)
   {
      return(IVR_FATAL);
   }

   // Create new voice module if needed
   if(pInfo->voiceAuthenticationLibrary == NULL)
   {
      HINSTANCE library = GetModuleHandle("Voice_enGenic.dll");
      PVoiceCreator voiceCreator;

      if(library == NULL)
      {
         library = LoadLibrary("Voice_enGenic.dll");
      }

      voiceCreator = GetVoiceCreator(library);

      if(voiceCreator == NULL)
      {
         pInfo->ivLog->Print("      ERROR: Unable to access '%s'\n", "Voice_enGenic.dll");
         pInfo->ivLog->Print("         This feature requires Voice_enGenic.dll and special licensing.\n"
                             "         Please contact enGenic for more information.\n");
         return(IVR_NOTSUPPORTED);
      }

      pInfo->voiceAuthentication = voiceCreator(pInfo->iAppID, pInfo->ivLog, pInfo->ivBoard);

      if(pInfo->voiceAuthentication == NULL)
      {
         pInfo->ivLog->Print("      ERROR: Unable to create voice module\n");
         return(IVR_FATAL);
      }

      pInfo->voiceAuthenticationLibrary = library;
   }

	pInfo->ivLog->Print("         Config path = '%s'\n", cfgPath);
	pInfo->ivLog->Print("         Model path  = '%s'\n", modelPath);

   format = TF_WAV_8KHZ_16BIT;
   
	bufsize = 8*2*iMaxTime;		// 8k samples per second, 2 byte samples

   if((status = pInfo->voiceAuthentication->VAStart(cfgPath, modelPath, bufsize)) != IVR_SUCCESS)
   {
      return(status);
   }

   // Create recording buffers
   audioBuf[0] = new char[bufsize];
   memset(audioBuf[0], 0, bufsize);
   audioBuf[1] = new char[bufsize];
   memset(audioBuf[1], 0, bufsize);
   audioBuf[2] = new char[bufsize];
   memset(audioBuf[2], 0, bufsize);

	// Prompt user and record, 3 times
   for(int iUtterance = 0; (iUtterance < 3) && (status == IVR_SUCCESS); iUtterance++)
   {
      status = pInfo->ivBoard->AudioPlay(playfile[iUtterance].c_str(),		// file to play
      						                 false,		// not interruptable
      						                 format,			// TeleFlow audio format, hardcoded to PCM8M16
      						                 0,			// normal volume (0)
      						                 false		// not fancy
				                            );

      if(status != IVR_SUCCESS)
      {
         // Release memory used up to this point
         delete[] audioBuf[0];
         delete[] audioBuf[1];
         delete[] audioBuf[2];
         return(status);
      }

		// Record user speech
		pInfo->ivLog->Print("         Record to buffer \n");
		status = pInfo->ivBoard->AudioRecordBuffer(audioBuf[iUtterance],             // buffer to record into
                                                 bufsize,            // buffer size (bytes)
                                                 false,              // not interruptable
                                                 TF_WAV_8KHZ_16BIT,  // SOUND_FORMAT format
                                                 0,             // int volLevel, 0 = normal
                                                 iMaxTime,
                                                 disableBeep,	      // bool disableBeep
                                                 false               // bool automaticGainControl
                                                );
	   if(status != IVR_SUCCESS)
	   {
			pInfo->ivLog->Print("         ERROR: Voiceprint recording, utterance %i, failed\n", iUtterance + 1);
         // Release memory used up to this point
         delete[] audioBuf[0];
         delete[] audioBuf[1];
         delete[] audioBuf[2];
	   	return(status);
	   }

      if((status = pInfo->voiceAuthentication->VASubmit(regpercent, audioBuf[iUtterance], bufsize)) != IVR_SUCCESS)
      {
         pInfo->voiceAuthentication->VAEnd();
         delete[] audioBuf[0];
         delete[] audioBuf[1];
         delete[] audioBuf[2];
         return(status);
      }
   } // end for(iUtterance <= 2)

   delete[] audioBuf[0];
   delete[] audioBuf[1];
   delete[] audioBuf[2];

   status = pInfo->voiceAuthentication->VARegister(parsedUserID.c_str());
   pInfo->voiceAuthentication->VAEnd();

   return(status);
}
