#ifndef COMMUNICATION_HUB_H
#define COMMUNICATION_HUB_H

#include <winsock2.h>
#include <windows.h>

#include <map>
using std::map;

#include "signalevt.h"

// Useless on its own, CommunicationPacket serves as the base class for message
// packets sent to a CommunicationHub and received by a CommunicationDestination
// Note that CommunicationPackets must be dynamically allocated by the sender and
// deallocated by the receiver
class CommunicationPacket
{
   public:

      CommunicationPacket(void)
      {
      };

      ~CommunicationPacket(void)
      {
      };
};

// Virtual base class for destination of CommunicationPacket sent to a CommunicationHub
// Note that CommunicationPackets must be dynamically allocated by the sender and
// deallocated by the receiver
template <class T> class CommunicationDestination
{
   public:

      CommunicationDestination(void)
      {
      };

      virtual ~CommunicationDestination(void)
      {
         mutex.Get();
      };

      // Returns the handle that is signalled when a message arrives
      virtual HANDLE GetMessageSignal(void)
      {
         HANDLE signal;

         mutex.Get();
         signal = messages.GetSignal();
         mutex.Release();

         return(signal);
      };

      // Returns the identier for the CommunicationDestination in a CommunicationHub
      // Note that IDs must be unique within a CommunicationHub
      virtual T ID(void) = 0;

      // Returns a pointer to first CommunicationPacket in the queue which must
      // then be cast to the appropriate derived class of CommunicationPacket
      // Returns NULL if there is no CommunicationPacket in the queue
      virtual CommunicationPacket* ReceiveMessage(void)
      {
         CommunicationPacket* message;

         mutex.Get();
         if(!messages.Get(message))
         {
            message = NULL;
         }
         mutex.Release();

         return(message);
      };

      // Adds a CommunicationPacket to the end of the queue
      virtual bool SendMessage(CommunicationPacket* message)
      {
         mutex.Get();
         messages.Add(message);
         mutex.Release();

         return(true);
      };

   protected:

      SignalQueue<CommunicationPacket*> messages;
      Mutex mutex;
};

// Base class for hub of CommunicationDestinations which receive CommunicationPackets
// Note that CommunicationPackets must be dynamically allocated by the sender and
// deallocated by the receiver
template <class T> class CommunicationHub
{
   public:

      CommunicationHub(void)
      {
      };

      ~CommunicationHub(void)
      {
         mutex.Get();
      };

      // Adds a CommunicationDestination to the hub's list
      // Note that the CommunicationDestination ID must be unique within the hub
      void Add(CommunicationDestination<T>* destination)
      {
         mutex.Get();
         if(destination != NULL)
         {
            destinations[destination->ID()] = destination;
         }
         mutex.Release();
      };

      // Returns a pointer to the CommunicationDestination identified by destinationID
      // Returns NULL if there is no CommunicationDestination identified by destinationID
      CommunicationDestination<T>* Get(T destinationID)
      {
         CommunicationDestination<T>* desiredDestination;

         mutex.Get();
         desiredDestination = destinations.count(destinationID) == 0 ? NULL : destinations[destinationID];

         mutex.Release();

         return(desiredDestination);
      };

      // Removes the CommunicationDestination identified by destinationID
      // Returns a pointer to the CommunicationDestination identified by destinationID
      // Returns NULL if there is no CommunicationDestination identified by destinationID
      CommunicationDestination<T>* Remove(T destinationID)
      {
         CommunicationDestination<T>* oldDestination;

         mutex.Get();
         oldDestination = Get(destinationID);
         destinations.erase(destinationID);
         mutex.Release();

         return(oldDestination);
      };

      // Passes the message to the destination
      // Returns true if there is a CommunicationDestination identified by destinationID
      // otherwise returns false
      bool SendMessage(T destinationID, CommunicationPacket* message)
      {
         bool result;

         mutex.Get();
         if(destinations.count(destinationID) == 0)
         {
            delete message;
            result = false;
         }
         else
         {
            result = destinations[destinationID]->SendMessage(message);
         }
         mutex.Release();

         return(result);
      };

   protected:

      map<T, CommunicationDestination<T>*> destinations;
      Mutex mutex;
};

#endif
