/* -------------------------------------------------------------------------
   File:        link.cpp

   Purpose:     Generic link list classes.

   Copyright (c) 1997-2001 enGenic. All rights reserved.
   ------------------------------------------------------------------------- */
#include "link.h"

/* -------------------------------------------------------------------------
   Name:        Linkage::Linkage                               (Constructor)
   ------------------------------------------------------------------------- */
Linkage::Linkage() {}

/* -------------------------------------------------------------------------
   Name:        Linkage::~Linkage                               (Destructor)
   ------------------------------------------------------------------------- */
Linkage::~Linkage() {}

/* -------------------------------------------------------------------------
   Name:        Head::Head                                     (Constructor)
   ------------------------------------------------------------------------- */
Head::Head () :
   first(0),
   last(0)
{
}
/*Head::Head*/

/* -------------------------------------------------------------------------
   Name:        Head::~Head                                     (Destructor)

   Purpose:     Remove elements from the link list.
   ------------------------------------------------------------------------- */
Head::~Head ()
{
   Clear();
}
/*Head::~Head*/

/* -------------------------------------------------------------------------
   Name:        Head::Suc

   Purpose:     Returns the first "Link" in the list.
   ------------------------------------------------------------------------- */
Link* Head::Suc() const
{
   return first;
}
/*Head::Suc*/

/* -------------------------------------------------------------------------
   Name:        Head::Pred

   Purpose:     Returns the last "Link" in the list.
   ------------------------------------------------------------------------- */
Link* Head::Pred() const
{
   return last;
}
/*Head::Pred*/

/* -------------------------------------------------------------------------
   Name:        Head::AddFirst

   Purpose:     Adds an element to the start of the list.  If the element is
                already in a list, it is removed before being added to this
                list.
   ------------------------------------------------------------------------- */
void Head::AddFirst (Link* element)
{
   if (!element)
      return;

   if (!first) {
      if (element->inList())
         element->Out();

      first = element;
      last = element;
      element->theList = this;
   } else {
      element->Precede(first);
      first = element;
   }
}
/*Head::AddFirst*/

/* -------------------------------------------------------------------------
   Name:        Head::AddLast

   Purpose:     Adds an element to the end of the list.  If the element is
                already in a list, it is removed before being added to this
                list.
   ------------------------------------------------------------------------- */
void Head::AddLast (Link* element)
{
   if (!element)
      return;

   if (last) {
      element->Follow(last);
      last = element;

   } else {
      if (element->inList())
         element->Out();

      first = element;
      last = element;
      element->theList = this;
   }
}
/*Head::AddLast*/

/* -------------------------------------------------------------------------
   Name:        Head::Cardinal

   Purpose:     Returns the number of elements in the list.
   ------------------------------------------------------------------------- */
long Head::Cardinal() const
{
   long  numberOfElements = 0;
   Link *tempPtr          = first;

   while (tempPtr) {
      ++numberOfElements;
      tempPtr = tempPtr->Suc();
   }

   return numberOfElements;
}
/*Head::Cardinal*/

/* -------------------------------------------------------------------------
   Name:        Head::Clear

   Purpose:     Removes all elements from the list.
   ------------------------------------------------------------------------- */
void Head::Clear()
{
   Link *tempPtr = first;
   Link *marker;

   while (tempPtr) {
      marker = tempPtr;
      tempPtr = tempPtr->Suc();
      delete marker;
   }

   first = last = 0;
}
/*Head::Clear*/

/* -------------------------------------------------------------------------
   Name:        Head::First

   Purpose:     Returns the first "Link" in the list.
   ------------------------------------------------------------------------- */
Link* Head::First() const
{
    return Suc();
}
/*Head::First*/

/* -------------------------------------------------------------------------
   Name:        Head::Last

   Purpose:     Returns the last "Link" in the list.
   ------------------------------------------------------------------------- */
Link* Head::Last() const
{
    return Pred();
}
/*Head::Last*/

/* -------------------------------------------------------------------------
   Name:        Head::Empty

   Purpose:     Determines if the list is empty.
   ------------------------------------------------------------------------- */
bool Head::Empty() const
{
    return (Cardinal() == 0);
}
/*Head::Empty*/



/* -------------------------------------------------------------------------
   Name:        Link::Link                                     (Constructor)

   Purpose:     Initialize the link as not belonging to any list.
   ------------------------------------------------------------------------- */
Link::Link() :
   prev(0),
   next(0),
   theList(0)
{
}
/*Link::Link*/

/* -------------------------------------------------------------------------
   Name:        Link::~Link                                     (Destructor)

   Purpose:     Before destroying link, ensure that is doesn't belong to any
                list.  If it does, remove it from the list.
   ------------------------------------------------------------------------- */
Link::~Link()
{
   RemoveElement();
}
/*Link::~Link*/

/* -------------------------------------------------------------------------
   Name:        Link::Pred

   Purpose:     Returns the previous "Link" in the list.
   ------------------------------------------------------------------------- */
Link* Link::Pred() const
{
    return prev;
}
/*Link::Pred*/

/* -------------------------------------------------------------------------
   Name:        Link::Suc

   Purpose:     Returns the next "Link" in the list.
   ------------------------------------------------------------------------- */
Link* Link::Suc() const
{
    return next;
}
/*Link::Suc*/

/* -------------------------------------------------------------------------
   Name:        Link::Out

   Purpose:     Remove "this" link from any list, return pointer to self.
   ------------------------------------------------------------------------- */
Link* Link::Out()
{
   RemoveElement();
   return this;
}
/*Link::Out*/

/* -------------------------------------------------------------------------
   Name:        Link::InTo

   Purpose:     Add "this" link to the bottom of a list.

                If list parameter is NULL, remove ourself from current list.
   ------------------------------------------------------------------------- */
void Link::InTo (Head* list)
{
   if (list) {
      list->AddLast (this);
      theList = list;
   } else
      Out();
}
/*Link::InTo*/

/* -------------------------------------------------------------------------
   Name:        Link::Precede

   Purpose:     Add "this" link to the list that "toPrecede" belongs to and
                place "this" link before "toPrecede".  If "this" link is
                already in a list, it is removed before being added to the
                "toPrecede" list.

                If "this" link belongs to a list but the "toPrecede" link is
                non-existent or doesn't belong to a list, "this" link is
                still removed from the list.
   ------------------------------------------------------------------------- */
void Link::Precede (Link* toPrecede)
{
   if (!toPrecede || !toPrecede->inList())
      Out();
   else {
      if (inList())
         Out();

      toPrecede->addBefore (this);
   }
}
/*Link::Precede*/

/* -------------------------------------------------------------------------
   Name:        Link::Precede

   Purpose:     Add "this" link to the end of the list.  If "this" link
                belongs to a list, it is removed before being added to the
                list.

                If "this" link belongs to a list but the new list is
                non-existent, "this" link is still removed from the list.
   ------------------------------------------------------------------------- */
void Link::Precede (Head* list)
{
    this->InTo (list);
}
/*Link::Precede*/

/* -------------------------------------------------------------------------
   Name:        Link::Follow

   Purpose:     Add "this" link to the list that "toFollow" belongs to and
                place "this" link after "toFollow".  If "this" link is
                already in a list, it is removed before being added to the
                "toFollow" list.

                If "this" link belongs to a list but the "toFollow" link is
                non-existent or doesn't belong to a list, "this" link is
                still removed from the list.
   ------------------------------------------------------------------------- */
void Link::Follow (Link* toFollow)
{
   if (!toFollow || !toFollow->inList())
      Out();
   else {
      if (inList())
         Out();

      toFollow->addAfter (this);
   }
}
/*Link::Follow*/

/* -------------------------------------------------------------------------
   Name:        Link::Follow

   Purpose:     Add "this" link to the start of the list.  If "this" link
                belongs to a list, it is removed before being added to the
                list.

                If "this" link belongs to a list but the new list is
                non-existent, "this" link is still removed from the list.
   ------------------------------------------------------------------------- */
void Link::Follow (Head* list)
{
   if (list)
      list->AddFirst (this);
   else
      Out();
}
/*Link::Follow*/

/* -------------------------------------------------------------------------
   Name:        Link::inList

   Purpose:     If this link in in a list return true otherwise false.
   ------------------------------------------------------------------------- */
bool Link::inList() const
{
    return (theList != 0);
}
/*Link::inList*/

/* -------------------------------------------------------------------------
   Name:        Link::RemoveElement

   Purpose:     Removes an element from a list.

                If the element belongs to the list, pointers are updated to
                ensure the list remains intact.
   ------------------------------------------------------------------------- */
void Link::RemoveElement()
{
   if (!theList)                       // If we don't belong to a list, ...
      return;                          // ... no need to continue.

   if (prev)                           // Link previous element to next.
      prev->next = next;

   if (next)                           // Link next element to previous.
      next->prev = prev;

   if (theList->first == this)         // If we are first element, ...
      theList->first = next;           // ... make next element first.

   if (theList->last == this)          // If we are last element, ...
      theList->last = prev;            // ... make previous element last.

   theList = 0;                        // We no longer belong to a list.
   prev = next = 0;                    // No previous or next elements.
}
/*Link::RemoveElement*/

/* -------------------------------------------------------------------------
   Name:        Link::addAfter                                       (local)

   Purpose:     Adds the "toAdd" link after "this" link and adjusts link
                references accordingly.

   Notes:       - This function is only called if "this" link belongs to a
                  list.
                - This function is only called if "toAdd" link does not
                  belong to a list.
   ------------------------------------------------------------------------- */
void Link::addAfter (Link* toAdd)
{
   toAdd->prev = this;
   toAdd->theList = theList;

   if (!next)
      next = toAdd;
   else {
      next->prev = toAdd;
      toAdd->next = next;
      next = toAdd;
   }

   if (theList->last == this)
      theList->last = toAdd;
}
/*Link::addAfter*/

/* -------------------------------------------------------------------------
   Name:        Link::addBefore

   Purpose:     Adds the "toAdd" link after "this" link and adjusts link
                references accordingly.

   Notes:       - This function is only called if "this" link belongs to a
                  list.
                - This function is only called if "toAdd" link does not
                  belong to a list.
   ------------------------------------------------------------------------- */
void Link::addBefore (Link* toAdd)
{
   toAdd->theList = theList;
   toAdd->next = this;

   if (!prev)
      prev = toAdd;
   else {
      prev->next = toAdd;
      toAdd->prev = prev;
      prev = toAdd;
   }

   if (theList->first == this)
      theList->first = toAdd;
}
/*Link::addBefore*/

